/**
 * Copyright (c) 2001-2006 eXo Platform SARL, http://www.exoplatform.com/
 *
 * All rights reserved. Please look at license.txt in info directory for
 * more license detail.
 *
 * -------------------------------------------------------------------------
 * $Id: RSASigner.java 10718 2006-11-29 02:54:54Z rmarins $
 * -------------------------------------------------------------------------
 */
package org.exoplatform.applications.signerclient.crypto.jce;

import java.io.IOException;
import java.security.InvalidKeyException;
import java.security.PrivateKey;
import java.security.PublicKey;
import java.security.SignatureException;
import java.security.interfaces.RSAPrivateKey;
import java.security.interfaces.RSAPublicKey;

import org.bouncycastle.asn1.ASN1Encodable;
import org.bouncycastle.asn1.DERObjectIdentifier;
import org.bouncycastle.asn1.nist.NISTObjectIdentifiers;
import org.bouncycastle.asn1.pkcs.PKCSObjectIdentifiers;
import org.bouncycastle.asn1.x509.AlgorithmIdentifier;
import org.bouncycastle.asn1.x509.DigestInfo;
import org.bouncycastle.asn1.x509.X509ObjectIdentifiers;
import org.bouncycastle.crypto.AsymmetricBlockCipher;
import org.bouncycastle.crypto.CipherParameters;
import org.bouncycastle.crypto.Digest;
import org.bouncycastle.crypto.encodings.PKCS1Encoding;
import org.bouncycastle.crypto.engines.RSAEngine;

/**
 * @version $Rev: 10718 $ $Date: 2006-11-29 09:54:54 +0700 (Wed, 29 Nov 2006) $
 * @author <a href="mailto:rafael.marins@exoplatform.com">Rafael Marins</a>
 */
class RSASigner implements PKCSObjectIdentifiers, X509ObjectIdentifiers, Signer {

	private String algorithm;

	private Digest digest;

	private AsymmetricBlockCipher cipher;

	private AlgorithmIdentifier algId;

	protected RSASigner(String algorithm, DERObjectIdentifier objId, Digest digest,
	        AsymmetricBlockCipher cipher) {
		super();

		this.algorithm = algorithm;
		this.digest = digest;
		this.cipher = new PKCS1Encoding(new RSAEngine());
		this.algId = new AlgorithmIdentifier(objId, null);
	}

	/* (non-Javadoc)
     * @see org.exoplatform.applications.signerclient.crypto.bc.Signer#getAlgorithm()
     */
    public String getAlgorithm() {
	    return algorithm;
    }

	/* (non-Javadoc)
     * @see org.exoplatform.applications.signerclient.crypto.bc.Signer#initVerify(java.security.PublicKey)
     */
	public void initVerify(PublicKey publicKey)
	        throws InvalidKeyException {
		if (!(publicKey instanceof RSAPublicKey)) {
			throw new InvalidKeyException("Supplied key (" + getType(publicKey)
			        + ") is not a RSAPublicKey instance");
		}

		CipherParameters param = RSAUtil
		        .generatePublicKeyParameter((RSAPublicKey) publicKey);

		digest.reset();
		cipher.init(false, param);
	}

	/* (non-Javadoc)
     * @see org.exoplatform.applications.signerclient.crypto.bc.Signer#initSign(java.security.PrivateKey)
     */
	public void initSign(PrivateKey privateKey)
	        throws InvalidKeyException {
		if (!(privateKey instanceof RSAPrivateKey)) {
			throw new InvalidKeyException("Supplied key ("
			        + getType(privateKey) + ") is not a RSAPrivateKey instance");
		}

		CipherParameters param = RSAUtil
		        .generatePrivateKeyParameter((RSAPrivateKey) privateKey);

		digest.reset();

		cipher.init(true, param);
	}

	private String getType(Object o) {
		if (o == null) {
			return null;
		}

		return o.getClass().getName();
	}

	/* (non-Javadoc)
     * @see org.exoplatform.applications.signerclient.crypto.bc.Signer#sign()
     */
	public byte[] sign() throws SignatureException {
		byte[] hash = new byte[digest.getDigestSize()];

		digest.doFinal(hash, 0);

		try {
			byte[] bytes = derEncode(hash);

			return cipher.processBlock(bytes, 0, bytes.length);
		} catch (ArrayIndexOutOfBoundsException e) {
			throw new SignatureException("key too small for signature type");
		} catch (Exception e) {
			throw new SignatureException(e.toString());
		}
	}

	/* (non-Javadoc)
     * @see org.exoplatform.applications.signerclient.crypto.bc.Signer#verify(byte[])
     */
	public boolean verify(byte[] sigBytes) throws SignatureException {
		byte[] hash = new byte[digest.getDigestSize()];

		digest.doFinal(hash, 0);

		byte[] sig;
		byte[] expected;

		try {
			sig = cipher.processBlock(sigBytes, 0, sigBytes.length);

			expected = derEncode(hash);
		} catch (Exception e) {
			return false;
		}

		if (sig.length == expected.length) {
			for (int i = 0; i < sig.length; i++) {
				if (sig[i] != expected[i]) {
					return false;
				}
			}
		} else if (sig.length == expected.length - 2) // NULL left out
		{
			int sigOffset = sig.length - hash.length - 2;
			int expectedOffset = expected.length - hash.length - 2;

			expected[1] -= 2; // adjust lengths
			expected[3] -= 2;

			for (int i = 0; i < hash.length; i++) {
				if (sig[sigOffset + i] != expected[expectedOffset + i]) // check
																		// hash
				{
					return false;
				}
			}

			for (int i = 0; i < sigOffset; i++) {
				if (sig[i] != expected[i]) // check header less NULL
				{
					return false;
				}
			}
		} else {
			return false;
		}

		return true;
	}

	private byte[] derEncode(byte[] hash) throws IOException {
		DigestInfo dInfo = new DigestInfo(algId, hash);

		return dInfo.getEncoded(ASN1Encodable.DER);
	}

//	private static PKCS1Encoding getPKCS1EncodingWithRSAEngine() {
//		final RSAEngine rsaEngine = new RSAEngine();
//		PKCS1Encoding encoding = (PKCS1Encoding) AccessController.doPrivileged(new PrivilegedAction() {
//    		public Object run() {
//    			return new PKCS1Encoding(rsaEngine);
//    		}
//    	});
//    	return encoding;
//	}

	static public class SHA1WithRSAEncryption extends RSASigner {
		public SHA1WithRSAEncryption(byte[] hash) {
			super("SHA1withRSA", id_SHA1, new EnclosingDigest(hash, "SHA1"),
					new PKCS1Encoding(new RSAEngine()));
		}
	}

	static public class SHA224WithRSAEncryption extends RSASigner {
		public SHA224WithRSAEncryption(byte[] hash) {
			super("SHA224withRSA", NISTObjectIdentifiers.id_sha224,
					 new EnclosingDigest(hash, "SHA224"), new PKCS1Encoding(new RSAEngine()));
		}
	}

	static public class SHA256WithRSAEncryption extends RSASigner {
		public SHA256WithRSAEncryption(byte[] hash) {
			super("SHA256withRSA", NISTObjectIdentifiers.id_sha256,
					 new EnclosingDigest(hash, "SHA256"), new PKCS1Encoding(new RSAEngine()));
		}
	}

	static public class SHA384WithRSAEncryption extends RSASigner {
		public SHA384WithRSAEncryption(byte[] hash) {
			super("SHA384withRSA", NISTObjectIdentifiers.id_sha384,
					 new EnclosingDigest(hash, "SHA384"), new PKCS1Encoding(new RSAEngine()));
		}
	}

	static public class SHA512WithRSAEncryption extends RSASigner {
		public SHA512WithRSAEncryption(byte[] hash) {
			super("SHA512withRSA", NISTObjectIdentifiers.id_sha512,
					 new EnclosingDigest(hash, "SHA512"), new PKCS1Encoding(new RSAEngine()));
		}
	}

	static public class MD2WithRSAEncryption extends RSASigner {
		public MD2WithRSAEncryption(byte[] hash) {
			super("MD2withRSA", md2,  new EnclosingDigest(hash, "MD2"), new PKCS1Encoding(
			        new RSAEngine()));
		}
	}

	static public class MD4WithRSAEncryption extends RSASigner {
		public MD4WithRSAEncryption(byte[] hash) {
			super("MD4withRSA", md4,  new EnclosingDigest(hash, "MD4"), new PKCS1Encoding(
			        new RSAEngine()));
		}
	}

	static public class MD5WithRSAEncryption extends RSASigner {
		public MD5WithRSAEncryption(byte[] hash) {
			super("MD5withRSA", md5,  new EnclosingDigest(hash, "MD5"), new PKCS1Encoding(
			        new RSAEngine()));
		}
	}

}
