/**
 * Copyright 2001-2006 The eXo Platform SARL         All rights reserved.  *
 * Please look at license.txt in info directory for more license detail.   *
 */

package org.exoplatform.services.jcr.impl.core;

import javax.jcr.AccessDeniedException;
import javax.jcr.InvalidItemStateException;
import javax.jcr.Item;
import javax.jcr.ItemNotFoundException;
import javax.jcr.Node;
import javax.jcr.PathNotFoundException;
import javax.jcr.ReferentialIntegrityException;
import javax.jcr.RepositoryException;
import javax.jcr.Session;
import javax.jcr.lock.LockException;
import javax.jcr.nodetype.ConstraintViolationException;
import javax.jcr.nodetype.ItemDefinition;
import javax.jcr.version.VersionException;

import org.apache.commons.logging.Log;
import org.exoplatform.services.jcr.datamodel.InternalQPath;
import org.exoplatform.services.jcr.datamodel.ItemData;
import org.exoplatform.services.jcr.impl.core.value.ValueFactoryImpl;
import org.exoplatform.services.jcr.impl.dataflow.TransientItemData;
import org.exoplatform.services.log.ExoLogger;

/**
 * Created by The eXo Platform SARL .
 * 
 * @author Gennady Azarenkov
 * @version $Id: ItemImpl.java 9819 2006-10-26 17:44:01Z geaz $
 */

public abstract class ItemImpl implements Item {

  protected static Log log = ExoLogger.getLogger("jcr.ItemImpl");
  
  protected final SessionImpl session;
  
  // ///////
  protected JCRPath location;

  //protected String internalUUID;

  protected NodeImpl parent;

  protected int state;

  protected int persistedVersion;

  // ///////

  // used for lazy initialization nodes/properties/parent
  protected ItemData data;

  protected SessionDataManager dataManager;

  protected LocationFactory locationFactory;

  protected ValueFactoryImpl valueFactory;

  protected boolean refreshToPersistedVersion; // [PN]

  protected long accessTime; // [PN]
  
  protected int itemHashCode = 0; // [PN] 19.04.06

  // new impl 
  ItemImpl(ItemData data, SessionImpl session) throws RepositoryException {

    this.session = session;
    this.data = data;
    this.location = session.getLocationFactory().createJCRPath(
        getData().getQPath());

    this.dataManager = session.getTransientNodesManager();
    this.locationFactory = session.getLocationFactory();
    this.valueFactory = (ValueFactoryImpl) session.getValueFactory();

    itemHashCode = (session.getWorkspace().getName() + data.getUUID())
        .hashCode();
  }

  public ItemData getData() {
    return data;
  }
  
  public String getParentUUID() {
    return getData().getParentUUID();
  }
  
  public InternalQPath getInternalPath() {
    return getData().getQPath();
  }
  
  protected ItemImpl item(JCRPath path) throws RepositoryException {
    return dataManager.getItem(path.getInternalPath(), true);
  }
  
  protected ItemImpl item(InternalQPath path) throws RepositoryException {
    return dataManager.getItem(path, true);
  }

  protected ItemImpl item(String uuid) throws RepositoryException {
    return dataManager.getItemByUUID(uuid, true);
  }

  protected NodeImpl parent() throws RepositoryException {
    return (NodeImpl) item(getData().getParentUUID());
  }


  ItemImpl(ItemData data, NodeImpl parentNode, int state, SessionImpl session)
      throws RepositoryException {
    
    this.session = session;
    this.state = state;
    //this.internalUUID = data.getUUID();
    this.persistedVersion = data.getPersistedVersion();

    this.location = session.getLocationFactory().createJCRPath(data.getQPath());
    this.parent = parentNode;

    this.data = data;

    this.dataManager = session.getTransientNodesManager();
    this.locationFactory = session.getLocationFactory();
    this.valueFactory = (ValueFactoryImpl) session.getValueFactory();
    
    // [PN] 19.04.06
    itemHashCode = (session.getWorkspace().getName() + getInternalUUID()).hashCode();
    //log.info("hashCode() INIT: " + getPath() + " " + session.getWorkspace().getName() + getInternalUUID() + " --> " + itemHashCode);
  }
  



  /* (non-Javadoc)
   * @see javax.jcr.Item#getPath()
   */
  public String getPath() {
    return getLocation().getAsString(false);
  }

  /* (non-Javadoc)
   * @see javax.jcr.Item#getName()
   */
  public String getName() {
    return getLocation().getName().getAsString();
  }

  /* (non-Javadoc)
   * @see javax.jcr.Item#getAncestor(int)
   */
  public Item getAncestor(int degree) throws ItemNotFoundException, AccessDeniedException,
      RepositoryException {
    try {
      // 6.2.8 If depth > n is specified then an ItemNotFoundException is
      // thrown.
      if (degree < 0)
        throw new ItemNotFoundException("Workspace.getAncestor() ancestor's degree < 0.");
      int n = getDepth() - degree;
      // log.debug("Item.getAncestor(" + degree + ") depth " +
      // getDepth()+" "+n);
      if (n == 0)
        return this;
      else if (n < 0)
        throw new ItemNotFoundException(
            "Workspace.getAncestor() ancestor's degree > depth of this item.");
      else {

        JCRPath ancestorLoc = getLocation().makeAncestorPath(n);

        return findItemByPath(ancestorLoc);
      }
    } catch (PathNotFoundException e) {
      throw new ItemNotFoundException(e.getMessage(), e);
    }
  }

  /* (non-Javadoc)
   * @see javax.jcr.Item#getParent()
   */
  public Node getParent() throws ItemNotFoundException, AccessDeniedException, RepositoryException {
    if (isRoot())
      throw new ItemNotFoundException("Root node does not have a parent");

    if (this.parent == null) {
      parent = (NodeImpl) findItemByPath(getLocation().makeParentPath());
    }

    if (this.parent == null) {
      reloadIfReferenceable();      
    }

    if (parent == null) {
      throw new RepositoryException("Parent not found for " + this.getPath() + " state: "
          + getState() + " at workspace: " + session.getWorkspace().getName());
    }
    return parent;

  }
  
  /* (non-Javadoc)
   * @see javax.jcr.Item#getSession()
   */
  public Session getSession() {
    return session;
  }

  /* (non-Javadoc)
   * @see javax.jcr.Item#getDepth()
   */
  public int getDepth() {
    return getLocation().getDepth();
  }

  /* (non-Javadoc)
   * @see javax.jcr.Item#isSame(javax.jcr.Item)
   */
  public boolean isSame(Item otherItem) {
    if (otherItem == null)
      return false;

    if (!this.getClass().equals(otherItem.getClass())) 
      return false;

    try {
      
      // UUID is not changed on ItemImpl
      return getInternalUUID().equals(((ItemImpl) otherItem).getInternalUUID());
    } catch (Exception e) {
      log.debug("Item.isSame() failed " + e);
      return false;
    }
  }

  /* (non-Javadoc)
   * @see javax.jcr.Item#isNew()
   */
  public boolean isNew() {
    return (getState() == ItemChangeState.ADDED);
  }

  /* (non-Javadoc)
   * @see javax.jcr.Item#remove()
   */
  public void remove() throws RepositoryException, ConstraintViolationException, VersionException,
      LockException {

    doRemove(true);
  }

  /* (non-Javadoc)
   * @see javax.jcr.Item#save()
   */
  public void save() throws ReferentialIntegrityException, AccessDeniedException, LockException,
      ConstraintViolationException, InvalidItemStateException, ReferentialIntegrityException,
      RepositoryException {

    if (isNew())
      throw new RepositoryException("It is impossible to call save() on the newly added item "
          + getPath());
    session.getTransientNodesManager().commit(this);
  }
  
  /* (non-Javadoc)
   * @see javax.jcr.Item#refresh(boolean)
   */
  public void refresh(boolean keepChanges) throws InvalidItemStateException, RepositoryException {
    if (!keepChanges)
      session.getTransientNodesManager().rollback(this);
    else
      session.getTransientNodesManager().refresh(this);
  }
  
//// ------------------ Impl
  
  void loadLocation(JCRPath location) {
    this.location = location;
  }

  
  /** 
   * For referenceable node item trying reload item by his internal uuid
   * For property item trying find parent by data's parent uuid and if 
   * it's referenceable will reload the property then
   * */
  boolean reloadIfReferenceable() throws RepositoryException {
    
      if (isNode()) {
        // It's a node
        NodeImpl thisNode = (NodeImpl) this;
        if (thisNode.isNodeType("mix:referenceable")) {
          // Ok, it's referenceable node. Trying retreive this node from workspace
          NodeImpl me = (NodeImpl) findItemByUUID(getInternalUUID());
          if (me != null && me != thisNode) {
            // This node live in workspace, get parent
            //log.info("===== Reloading referenceable node: " + me.getPath() + "  " + getInternalUUID());
            try {
              loadData(me.storedNodeData()); // Data from WDM
              log.info("===== Reload referenceable node DONE: " + me.getPath() + "  " + getInternalUUID());
              return true;
            } catch (RepositoryException e) {
              // Sorry, no parent available. Wrong state:(
            }            
          }
        }
      } else if (data != null) {
        // It's a property. Try find parent by data's uuid
        NodeImpl myParent = (NodeImpl) findItemByUUID(data.getParentUUID());
        if (myParent != null && myParent != parent && myParent.isNodeType("mix:referenceable")) {
          // So, the uuid is true and if parent is referenceable- load it
          parent = myParent;
          JCRPath myLocation = session.getLocationFactory()
            .createJCRPath(myParent.getLocation(), getName());
          loadLocation(myLocation);
          ItemData itemData = session.getTransientNodesManager().getTransactManager().getItemData(myLocation.getInternalPath());
          if (itemData != null) {
            loadData(itemData);
          }
          //log.info("===== Reload property with referenceable parent DONE: " + getPath() + "  " + getActualItemData().getUUID());
          return true;
        }
      }
      return false;    
  }

  
  /**
   * For internal use;
   * Remove from parent and mark as DELETED 
   * @author [PN] 11.04.06
   * */
  public void removeFromParent() throws RepositoryException, LockException {
    removeFromParent(ItemChangeState.DELETED);
  }
  
  /**
   * For internal use;
   * Remove from parent and set state 
   * @author [PN] 11.04.06
   * */  
  public void removeFromParent(int setState) throws RepositoryException, LockException {
    if (isRoot())
      throw new RepositoryException("Can't remove ROOT node.");

    parent = (NodeImpl) getParent();
    
    // Check locking
    if (!parent.checkLocking())
      throw new LockException("Node " + parent.getPath() + " is locked ");

    // [2] remove from parent
    if (isNode()) {
      if (parent.removeChildNode(this) != null) {
        parent.reorderChildNodes(true); // in case updates (reordered), do it direct from storage.
      }
    } else {
      parent.removeChildProperty(this);
    }
    state = setState;
  }  
  
  /**
   * @param externalValidation
   * @throws RepositoryException
   * @throws ConstraintViolationException
   * @throws VersionException
   * @throws LockException
   */
  public void doRemove(boolean externalValidation) throws RepositoryException,
      ConstraintViolationException, VersionException, LockException {

    if (isRoot())
      throw new RepositoryException("Can't remove ROOT node.");

    // Check constraints
    if (externalValidation) {
      ItemDefinition def;
      if (isNode())
        def = ((NodeImpl) this).getDefinition();
      else
        def = ((PropertyImpl) this).getDefinition();

      if (def.isMandatory() || def.isProtected())
        throw new ConstraintViolationException("Can't remove mandatory or protected item "
            + getPath());
    }
    // Check if versionable ancestor is not checked-in
    parent = (NodeImpl) getParent();
    if (!parent.isCheckedOut())
      throw new VersionException("Node " + parent.getPath()
          + " or its nearest ancestor is checked-in");

    // Check locking
    if (!parent.checkLocking())
      throw new LockException("Node " + parent.getPath() + " is locked ");

    // [2] remove from parent
    if (isNode()) {
      parent.removeChildNode(this);
    } else {
      parent.removeChildProperty(this);
    }

    // [1] remove from datamanager
    dataManager.delete(this);
  }


  // -------------------- Implementation ----------------------

  public abstract TransientItemData getActualItemData() throws RepositoryException;

  public String getInternalUUID() {
    return data.getUUID();
  }

  public JCRPath getLocation() {
    return this.location;
  }

  public boolean isRoot() {
    return (getDepth() == 0);
  }

  public int getState() {
    return state;
  }

  /**
   * @param newState The state to set.
   */
  protected void setState(int newState) {
    // old new result
    // A -> U = A
    // D -> ? = D
    // D -> A = A
    // A -> D = U

    int origState = newState;
    if ((this.state == ItemChangeState.ADDED || this.state == ItemChangeState.AUTO_ADDED)
        && newState == ItemChangeState.UPDATED) {
      newState = state; // not change
    } else if (this.state == ItemChangeState.DELETED || this.state == ItemChangeState.AUTO_DELETED) {
      // [PN] 04.04.06, 12.04.06
      if (newState != ItemChangeState.UNCHANGED && newState != ItemChangeState.AUTO_DELETED) {
        newState = state;
      }
    } else if (this.state == ItemChangeState.ADDED && newState == ItemChangeState.DELETED) {
      newState = ItemChangeState.UNCHANGED;
    }

    this.state = newState;
    if (log.isDebugEnabled())
      log.debug("Setting state: " + ItemChangeState.getStateName(origState) 
          + " set: " + ItemChangeState.getStateName(newState) + " " + getPath());
  }

//  // set state direct, no setState() usage
//  @Deprecated
//  protected void markUpdated() {
//    state = ItemChangeState.UPDATED;
//  }

  public boolean isRefreshToPersistedVersion() {
    return refreshToPersistedVersion;
  }

  public void setRefreshToPersistedVersion(boolean refreshToPersistedVersion) {
    this.refreshToPersistedVersion = refreshToPersistedVersion;
  }

  abstract void loadData(ItemData data) throws RepositoryException;

  public void setPersistedVersion(int persistedVersion) {
    this.persistedVersion = persistedVersion;
  }

  public boolean hasPermission(String action) throws RepositoryException {

//    // versionStorage's children should have READ permission for all
//    if (isRoot()
//        || (location.isDescendantOf(locationFactory.parseAbsPath("/jcr:system/jcr:versionStorage"),
//            false) && action.equals(PermissionType.READ)))
//      return true;
//    
//    else if (parent == null) {
//      parent = (NodeImpl) this.dataManager.getItemByUUID(data.getParentUUID());      
//    }
//
//    if (parent == null) {
//      parent = (NodeImpl) getParent();
//    }
//    if (parent == null) {
//      throw new RepositoryException("FATAL: parent not found for " + this.getPath());
//    }
//
//    boolean hasPerm = session.getAccessManager().hasPermission(
//        ((NodeImpl) parent).getACL(), action, session.getUserID());
//    return hasPerm;
    
//  CODE below CAUSES 1 TCK fails (WITH code above IT IS OK):
//    junit.framework.AssertionFailedError: Workspace.restore(Version[], boolean) with a version that has no corresponding node must succeed if a version of a 
//    parent with correspondance is present in the version array.
//    at junit.framework.Assert.fail(Assert.java:47)
//    at org.apache.jackrabbit.test.api.version.WorkspaceRestoreTest.testWorkspaceRestoreWithParent(WorkspaceRestoreTest.java:174)
    
    NodeImpl testNode;
    if(isNode())
      testNode = (NodeImpl)this;
    else {
      testNode = (NodeImpl) dataManager.getItemByUUID(data.getParentUUID());
      // possible in a case if new parent node is not valid
      if (testNode == null) {
        testNode = (NodeImpl) getParent();
      }
      if (testNode == null) {
        throw new RepositoryException("FATAL: parent not found for " + this.getPath());
      }
    }
    
    return session.getAccessManager().hasPermission(
     testNode.getACL(), action, session.getUserID());
  }


  public int getPersistedVersion() {
    return persistedVersion;
  }

  protected Item findItemByPath(JCRPath path) throws RepositoryException {
    return dataManager.getAccessibleItem(path);
  }

  protected Item findItemByUUID(String uuid) throws RepositoryException {
    return dataManager.getActualItemByUUID(uuid);
  }
  
  // new impl methods? ////////////////
  
  
  
  ////////////////////////////////

  
//  /**
//   * @return Returns the accessTime.
//   */
//  public long getAccessTime() {
//    return accessTime;
//  }
//
//  /**
//   * @param accessTime The accessTime to set.
//   */
//  public void setAccessTime(long accessTime) {
//    this.accessTime = accessTime;
//  }  

  /**
   * @author [PN] 18.04.06
   * @see java.lang.Object#equals(java.lang.Object)
   */
  @Override
  public boolean equals(Object obj) {
    if (obj instanceof ItemImpl) {
      ItemImpl otherItem = (ItemImpl) obj;
      try {
        return getInternalUUID().equals(otherItem.getInternalUUID());
      } catch(Exception e) {
        return false;
      }
    }
    return false;
  }

  /**
   * @see java.lang.Object#hashCode()
   */
  @Override
  public int hashCode() {
    return itemHashCode;
  }
  
  
}