/***************************************************************************
 * Copyright 2001-2003 The eXo Platform SARL         All rights reserved.  *
 * Please look at license.txt in info directory for more license detail.   *
 **************************************************************************/

package org.exoplatform.services.jcr.impl.core;

import java.io.IOException;
import java.io.InputStream;
import java.security.AccessControlException;
import java.util.ArrayList;
import java.util.Calendar;
import java.util.Collections;
import java.util.Comparator;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.concurrent.CopyOnWriteArrayList;

import javax.jcr.AccessDeniedException;
import javax.jcr.InvalidItemStateException;
import javax.jcr.Item;
import javax.jcr.ItemExistsException;
import javax.jcr.ItemNotFoundException;
import javax.jcr.ItemVisitor;
import javax.jcr.MergeException;
import javax.jcr.NoSuchWorkspaceException;
import javax.jcr.Node;
import javax.jcr.NodeIterator;
import javax.jcr.PathNotFoundException;
import javax.jcr.Property;
import javax.jcr.PropertyIterator;
import javax.jcr.PropertyType;
import javax.jcr.RepositoryException;
import javax.jcr.UnsupportedRepositoryOperationException;
import javax.jcr.Value;
import javax.jcr.ValueFormatException;
import javax.jcr.lock.Lock;
import javax.jcr.lock.LockException;
import javax.jcr.nodetype.ConstraintViolationException;
import javax.jcr.nodetype.ItemDefinition;
import javax.jcr.nodetype.NoSuchNodeTypeException;
import javax.jcr.nodetype.NodeDefinition;
import javax.jcr.nodetype.NodeType;
import javax.jcr.nodetype.PropertyDefinition;
import javax.jcr.observation.Event;
import javax.jcr.version.Version;
import javax.jcr.version.VersionException;
import javax.jcr.version.VersionHistory;

import org.apache.commons.logging.Log;
import org.exoplatform.services.jcr.access.AccessControlList;
import org.exoplatform.services.jcr.access.AccessControlPolicy;
import org.exoplatform.services.jcr.access.PermissionType;
import org.exoplatform.services.jcr.access.SystemIdentity;
import org.exoplatform.services.jcr.core.ExtendedNode;
import org.exoplatform.services.jcr.core.ExtendedPropertyType;
import org.exoplatform.services.jcr.datamodel.InternalQName;
import org.exoplatform.services.jcr.datamodel.ItemData;
import org.exoplatform.services.jcr.datamodel.NodeData;
import org.exoplatform.services.jcr.datamodel.PropertyData;
import org.exoplatform.services.jcr.datamodel.Uuid;
import org.exoplatform.services.jcr.impl.Constants;
import org.exoplatform.services.jcr.impl.core.itemfilters.ItemFilter;
import org.exoplatform.services.jcr.impl.core.itemfilters.NamePatternFilter;
import org.exoplatform.services.jcr.impl.core.lock.LockImpl;
import org.exoplatform.services.jcr.impl.core.nodetype.ItemDefinitionImpl;
import org.exoplatform.services.jcr.impl.core.nodetype.NodeDefinitionImpl;
import org.exoplatform.services.jcr.impl.core.nodetype.NodeTypeImpl;
import org.exoplatform.services.jcr.impl.core.nodetype.NodeTypeManagerImpl;
import org.exoplatform.services.jcr.impl.core.nodetype.PropertyDefinitionImpl;
import org.exoplatform.services.jcr.impl.core.nodetype.PropertyDefinitions;
import org.exoplatform.services.jcr.impl.core.value.PermissionValue;
import org.exoplatform.services.jcr.impl.core.version.MergeVisitor;
import org.exoplatform.services.jcr.impl.core.version.VersionHistoryImpl;
import org.exoplatform.services.jcr.impl.core.version.VersionImpl;
import org.exoplatform.services.jcr.impl.dataflow.TransientNodeData;
import org.exoplatform.services.jcr.impl.dataflow.TransientPropertyData;
import org.exoplatform.services.jcr.impl.util.EntityCollection;
import org.exoplatform.services.jcr.util.UUIDGenerator;
import org.exoplatform.services.log.ExoLogger;

/**
 * Created by The eXo Platform SARL.
 * 
 * @author <a href="mailto:geaz@users.sourceforge.net">Gennady Azarenkov </a>
 * @version $Id: NodeImpl.java 9853 2006-10-28 15:22:26Z geaz $
 */

public class NodeImpl extends ItemImpl implements ExtendedNode {
  
  protected static Log log = ExoLogger.getLogger("jcr.NodeImpl");

  protected LocationFactory sysLocFactory;

  protected UUIDGenerator   uuidGenerator;

  // //////
  private List              childNodes;

  private List              childProperties;

  private JCRName           primaryType;

  private JCRName[]         mixinTypes;

  private int               orderNumber;

  private int orderDirection = 0;

  private NodeDefinition    definition;
  
  /**
   * New constructor
   * @param id
   * @param session
   */
  public NodeImpl(NodeData data, SessionImpl session) throws RepositoryException {
    super(data, session);
    sysLocFactory = session.getSystemLocationFactory();
    loadData(data);
  }  

  
  public NodeImpl(NodeData data, NodeImpl parentNode, int state, SessionImpl session
      ) throws PathNotFoundException, RepositoryException,
      ConstraintViolationException {

    super(data, parentNode, state, session);
    // nodeType names should conform NamespaceRegistry's mapping
    sysLocFactory = session.getSystemLocationFactory();
    uuidGenerator = session.getUuidGenerator();
    loadData(data);
  }

  public void loadData(ItemData itemData) throws RepositoryException, InvalidItemStateException,
      ConstraintViolationException {

    if (itemData == null)
      throw new InvalidItemStateException("Data is null for " + this.getPath()
          + " Probably was deleted by another session and can not be loaded from container ");

    if (!(itemData instanceof NodeData))
      throw new RepositoryException("Load data: NodeData is expected. But have " + itemData);

    NodeData nodeData = (NodeData) itemData;
    if (nodeData.getPrimaryTypeName() == null)
      throw new RepositoryException("Load data: NodeData has no primaryTypeName. Null value found. " 
          + (nodeData.getQPath() != null ? nodeData.getQPath().getAsString() : "[null path node]") 
          + " " + nodeData);
    if (nodeData.getMixinTypeNames() == null)
      throw new RepositoryException("Load data: NodeData has no mixinTypeNames. Null value found. " 
          + (nodeData.getQPath() != null ? nodeData.getQPath().getAsString() : "[null path node]"));

    this.primaryType = sysLocFactory.createJCRName(nodeData.getPrimaryTypeName());
    this.mixinTypes = new JCRName[nodeData.getMixinTypeNames().length];
    for (int i = 0; i < mixinTypes.length; i++)
      mixinTypes[i] = sysLocFactory.createJCRName(nodeData.getMixinTypeNames()[i]);
    this.orderNumber = nodeData.getOrderNumber();
    //this.childrenACL = initACL(nodeData);
    this.childNodes = null;
    
    this.childProperties = null;

    this.persistedVersion = nodeData.getPersistedVersion();

    this.data = nodeData;

    if (isNodeType("mix:referenceable")) {
      // we have to check and reload actual location and parent if needed
      try {
        if (!getLocation().getInternalPath().equals(data.getQPath())) {
          // load new location
          JCRPath myLocation = session.getLocationFactory().createJCRPath(data.getQPath());
          loadLocation(myLocation);
          //log.info("=== Reload referenceable node location DONE: " + getPath());
          try {
            // log.info("=== Reloading referenceable node parent: " +
            // data.getParentUUID());
            NodeImpl myParent = (NodeImpl) findItemByUUID(data.getParentUUID());
            if (myParent != null && myParent != parent) {
              // load new parent
              parent = myParent;
              //log.info("=== Reload referenceable node parent DONE: " + parent.getPath() + "  " + parent.getInternalUUID());
            }
          } catch (RepositoryException e) {
            log.error("Error load parent for referenceable node: " + getPath(), e);
          }
        }
      } catch (RepositoryException e) {
        log.error("Error load location for referenceable node: " + getPath(), e);
      }
    }
  }

  private void initDefinition() throws RepositoryException, ConstraintViolationException {

    if (this.isRoot()) { // root - no parent
      NodeDefinitionImpl defNodeDef = new NodeDefinitionImpl();
      defNodeDef.setRequiredNodeTypes(new NodeType[] {nodeType(Constants.NT_BASE)});
      this.definition = defNodeDef;
      return;
    }

    NodeImpl parent = (NodeImpl) getParent();

    NodeType[] nodetypes = parent.getAllNodeTypes();
    NodeDefinition residualDef = null;
    for (int i = 0; i < nodetypes.length; i++) {

      NodeDefinition def = ((NodeTypeImpl) nodetypes[i]).getChildNodeDefinition(getName());

      if (def != null) {
        if (def.getName() == ItemDefinitionImpl.RESIDUAL_SET)
          residualDef = def;
        else {
          if (log.isDebugEnabled()) log.debug("Node definition found " + def.getName());
          this.definition = def;
          return;
          // return def;
        }
      }
    }

    if (residualDef != null)
      this.definition = residualDef;
    else
      throw new ConstraintViolationException(
          "NodeImpl.getDefinition failed. Definition not found for " + getPath());

  }

  private AccessControlList initACL(NodeData nodeData) {
    AccessControlList dataACL = nodeData.getACL();
    if (dataACL == null) { // default one
      throw new RuntimeException("ACL is NULL for "+getPath());
//      System.out.println(">>>>>>ACL is NULL for "+getPath());
//      return new AccessControlList(true);
    } else {
      return dataACL;
    }
  }

  public TransientNodeData getActualItemData() throws RepositoryException {

    return traverseData(this);
  }

  private synchronized TransientNodeData traverseData(NodeImpl node) throws RepositoryException {
    InternalQName[] mixinNames = new InternalQName[node.getMixinTypes().length];
    for (int i = 0; i < node.getMixinTypes().length; i++) {
      mixinNames[i] = node.getMixinTypes()[i].getInternalName();
    }
   
    TransientNodeData currentData = new TransientNodeData(node.getLocation().getInternalPath(),
        node.storedNodeData().getUUID(), node.getPersistedVersion(), node.getPrimaryType()
            .getInternalName(), mixinNames, node.getOrderNumber(), node.storedNodeData()
            .getParentUUID(), getOrderDirection(), node.getACL());

    // This flag will be switched off in node.getNodes() if on. 
    // Remember it for use if child's properties will be not initialized  
    //boolean reload = node.isRefreshToPersistedVersion();
    
    if (node.isChildPropertiesInitialized()) { //     
      PropertyIterator props = node.getProperties();
      while (props.hasNext()) {
        PropertyImpl prop = (PropertyImpl) props.next();
        currentData.addProperty((TransientPropertyData) prop.getActualItemData());
      }
    }
    
    if (node.isChildNodesInitialized()) {
      NodeIterator nodes = node.getNodes();
      while (nodes.hasNext()) {
        NodeImpl childNode = (NodeImpl) nodes.next();
        
        // [PN] 30.03.06 using refreshToPersistedVersion flag for all descendants
        boolean reloadState = childNode.isRefreshToPersistedVersion();
        try {
          childNode.setRefreshToPersistedVersion(isRefreshToPersistedVersion());
          //System.out.println("child "+childNode.getPath());
          if (childNode.getPath().equals(node.getPath()))
            throw new RepositoryException("FATAL: internal error child node path equals parent "
                + node.getPath());
          currentData.addNode(traverseData(childNode));
        } finally {
          childNode.setRefreshToPersistedVersion(reloadState);
        }
      }
    }
    
    return currentData;
  }

  public NodeData storedNodeData() {
    return (NodeData) data;
  }

  // /// ------------ Reading

  /**
   * Finds parent. It can be this node or this node's child.
   * 
   * @param itemPath
   * @return this node or find parent by path
   * @throws PathNotFoundException if not found
   */
  private NodeImpl findParent(JCRPath itemPath) throws RepositoryException, PathNotFoundException,
      ConstraintViolationException {

    JCRPath parentPath = itemPath.makeParentPath();
    if (getLocation().equals(parentPath))
      return this;
    else {
      ItemImpl parentNode = dataManager.getItem(parentPath);
      if (parentNode == null)
        throw new PathNotFoundException("Parent not found for " + itemPath.getAsString(false));
      else if (!parentNode.isNode())
        throw new ConstraintViolationException("Parent is a property " + parentNode.getPath());
      else
        return (NodeImpl) parentNode;
    }
  }

  /**
   * @see javax.jcr.Node#getNode
   */
  public Node getNode(String relPath) throws PathNotFoundException, RepositoryException {

    JCRPath itemPath = locationFactory.createJCRPath(getLocation(), relPath);
    NodeImpl myParent = findParent(itemPath);
    //log.info("getNode() " + itemPath.getAsString(false));

    for (NodeIterator nodes = myParent.getNodes(itemPath.getName().getAsString()); nodes.hasNext();) {
      Node nextNode = nodes.nextNode();
      if (nextNode.getIndex() == itemPath.getIndex())
        return nextNode;
    }
    throw new PathNotFoundException("Node not found " + itemPath.getAsString(true));
  }

  /**
   * @see javax.jcr.Node#getNodes
   */
  public NodeIterator getNodes() throws RepositoryException {

    loadChildNodes();
    return new EntityCollection(childNodes);
  }

//  public void clearChildNodes() {
//    childNodes = null;
//  }

  public List loadChildNodes() throws RepositoryException, AccessDeniedException {

    // [PN] I'm not sure that that will be right
    // A referenceable node or referenceable parent of property
    // will be reloaded every time entering here
    // reloadIfReferenceable();

    if (isRefreshToPersistedVersion()) {
      reloadNodeData();
      setRefreshToPersistedVersion(false);
    }
    // return if initialized
    if (childNodes == null) {
      childNodes = getChildNodes();
    }
    return childNodes;
  }

  /**
   * @return list with actual nodes, that stored in persisten storage
   */
  public List getChildNodes() throws RepositoryException, AccessDeniedException {

    List storedNodes = dataManager.getChildren(this, true);
    
    Collections.sort(storedNodes, new NodesOrderComparator());
    return new CopyOnWriteArrayList(storedNodes);
  }

  private static class NodesOrderComparator implements Comparator {
    public int compare(Object o1, Object o2) {
      NodeImpl n1 = (NodeImpl) o1;
      NodeImpl n2 = (NodeImpl) o2;
      return n1.getOrderNumber() - n2.getOrderNumber();
    }
  }

  /**
   * @see javax.jcr.Node#getNodes
   */
  public NodeIterator getNodes(String namePattern) throws RepositoryException {
    ItemFilter filter = new NamePatternFilter(namePattern);
    ArrayList list = new ArrayList();
    NodeIterator children = getNodes();
    // System.err.println(">>> getNodes >>");
    while (children.hasNext()) {

      NodeImpl item = (NodeImpl) children.nextNode();
      // System.err.println(">>> getNodes >>" +this.getPath()+" "+item);

      if (filter.accept(item))
        list.add(item);
    }
    return new EntityCollection(list);
  }

  /**
   * @see javax.jcr.Node#getProperty
   */
  public Property getProperty(String relPath) throws PathNotFoundException, RepositoryException {

    JCRPath itemPath = locationFactory.createJCRPath(getLocation(), relPath);
  
    //log.info("getProperty() " + itemPath.getAsString(false));
    
    // [PN] 15.04.06 Don't traversing all properties if need only one.
    // TODO But we have to search for nodes in both persistent/session storages. 
    // Search in parent props
    NodeImpl myParent = findParent(itemPath);

    PropertyIterator children = myParent.getProperties();
    while (children.hasNext()) {
      PropertyImpl item = (PropertyImpl) children.nextProperty();
      if (item.getActualItemData().getQPath().getName().equals(itemPath.getInternalPath().getName())) {
        // [PN] 19.04.06 Some magic for retrieve most actual item,
        // Use case: jcr:predecessors which can be changed direct in workspace and should be accessible
        // on the session's item immediately
        // TODO Do refactoring and remove.
        if (item.isNew()) {
          return item;
        }
        PropertyImpl sessionItem = (PropertyImpl) dataManager.getAccessibleItem(item.getLocation());
        if (sessionItem != null && sessionItem.getPersistedVersion() > item.getPersistedVersion()) {
          int itemIndex = childProperties.indexOf(item);
          if (itemIndex >= 0) {
            // [PN] Fixing on-node child properties list.
            childProperties.set(itemIndex, sessionItem);
          }
          return sessionItem;
        }
        return item;
      }
    }
    
    throw new PathNotFoundException("Property not found " + itemPath.getAsString(false));
  }

  /**
   * @see javax.jcr.Node#getProperties
   */
  public PropertyIterator getProperties() throws RepositoryException {
    
    loadChildProperties();
    return new EntityCollection(childProperties);
  }

  public List loadChildProperties() throws RepositoryException, AccessDeniedException {

    // not yet initialized
    if (this.childProperties == null) {
      childProperties = getChildProperties();
    }
    return childProperties;
  }

  /**
   * @return list with actual properties, that stored in persisten storage
   */
  public List getChildProperties() throws RepositoryException, AccessDeniedException {

    List storedNodes = dataManager.getChildren(this, false);
    // 0 - primaryType
    // 1 - mixinTypes
    // 2 - uuid
    // ....

    Collections.sort(storedNodes, new PropertiesOrderComparator());
    return new CopyOnWriteArrayList(storedNodes);
  }

  private static class PropertiesOrderComparator implements Comparator {
    public int compare(Object o1, Object o2) {
      int r = 0;
      try {
        PropertyImpl p1 = (PropertyImpl) o1;
        InternalQName qname1 = p1.getLocation().getName().getInternalName();
        PropertyImpl p2 = (PropertyImpl) o2;
        InternalQName qname2 = p2.getLocation().getName().getInternalName();
        if (qname1.equals(Constants.JCR_PRIMARYTYPE)) {
          r = Integer.MIN_VALUE;
        } else if (qname2.equals(Constants.JCR_PRIMARYTYPE)) {
          r = Integer.MAX_VALUE;
        } else if (qname1.equals(Constants.JCR_MIXINTYPES)) {
          r = Integer.MIN_VALUE + 1;
        } else if (qname2.equals(Constants.JCR_MIXINTYPES)) {
          r = Integer.MAX_VALUE - 1;
        } else if (qname1.equals(Constants.JCR_UUID)) {
          r = Integer.MIN_VALUE + 2;
        } else if (qname2.equals(Constants.JCR_UUID)) {
          r = Integer.MAX_VALUE - 2;
        } else {
          r = qname1.getAsString().compareTo(qname2.getAsString());
        }
      } catch (Exception e) {
        System.err.println("PropertiesOrderComparator error: " + e);
      }
      return r;
    }
  }

  /**
   * @see javax.jcr.Node#getProperties
   */
  public PropertyIterator getProperties(String namePattern) throws RepositoryException {
    ItemFilter filter = new NamePatternFilter(namePattern);
    ArrayList list = new ArrayList();
    PropertyIterator children = getProperties();
    while (children.hasNext()) {
      PropertyImpl item = (PropertyImpl) children.nextProperty();
      if (filter.accept(item))
        list.add(item);
    }

    return new EntityCollection(list);
  }

  /**
   * @see javax.jcr.Node#getIndex
   */
  public int getIndex() throws RepositoryException {
    return getLocation().getIndex();
  }

  /**
   * @see javax.jcr.Node#getPrimaryItem
   */
  public Item getPrimaryItem() throws ItemNotFoundException, RepositoryException {

    NodeType[] types = this.getAllNodeTypes();
    for (int i = 0; i < types.length; i++) {
      String itemName = types[i].getPrimaryItemName();
      if (itemName != null) {
        Item primaryItem = dataManager.getItem(locationFactory.createJCRPath(getLocation(),
            itemName));
        if (primaryItem != null)
          return primaryItem;
      }
    }

    throw new ItemNotFoundException("Primary item not found for " + getPath());
  }

  /**
   * @see javax.jcr.Node#getUUID
   */
  public String getUUID() throws UnsupportedRepositoryOperationException, RepositoryException {
    if (isNodeType("mix:referenceable")) {
      return this.getInternalUUID();
    } else {
      throw new UnsupportedRepositoryOperationException("Node " + getPath()
          + " is not referenceable");
    }

  }

  /**
   * Returns saved only references (allowed by specs)
   * 
   * @see javax.jcr.Node#getReferences
   */
  public PropertyIterator getReferences() throws RepositoryException {
    EntityCollection refs = new EntityCollection();
    for (PropertyData propertyData: dataManager.getReferences(getUUID())) {
      refs.add(new PropertyImpl((TransientPropertyData)propertyData, null, ItemChangeState.UNCHANGED, session));
    }
    return refs;
  }

  /**
   * @see javax.jcr.Node#hasNode
   */
  public boolean hasNode(String relPath) throws RepositoryException {
    try {
      getNode(relPath);
    } catch (PathNotFoundException e) {
      return false;
    }
    return true;
  }

  /**
   * @see javax.jcr.Node#hasNodes
   */
  public boolean hasNodes() throws RepositoryException {
    return getNodes().hasNext();
  }

  /**
   * @see javax.jcr.Node#hasProperty
   */
  public boolean hasProperty(String relPath) throws RepositoryException {
    try {
      getProperty(relPath);
      return true;
    } catch (RepositoryException e) {
      return false;
    }
  }

  /**
   * @see javax.jcr.Node#hasProperties
   */
  public boolean hasProperties() throws RepositoryException {
    return getProperties().hasNext();
  }

  /**
   * @see javax.jcr.Item#isModified
   */
  public boolean isModified() {
    return (getState() == ItemChangeState.UPDATED)
        || (!isNew() && session.getTransientNodesManager().hasAddedOrRemovedDescendant(
            getLocation(), true));
  }

  // // -------------- Writting

  /**
   * @see javax.jcr.Node#addNode
   */
  public Node addNode(String relPath) throws  PathNotFoundException,
      ConstraintViolationException, RepositoryException, VersionException, LockException {

    // Parent can be not the same as this node
    JCRPath itemPath = locationFactory.createJCRPath(getLocation(), relPath);

    // Check if there no final index
    if (itemPath.isIndexSetExplicitly())
      throw new RepositoryException(
          "The relPath provided must not have an index on its final element. "
              + itemPath.getAsString(false));

    NodeImpl parent = this.findParent(itemPath);
    // Check if parent exists
    if (parent == null)
      throw new PathNotFoundException("Parent not found for " + itemPath.getAsString(true));

    // find node type
    NodeType[] nodeTypes = parent.getAllNodeTypes();
    JCRName nodeTypeName = findNodeType(nodeTypes, itemPath.getName().getAsString());

    // try to make new node
    NodeImpl newNode = doAddNode(parent, itemPath.getName(), nodeTypeName, null, false, true);
    return newNode;

  }

  private JCRName findNodeType(NodeType[] nodeTypes, String name) throws RepositoryException,
      ConstraintViolationException {
    String residualNodeTypeName = null;
    for (int j = 0; j < nodeTypes.length; j++) {
      NodeDefinition[] nodeDefs = nodeTypes[j].getChildNodeDefinitions();
      for (int i = 0; i < nodeDefs.length; i++) {
        NodeDefinition nodeDef = nodeDefs[i];
        if (nodeDef.getName().equals(name)) {
          NodeType nt = nodeDef.getDefaultPrimaryType();
          // [PN] 25.09.06
          if (nt != null)
            return sysLocFactory.parseJCRName(nt.getName());
          // else: A node type of the new child node cannot be determined automatically and no primary node type is explicitly specified.
          // E.g. nt:file jcr:content with jcr:defaultPrimaryType = null
        } else if (nodeDef.getName().equals(NodeDefinitionImpl.RESIDUAL_SET)) {
//          System.err.println("nodeDef.getDefaultPrimaryType().getName(): " + nodeDef.getName()
//              + " " + nodeDef.getDefaultPrimaryType()
//              + " " + (nodeDef.getDefaultPrimaryType() != null ? nodeDef.getDefaultPrimaryType().getName() : "[null]"));
          residualNodeTypeName = nodeDef.getDefaultPrimaryType().getName();
        }
      }
    }
    if (residualNodeTypeName == null)
      throw new ConstraintViolationException("Can not define node type for " + name);
    return sysLocFactory.parseJCRName(residualNodeTypeName);
  }

  /**
   * @see javax.jcr.Node#addNode
   */
  public Node addNode(String relPath, String nodeTypeName) throws ItemExistsException,
      PathNotFoundException, NoSuchNodeTypeException, ConstraintViolationException,
      RepositoryException, VersionException, LockException {

    // Parent can be not the same as this node
    JCRPath itemPath = locationFactory.createJCRPath(getLocation(), relPath);

    // Check if there no final index
    if (itemPath.isIndexSetExplicitly())
      throw new RepositoryException(
          "The relPath provided must not have an index on its final element. "
              + itemPath.getAsString(false));

    NodeImpl parent = this.findParent(itemPath);

    // Check if parent exists
    if (parent == null)
      throw new PathNotFoundException("Parent not found for " + itemPath.getAsString(true));

    return doAddNode(parent, itemPath.getName(), sysLocFactory.parseJCRName(nodeTypeName), null,
        false, true);
  }

  /**
   * Creates child node without adding to session's transient storage
   * 
   * @param name
   * @param primaryTypeName
   * @param addAutoCreated if add auto-created items (false value is useful for
   *          node import/copy/move)
   * @return new child node
   * @throws ItemExistsException
   * @throws RepositoryException
   * @throws ConstraintViolationException
   */
  public NodeImpl createChildNode(String name, String primaryTypeName, boolean addAutoCreated)
      throws ItemExistsException, RepositoryException, ConstraintViolationException {
    return doAddNode(this, locationFactory.parseJCRName(name), sysLocFactory
        .parseJCRName(primaryTypeName), null, true, addAutoCreated);
  }

  public NodeImpl createChildNode(String name, String primaryTypeName, String[] mixinTypeNames,
      String uuid, boolean autoCreated, boolean addAutoCreated) throws ItemExistsException,
      RepositoryException, ConstraintViolationException {

    // log.debug(">>> clone mixin "+mixinTypeNames);
    NodeImpl node = doAddNode(this, locationFactory.parseJCRName(name), sysLocFactory
        .parseJCRName(primaryTypeName), uuid, autoCreated, addAutoCreated);

    for (int i = 0; i < mixinTypeNames.length; i++) {
      // node.doAddMixin(mixinTypeNames[i], autoCreated, addAutoCreated);
      // [PN] 31.03.06 node.doAddMixin() instead doAddMixin()  
      node.doAddMixin(mixinTypeNames[i], true, !autoCreated, addAutoCreated);
    }
    return node;
  }

  public NodeImpl createChildNode(String name, String primaryTypeName, boolean autoCreated,
      boolean addAutoCreated) throws ItemExistsException, RepositoryException,
      ConstraintViolationException {

    return doAddNode(this, locationFactory.parseJCRName(name), sysLocFactory
        .parseJCRName(primaryTypeName), null, autoCreated, addAutoCreated);
  }

  // with mixin
  public NodeImpl createChildNode(String name, String primaryTypeName, String[] mixinTypeNames,
      boolean addAutoCreated) throws ItemExistsException, RepositoryException,
      ConstraintViolationException {

    NodeImpl node = doAddNode(this, locationFactory.parseJCRName(name), sysLocFactory
        .parseJCRName(primaryTypeName), null, true, addAutoCreated);
    
//  [PN] 10.02.06. Proposed but throws errors in TCK API WorkspaceCloneReferenceableTest.txt    
//  InternalQName nameQName = locationFactory.parseJCRName(name).getInternalName();
//  InternalQName ptQName = sysLocFactory.parseJCRName(primaryTypeName).getInternalName();
//  NodeImpl node = doAddNode(this, nameQName, ptQName, 
//      null, // mixinTypeNames, will be added below 
//      null, // uuid 
//      -1,   // -1 if calculate
//      false, // registerInDataManager 
//      false, // doExternalValidation 
//      addAutoCreated);
    
    for (int i = 0; i < mixinTypeNames.length; i++) {
      // node.doAddMixin(mixinTypeNames[i], true, false);
      node.doAddMixin(mixinTypeNames[i], false, false, false);
    }

    return node;
  }

  public NodeImpl createChildNodeInmemory(InternalQName name, InternalQName primaryTypeName,
      InternalQName[] mixinTypeNames, String uuid, int index, boolean registerInDataManager,
      boolean doExternalValidation, boolean addAutoCreated) throws ItemExistsException,
      RepositoryException, ConstraintViolationException, VersionException, LockException {

    return doAddNode(this, name, primaryTypeName, mixinTypeNames, uuid, index,
        registerInDataManager, doExternalValidation, addAutoCreated);

  }

  public NodeImpl copyChildNodeInmemory(InternalQName qName, NodeData srcNode, boolean keepUUID,
      boolean keepIndex, boolean registerInDataManager, boolean doExternalValidation,
      boolean addAutoCreated) throws ItemExistsException, RepositoryException,
      ConstraintViolationException, VersionException, LockException {

    InternalQName ptQName = srcNode.getPrimaryTypeName();
    InternalQName[] mtQNames = srcNode.getMixinTypeNames();
    int index = (keepIndex) ? srcNode.getQPath().getIndex() : -1;
    String uuid = (keepUUID) ? srcNode.getUUID() : null;

    NodeImpl nodeImpl = doAddNode(this, qName, ptQName, mtQNames, uuid, index,
        registerInDataManager, doExternalValidation, addAutoCreated);

    return nodeImpl;

  }

  /**
   *  Checks for location changes if the node was persisted before.
   *  And idea in next. We look for this node in storage and if found it
   *  we reading it path (location) only (as we can have unsaved changes here). 
   * */
  protected void fixRefLocation() throws RepositoryException {
    // [PN] 06.04.06
    if (!isNew() && isNodeType(Constants.MIX_REFERENCEABLE)) {
      ItemData storedData = session.getTransientNodesManager().getWorkspaceDataManager().getItemData(getInternalUUID());
      if (storedData != null
          && storedData.getPersistedVersion() >= getPersistedVersion() 
          && !storedData.getQPath().equals(getLocation().getInternalPath())) {
        // We have new location in storage with more fresh version
        String prevPath = getPath();
        loadLocation(locationFactory.createJCRPath(storedData.getQPath()));
        System.out.println("====== reload location from " + prevPath + " to " + getPath());
      }
    }
  }
  
  /**
   * Main method for node adding
   * 
   * @param parentNode
   * @param name
   * @param primaryTypeName
   * @param mixinTypeNames (null or empty if undefined)
   * @param uuid (null if undefined)
   * @param index (-1 if calculate)
   * @param registerInDataManager if register this node with transient session
   *          including storage for changed items and session cache storage (as
   *          a rule: true for Session's method, false for Workspace's and for
   *          auto created)
   * @param doExternalValidation if do validation for externally added nodes (as
   *          a rule: true for manually added node (addNode(..)), false for auto
   *          created/imported etc)
   * @param addAutoCreated if auto-created items will be added as well (as a
   *          rule: true for just created node, false for copied/moved/imported
   *          etc)
   * @return new node
   * @throws ItemExistsException
   * @throws RepositoryException
   */
  private NodeImpl doAddNode(NodeImpl parentNode, InternalQName name,
      InternalQName primaryTypeName, InternalQName[] mixinTypeNames, String uuid, int index,
      boolean registerInDataManager, boolean doExternalValidation, boolean addAutoCreated)
      throws ItemExistsException, RepositoryException, ConstraintViolationException,
      VersionException, LockException {

    // Check if nodeType exists and not mixin
    String ptStr = sysLocFactory.createJCRName(primaryTypeName).getAsString();
    //NodeTypeImpl type = (NodeTypeImpl) nodeType(ptStr);
    NodeTypeImpl type = nodeType(primaryTypeName);

    if (type.isMixin())
      throw new ConstraintViolationException("Add Node failed: Node Type <" + ptStr
          + "> is MIXIN type!");

    // Set mixin types to empty array if null
    if (mixinTypeNames == null)
      mixinTypeNames = new InternalQName[0];

    // [PN] 06.04.06 Check for referenceable node location change
    fixRefLocation();
    
    // Create new NodeData inmemory
    String nameStr = sysLocFactory.createJCRName(name).getAsString();
    JCRPath path = locationFactory.createJCRPath(parentNode.getLocation(), nameStr);

    // set index if any -- will be set later else
    if (index > 0) 
      path.setIndex(index);
    
    // Generate UUID if not found
    if (uuid == null)
      uuid = uuidGenerator.generateUUID(path.getAsString(true)).getString();

    // Find Order number
    int orderNum = 0;
    try {
      // [PN] 04.05.06
      // orderNum = (int) getNodes().getSize();
      orderNum = (int) parentNode.getNodes().getSize();
    } catch (AccessDeniedException e) {
      log.warn("Access denied for read order num == 0 " + getPath());
    }

    // create new nodedata
    NodeData nodeData = new TransientNodeData(path.getInternalPath(), uuid, -1, primaryTypeName,
        mixinTypeNames, orderNum, parentNode.storedNodeData().getUUID(), parentNode.storedNodeData().getACL());

    // Create new Node
    int state = ItemChangeState.ADDED;
    NodeImpl node = new NodeImpl(nodeData, parentNode, state, session);

    // Check if new node's node type is allowed by its parent definition
    NodeType[] types = parentNode.getAllNodeTypes();
    for (int i = 0; i < types.length; i++) {
      NodeTypeImpl t = (NodeTypeImpl) types[i];
      if (t.isChildNodePrimaryTypeAllowed(locationFactory.createJCRName(primaryTypeName)
          .getAsString())) {
        break;
      } else if (i == types.length - 1) {
        throw new ConstraintViolationException("Can't add node " + name.getAsString() + " to "
            + parentNode.getPath() + " node type " + primaryTypeName.getAsString()
            + " is not allowed as child's node type for parent node type ");
      }
    }

    // Validation
    if (doExternalValidation) {

      // Check if node is not protected
      if (node.getDefinition().isProtected()) // nodeTypeValidator.isItemProtected(this))
        throw new ConstraintViolationException("Can't add protected node " + name.getAsString()
            + " to " + parentNode.getPath());

      // Check if versionable ancestor is not checked-in
      if (!parentNode.isCheckedOut())
        throw new VersionException("Node " + parentNode.getPath()
            + " or its nearest ancestor is checked-in");

      // Check locking
      if (!parentNode.checkLocking())
        throw new LockException("Node " + parentNode.getPath() + " is locked ");

    }

    // if samename exists check samename siblings and calculate index
    if (index == -1) { 
      node.getLocation().setIndex(calculateIndex(parentNode.getNodes(), node));
    } else if (index > 1) {
      // [PN] 01.06.06 fixed for ItemExistsException throw; was detected in TCK's SerializationTest 
      if (!node.getDefinition().allowsSameNameSiblings()) {
        throw new ItemExistsException("A node with this name (" 
            + name.getAsString() + ") is already exists. "
            + "Overwriting an existing node during import is impossible.");
      }
    }

    // add node to parent
    parentNode.addChildNode(node);

    // Register node with dataManager
    if (registerInDataManager)
      dataManager.update(node);

    if (addAutoCreated) {
      // Add autocreated items
      node.addAutoCreatedItems(node.getPrimaryNodeType(), false);
    }

    if (log.isDebugEnabled())
      log.debug("DO ADD NODE >>>> new node : " + node.getPath() + " name: "
          + " primaryType: " + node.getPrimaryNodeType().getName() + " index: " + node.getIndex()
          + " state: " + node.getState() + " parent: " + parentNode);

    if (registerInDataManager) {
      return (NodeImpl) dataManager.cacheItem(node, true);
    } else
      return node;
  }

  /**
   * @deprecated
   */
  private NodeImpl doAddNode(NodeImpl parentNode, JCRName name, JCRName primaryTypeName,
      String uuid, boolean autoCreated, boolean addAutoCreated) throws ItemExistsException,
      RepositoryException, ConstraintViolationException, VersionException, LockException {

    //long startDoAddNode = System.currentTimeMillis();
    
    // Check if nodeType exists and not mixin
    //NodeTypeImpl type = (NodeTypeImpl) nodeType(primaryTypeName.getAsString());
    NodeTypeImpl type = nodeType(primaryTypeName.getInternalName());
    if (type.isMixin())
      throw new ConstraintViolationException("Add Node failed: Node Type <" + primaryTypeName
          + "> is MIXIN type!");

    // [PN] 06.04.06 Check for referenceable node location change
    fixRefLocation();
    
    // Create new NodeData inmemory
    JCRPath path = locationFactory.createJCRPath(parentNode.getLocation(), name.getAsString());
    if (uuid == null)
      uuid = uuidGenerator.generateUUID(path.getAsString(true)).getString();
    int orderNum = 0;
    //long startOrderNumb = System.currentTimeMillis();
    try {
      // [PN] 04.05.06
      // orderNum = (int) getNodes().getSize();
      orderNum = (int) parentNode.getNodes().getSize();
    } catch (AccessDeniedException e) {
      log.warn("Access denied for read order num == 0 " + getPath());
    }
    //if (log.isDebugEnabled())
    //  log.info("      orderNum:\t\t\t" + (System.currentTimeMillis() - startOrderNumb) + "ms, " + (System.currentTimeMillis() - startDoAddNode) + "ms");

    NodeData nodeData = new TransientNodeData(path.getInternalPath(), uuid, -1, primaryTypeName
        .getInternalName(), new InternalQName[0], orderNum, parentNode.storedNodeData().getUUID(), parentNode.storedNodeData().getACL());

    // Create new Node
    int state = ItemChangeState.ADDED;
    NodeImpl node = new NodeImpl(nodeData, parentNode, state, session);

    // Check if new node's node type is allowed by its parent definition
    NodeType[] types = parentNode.getAllNodeTypes();
    for (int i = 0; i < types.length; i++) {
      NodeTypeImpl t = (NodeTypeImpl) types[i];
      if (t.isChildNodePrimaryTypeAllowed(primaryTypeName.getAsString())) {
        break;
      } else if (i == types.length - 1) {
        throw new ConstraintViolationException("Can't add node " + name.getAsString() + " to "
            + parentNode.getPath() + " node type " + primaryTypeName.getAsString()
            + " is not allowed as child's node type for parent node type " 
            + parentNode.getPrimaryNodeType().getName());
      }
    }

    // For manually added
    if (!autoCreated) {

      //long startNotAutoCreated = System.currentTimeMillis();
      
      // Check if node is not protected
      if (node.getDefinition().isProtected()) // nodeTypeValidator.isItemProtected(this))
        throw new ConstraintViolationException("Can't add protected node " + name.getAsString()
            + " to " + parentNode.getPath());

      // Check if versionable ancestor is not checked-in
      if (!parentNode.isCheckedOut())
        throw new VersionException("Node " + parentNode.getPath()
            + " or its nearest ancestor is checked-in");

      // Check locking
      if (!parentNode.checkLocking())
        throw new LockException("Node " + parentNode.getPath() + " is locked ");
      
      //if (log.isDebugEnabled())
      //  log.info("      notAutoCreated:\t\t" + (System.currentTimeMillis() - startNotAutoCreated) + "ms, " + (System.currentTimeMillis() - startDoAddNode) + "ms");
    }

    // if samename exists check samename siblings and calculate index
    //long startCalculateIndex = System.currentTimeMillis();
    node.getLocation().setIndex(calculateIndex(parentNode.getNodes(), node));
    //log.info("      calculateIndex:\t\t" + (System.currentTimeMillis() - startCalculateIndex) + "ms, " + (System.currentTimeMillis() - startDoAddNode) + "ms");

    if (log.isDebugEnabled())
      log.debug("DO ADD NODE >>>> new node : " + node.getPath() + " name: " + " primaryType: "
          + node.getPrimaryNodeType().getName() + " index: " + node.getIndex() + " state: "
          + node.getState() + " parent: " + parentNode);

    // add node to parent
    //long startAddChild = System.currentTimeMillis();
    parentNode.addChildNode(node);
    //if (log.isDebugEnabled())
    //  log.info("      addChildNode:\t\t" + (System.currentTimeMillis() - startAddChild) + "ms, " + (System.currentTimeMillis() - startDoAddNode) + "ms");

    // Register node with dataManager
    if (!autoCreated) {
      //long startDMUpdate = System.currentTimeMillis();
      dataManager.update(node);
      //if (log.isDebugEnabled())
      //  log.info("      dataManager.update:\t\t" + (System.currentTimeMillis() - startDMUpdate) + "ms, " + (System.currentTimeMillis() - startDoAddNode) + "ms");
    }

    if (addAutoCreated) {
      // Add autocreated items
      //long startInitAuto = System.currentTimeMillis();
      node.addAutoCreatedItems(node.getPrimaryNodeType(), false);
      //if (log.isDebugEnabled())
      //  log.info("      addAutoCreatedItems:\t\t" + (System.currentTimeMillis() - startInitAuto) + "ms, " + (System.currentTimeMillis() - startDoAddNode) + "ms");
    }

    // return node;
    //long startCacheItem = System.currentTimeMillis();
    NodeImpl thisNode = (NodeImpl) dataManager.cacheItem(node, true);
    //if (log.isDebugEnabled())
    //  log.info("      dataManager.cacheItem:\t" + (System.currentTimeMillis() - startCacheItem) + "ms, " + (System.currentTimeMillis() - startDoAddNode) + "ms");
    //if (log.isDebugEnabled())
    //  log.info("   doAddNode: " + (System.currentTimeMillis() - startDoAddNode) + "ms");
    return thisNode;
  }

  private int calculateIndex(NodeIterator children, NodeImpl childToAdd)
      throws RepositoryException, ItemExistsException {
    NodeDefinition def = childToAdd.getDefinition();
    int ind = 0;
    while (children.hasNext()) {
      NodeImpl child = (NodeImpl) children.nextNode();
      if (child.getState() != ItemChangeState.DELETED
          && child.getName().equals(childToAdd.getName())) {
        if (log.isDebugEnabled()) 
          log.debug(">>> Calculate index for " + childToAdd.getPath() + " " + child.getPath() + " " + child.getState());

        if (def.allowsSameNameSiblings())
          ind++;
        else
          throw new ItemExistsException("The node " + childToAdd.getName() + " already exists in "
              + getPath() + " and same name sibling is not allowed ");

      }
    }
    return ind + 1;

  }

  /**
   * Initializes children if neccessary and Adds child node
   * 
   * @param node to add
   * @throws RepositoryException
   */
  public void addChildNode(NodeImpl node) throws RepositoryException {
    loadChildNodes();
    int removeIndex = -1;
    for (int i = 0; i < childNodes.size() && removeIndex < 0; i++) {
      if (((Node) childNodes.get(i)).getPath().equals(node.getPath())) {
        // Here is a possible error:
        // when we have a node with state DELETED and will try create a new node
        // with this path
        removeIndex = i;
      }
    }
    // [PN] 20.07.06
    if (removeIndex>=0) {
      // [PN] 04.05.06 caused by restore procedure behaviour, and it's more true
      NodeImpl oldNode = (NodeImpl) childNodes.remove(removeIndex);
      node.setOrderNumber(removeIndex);
      childNodes.add(removeIndex, node);
      // [PN] TODO May be it's not true place
      oldNode.setState(ItemChangeState.DELETED);
    } else {
      childNodes.add(node);
    }
    if (log.isDebugEnabled())
      log.debug("Add child node " + (node != null ? node.getPath() : "[null]") 
          + ", nodes count: " + childNodes.size());
  }
  
  /**
   * @param node
   * @throws RepositoryException
   */
  NodeImpl removeChildNode(ItemImpl node) throws RepositoryException {
    loadChildNodes();
    NodeImpl removedNode = null;
    int removeIndex = -1;
    for (int i = 0; i < childNodes.size() && removeIndex < 0; i++) {
      NodeImpl nextNode = (NodeImpl) childNodes.get(i);
      if (nextNode.getInternalUUID().equals(node.getInternalUUID())) {
        removeIndex = i;
      }
    }
    if (removeIndex >= 0) {
      removedNode = (NodeImpl) childNodes.remove(removeIndex);
      if (log.isDebugEnabled())
        log.debug("Remove child node " + (removedNode != null ? removedNode.getPath() 
            + " " + removedNode.getOrderNumber() : "[null]") 
            + ", nodes count:" + childNodes.size());
    }
    return removedNode;
  }

  void reorderChildNodes(boolean updateInStorage) throws RepositoryException {
    for (int i = 0; i < childNodes.size(); i++) {
      NodeImpl children = (NodeImpl) childNodes.get(i);
      if (log.isDebugEnabled())
        log.debug("Reorder child node " + (children != null ? children.getPath() 
            + " " + children.getOrderNumber() + "-->" + i : "[null]"));
      children.setOrderNumber(i);
      // [PN] 21.07.06
      if (!children.isNew()) {
        children.setState(ItemChangeState.UPDATED);
        if (updateInStorage) {
          //children.markUpdated(); // [PN] 14.04.06 TODO remove usage of markUpdated() call
          dataManager.getTransactManager().saveItem(children);
          //dataManager.resetItemState(children);
          children.setState(ItemChangeState.UNCHANGED);
        } else {
          //children.setState(ItemChangeState.UPDATED);
          dataManager.update(children);
        }
      }
    }
  }
  
  /**
   * @deprecated [PN] 06.05.06
   * 
   * Reindex child nodes with same name sibs. Called from
   * SessionDataManager.commit() for DELETED items.
   */
  void reindexChildNodes(NodeImpl removedNode) throws RepositoryException {

    //log.debug("===== reindexChildNodes(): removed SNS'Node: " + removedNode.getPath());
    for (NodeIterator childs = getNodes(removedNode.getName()); childs.hasNext();) {
      NodeImpl children = (NodeImpl) childs.next();
      if (children.getName().equals(removedNode.getName())
          && children.getIndex() > removedNode.getIndex()) {
        //log.debug("=== sameName child for pack " + children.getPath() + " -> " + children.getIndex());
        children.getLocation().setIndex(children.getIndex() - 1);
        //log.debug("=== sameName child renewed index " + children.getPath() + " -> " + children.getIndex());
      }
    }
  }
  
  /**
   * Reindex child nodes with same name sibs. Called from
   * SessionDataManager.commit() for DELETED items.
   * 
   * @param nodeName
   * @throws RepositoryException
   */
  void reindexChildNodesByName(String nodeName) throws RepositoryException {
    
    int i = 1;
    for (NodeIterator childs = getNodes(nodeName); childs.hasNext();) {
      NodeImpl child = (NodeImpl) childs.next();
      //int prevIndex = child.getIndex();
      child.getLocation().setIndex(i++);
      //log.debug("reindexChildNodesByName() " + child.getPath() + " " + prevIndex + " -> " + child.getIndex() + "  " + child.getInternalUUID());
    }
  }
  
  protected int getValuesType(Value[] values) throws ValueFormatException {

    int type = PropertyType.UNDEFINED;

    if (values != null && values.length > 0) {
      for (Value v : values) {
        if (v != null) {
          if (type != PropertyType.UNDEFINED) {
            if (v.getType() != type) {
              throw new ValueFormatException(
                  "Can not set multi-valued property with array of different type values  "
                      + getPath());
            }
          } else {
            type = v.getType();
          }
        }
      }
    }

    return type;
  }

  /**
   * @see javax.jcr.Node#setProperty
   */
  public Property setProperty(String name, Value[] values, int type) throws ValueFormatException,
      VersionException, LockException, ConstraintViolationException, RepositoryException {

    getValuesType(values);

    return doSetProperty(name, values, type, true, true, true);
  }

  /**
   * @see javax.jcr.Node#setProperty
   */
  public Property setProperty(String name, Value[] values) throws ValueFormatException,
      VersionException, LockException, ConstraintViolationException, RepositoryException {

    // if (values != null && values.length > 0 && values[0] != null)
    // type = values[0].getType();

    int type = getValuesType(values);

    return doSetProperty(name, values, type, true, true, true);
  }

  /**
   * @see javax.jcr.Node#setProperty
   */
  public Property setProperty(String name, String[] values) throws ValueFormatException,
      VersionException, LockException, ConstraintViolationException, RepositoryException {

    Value[] strValues;
    if (values == null)
      strValues = null;
    else {
      List valList = packStrValues(values);
      strValues = new Value[valList.size()];
      for (int i = 0; i < valList.size(); i++) {
        strValues[i] = valueFactory.createValue((String) valList.get(i));
      }
    }
    return doSetProperty(name, strValues, PropertyType.STRING, true, true, true);
  }

  /**
   * @see javax.jcr.Node#setProperty
   */
  public Property setProperty(String name, String[] values, int type) throws ValueFormatException,
      VersionException, LockException, ConstraintViolationException, RepositoryException {
    Value[] strValues;
    if (values == null)
      strValues = null;
    else {
      List valList = packStrValues(values);
      strValues = new Value[valList.size()];
      for (int i = 0; i < valList.size(); i++) {
        strValues[i] = valueFactory.createValue((String) valList.get(i));
      }
    }

    return doSetProperty(name, strValues, type, true, true, true);
  }

  private List packStrValues(String values[]) {
    ArrayList valList = new ArrayList();
    for (int i = 0; i < values.length; i++) {
      if (values[i] != null)
        valList.add(values[i]);
    }
    return valList;
  }

  /**
   * @see javax.jcr.Node#setProperty
   */
  public Property setProperty(String name, Value value) throws ValueFormatException,
      VersionException, LockException, ConstraintViolationException, RepositoryException {
    int valType;
    Value[] values;
    if (value == null) {
      valType = PropertyType.UNDEFINED;
      values = null;
    } else {
      valType = value.getType();
      values = new Value[] { value };
    }

    return doSetProperty(name, values, valType, false, true, true);
  }

  /**
   * @see javax.jcr.Node#setProperty
   */
  public Property setProperty(String name, Value value, int type) throws ValueFormatException,
      VersionException, LockException, ConstraintViolationException, RepositoryException {

    Value[] newValues = value != null ? new Value[] { value } : null;
    // if (value != null) {
    // newValues = valueFactory.convert(new Value[] { value }, type, false);
    // }
    return doSetProperty(name, newValues, type, false, true, true);
  }

  /**
   * @see javax.jcr.Node#setProperty
   */
  public Property setProperty(String name, String value, int type) throws ValueFormatException,
      VersionException, LockException, ConstraintViolationException, RepositoryException {

    Value val = null;
    // Hm, It's copy of code used in PropertyImpl.doSetValue(),
    // for TCK's SetPropertyAssumeTypeTest
    try {
      val = value == null ? null : valueFactory.createValue(value, type);
    } catch (ValueFormatException e) {
      // We have to know what type of Exception tyo throw,
      // ValueFormatException or ConstraintViolationException for property type
      // violated the node type definition

      throw new ConstraintViolationException(
          "The change would violate a requred node type with property '" + name
              + "'. A property value '" + value + "' can't be of type "
              + PropertyType.nameFromValue(type) + ". Converting is impossible: " + e.getMessage(),
          e);
    }

    return setProperty(name, val, type);

  }

  /**
   * @see javax.jcr.Node#setProperty
   */
  public Property setProperty(String name, String value) throws ValueFormatException,
      VersionException, LockException, ConstraintViolationException, RepositoryException {

    Value val = value == null ? null : valueFactory.createValue(value);
    return setProperty(name, val, PropertyType.STRING);
  }

  /**
   * @see javax.jcr.Node#setProperty
   */
  public Property setProperty(String name, InputStream value) throws ValueFormatException,
      VersionException, LockException, ConstraintViolationException, RepositoryException {

    Value val = value == null ? null : valueFactory.createValue(value);
    return setProperty(name, val, PropertyType.BINARY);
  }

  /**
   * @see javax.jcr.Node#setProperty
   */
  public Property setProperty(String name, boolean value) throws ValueFormatException,
      VersionException, LockException, ConstraintViolationException, RepositoryException {

    return setProperty(name, valueFactory.createValue(value), PropertyType.BOOLEAN);
  }

  /**
   * @see javax.jcr.Node#setProperty
   */
  public Property setProperty(String name, Calendar value) throws ValueFormatException,
      VersionException, LockException, ConstraintViolationException, RepositoryException {

    Value val = value == null ? null : valueFactory.createValue(value);
    return setProperty(name, val, PropertyType.DATE);
  }

  /**
   * @see javax.jcr.Node#setProperty
   */
  public Property setProperty(String name, double value) throws ValueFormatException,
      VersionException, LockException, ConstraintViolationException, RepositoryException {

    return setProperty(name, valueFactory.createValue(value), PropertyType.DOUBLE);
  }

  /**
   * @see javax.jcr.Node#setProperty
   */
  public Property setProperty(String name, long value) throws ValueFormatException,
      VersionException, LockException, ConstraintViolationException, RepositoryException {

    return setProperty(name, valueFactory.createValue(value), PropertyType.LONG);

  }

  /**
   * @see javax.jcr.Node#setProperty
   */
  public Property setProperty(String name, Node value) throws ValueFormatException,
      VersionException, LockException, ConstraintViolationException, RepositoryException {

    Value val = (value == null) ? null : valueFactory.createValue(value);
    return setProperty(name, val, PropertyType.REFERENCE);
  }

  /**
   * @param qname
   * @param values
   * @param type
   * @param multiValued
   * @param registerInDataManager if register this prop with transient session
   *          including storage for changed items and session cache storage (as
   *          a rule: true for Session's method, false for Workspace's and for
   *          auto created)
   * @param doExternalValidation if do validation for externally added props (as
   *          a rule: true for manually added node (setProperty(..)), false for
   *          auto created/imported etc)
   * @return
   */
  protected PropertyImpl doSetProperty(InternalQName qname, Value[] values, int type,
      boolean multiValued, boolean registerInDataManager, boolean doExternalValidation)
      throws ValueFormatException, VersionException, LockException, ConstraintViolationException,
      RepositoryException {

    // [PN] 06.04.06 Check for referenceable node location change
    fixRefLocation();
    
    String name = this.locationFactory.createJCRName(qname).getAsString();
    JCRPath path = locationFactory.createJCRPath(getLocation(), name);
    PropertyImpl property;
    int propState;
    try {
      property = (PropertyImpl) getProperty(name);
      propState = ItemChangeState.UPDATED;

    } catch (PathNotFoundException e) {
      // create new property
      // JCRPath path = locationFactory.createJCRPath(getLocation(), name);
      String uuid = uuidGenerator.generateUUID(path.getAsString(true)).getString();
      TransientPropertyData propertyData = new TransientPropertyData(path.getInternalPath(), uuid, -1, type,
          getInternalUUID(), multiValued);
      // data.setMultiValued(multiValued);
      propState = ItemChangeState.ADDED;
      property = new PropertyImpl(propertyData, this, propState, session);      
    }

    // System.out.println("SET VALUES >>>>>>> "+property.getType());

    // set new values
    property.doSetValue(values, multiValued, type, propState, registerInDataManager,
        doExternalValidation);

    // [PN] 20.07.06
//    if (property.getState() != ItemChangeState.DELETED)
//      addChildProperty(property);

    if (registerInDataManager)
      return (PropertyImpl) dataManager.cacheItem(property, true);
    else
      return property;

  }
  
  /**
   * Create a child node in memory using InternalQName name, value, multiple flag and property type. 
   * */
  public PropertyImpl createChildPropertyInmemory(InternalQName qname, int type, Value[] values,  
      boolean multiValued, boolean registerInDataManager, boolean doExternalValidation)
      throws ValueFormatException, VersionException, LockException, ConstraintViolationException,
      RepositoryException {

    return doSetProperty(qname, values, type, multiValued, registerInDataManager,
        doExternalValidation);
  }   

  /**
   * Create a child node in memory using InternalQName, value and property type.
   * Multiple flag will be used from property definition. 
   * If no property definition found then multiple-valued value will be created. 
   * */
  public PropertyImpl createChildPropertyInmemory(InternalQName qname, int type, Value[] values, 
      boolean registerInDataManager, boolean doExternalValidation)
      throws ValueFormatException, VersionException, LockException, ConstraintViolationException, RepositoryException {
      
    JCRName propertyName = session.getLocationFactory().createJCRName(qname);
    PropertyDefinition propertyDefinition = findPropertyDefinition(propertyName.getAsString());
    boolean multiValued = propertyDefinition != null ? propertyDefinition.isMultiple() : true; 
    return createChildPropertyInmemory(qname, type, values, multiValued, 
          registerInDataManager, doExternalValidation);
  }
  
  /**
   * Create a child node in memory using string name, value, multiple flag and property type 
   * */
  public PropertyImpl createChildPropertyInmemory(String name, int type, Value[] values, 
    boolean multiValued, boolean registerInDataManager, boolean doExternalValidation)
    throws ValueFormatException, VersionException, LockException, ConstraintViolationException, RepositoryException {
    
    InternalQName qname = session.getLocationFactory().parseJCRName(name).getInternalName();
    return createChildPropertyInmemory(qname, type, values, multiValued, 
        registerInDataManager, doExternalValidation);
  }

  /**
   * Create a child node in memory using string name, value and property type.
   * Multiple flag will be used from property definition. 
   * If no property definition found then multiple-valued value will be created. 
   * */
  public PropertyImpl createChildPropertyInmemory(String name, int type, Value[] values, 
      boolean registerInDataManager, boolean doExternalValidation)
      throws ValueFormatException, VersionException, LockException, ConstraintViolationException, RepositoryException {
      
    PropertyDefinition propertyDefinition = findPropertyDefinition(name);
    boolean multiValued = propertyDefinition != null ? propertyDefinition.isMultiple() : true; 
    InternalQName qname = session.getLocationFactory().parseJCRName(name).getInternalName();
    return createChildPropertyInmemory(qname, type, values, multiValued, 
          registerInDataManager, doExternalValidation);
  }
  
  /**  
   * @deprecated 
   * */
  public PropertyImpl createChildProperty(String name, Value[] values, int type)
      throws ValueFormatException, VersionException, LockException, ConstraintViolationException, RepositoryException {

    PropertyDefinition anyDef = findPropertyDefinition(name);
//    if (name.equals("jcr:created")) {
//      log.info("jcr:created >>> " + OnParentVersionAction.nameFromValue(anyDef.getOnParentVersion()));
//    }
    if (anyDef != null) {
      return doSetProperty(name, values, type, anyDef.isMultiple(), false, false);
    } else {
      throw new ConstraintViolationException("Property " + name + " is not defined for "
          + getPath());
    }
  }

  /** 
   *  @deprecated
   * 
   * @param name
   * @param values
   * @param type
   * @param multiValued
   * @param registerInDataManager
   * @param doExternalValidation
   * @return
   * @throws ValueFormatException
   * @throws VersionException
   * @throws LockException
   * @throws ConstraintViolationException
   * @throws RepositoryException
   */
  protected PropertyImpl doSetProperty(String name, Value[] values, int type, boolean multiValued,
      boolean registerInDataManager, boolean doExternalValidation) throws ValueFormatException,
      VersionException, LockException, ConstraintViolationException, RepositoryException {

    // [PN] 06.04.06 Check for referenceable node location change
    fixRefLocation();
    
    PropertyImpl property;
    int propState;

    try {
      property = (PropertyImpl) getProperty(name);
      propState = ItemChangeState.UPDATED;

    } catch (PathNotFoundException e) {
      // create new property
      JCRPath path = locationFactory.createJCRPath(getLocation(), name);
      String uuid = uuidGenerator.generateUUID(path.getAsString(true)).getString();
      TransientPropertyData data = new TransientPropertyData(path.getInternalPath(), uuid, -1, type,
          getInternalUUID(), multiValued);
      // data.setMultiValued(multiValued);
      propState = ItemChangeState.ADDED;
      // state = autoCreated?ItemChangeState.AUTO_ADDED:ItemChangeState.ADDED;
      property = new PropertyImpl(data, this, propState, session);
    }

    // set new values
    property.doSetValue(values, multiValued, type, propState, registerInDataManager, doExternalValidation);

    // [PN] 20.07.06
//  if (property.getState() != ItemChangeState.DELETED)
//    addChildProperty(property);
    
    // [PN] 11.04.06 usage of registerInDataManager flag
    if (registerInDataManager) {
      return (PropertyImpl) dataManager.cacheItem(property, true);
    } else {
      return property;
    }
  }

  void addChildProperty(PropertyImpl property) throws RepositoryException {

    loadChildProperties();
    
    // [PN] 20.07.06
    //    for (int i = 0; i < childProperties.size(); i++) {
    //      if (((Property) childProperties.get(i)).getPath().equals(property.getPath())) {
    //        childProperties.remove(i);
    //      }
    //    }
    // [PN] 20.07.06 By UUID instead Path
    int removeIndex = -1;
    for (int i = 0; i < childProperties.size() && removeIndex < 0; i++) {
      PropertyImpl nextProp = (PropertyImpl) childProperties.get(i);
      if (nextProp.getInternalUUID().equals(property.getInternalUUID())) {
        removeIndex = i;
      }
    }
    if (removeIndex>=0)
      childProperties.remove(removeIndex);
    childProperties.add(property);
    if (log.isDebugEnabled())
      log.debug("Add child property " + (property != null ? property.getPath() : "[null]") 
          + ", properties count: " + childProperties.size());
  }

  PropertyImpl removeChildProperty(ItemImpl property) throws RepositoryException {
    loadChildProperties();    
    PropertyImpl removedProperty = null;
    int removeIndex = -1;
    for (int i = 0; i < childProperties.size() && removeIndex < 0; i++) {
      PropertyImpl nextProp = (PropertyImpl) childProperties.get(i);
      if (nextProp.getInternalUUID().equals(property.getInternalUUID())) {
        removeIndex = i;
      }
    }
    if (removeIndex>=0) {
      removedProperty = (PropertyImpl) childProperties.remove(removeIndex);
      if (log.isDebugEnabled())
        log.debug("Remove child property " + (removedProperty != null ? removedProperty.getPath() : "[null]") 
            + ", properties count: " + childProperties.size());
    }
    return removedProperty;
  }

  /**
   * @see javax.jcr.Node#getPrimaryNodeType
   */
  public NodeType getPrimaryNodeType() throws RepositoryException {
    return nodeType(primaryType.getInternalName());
  }

  /**
   * @see javax.jcr.Node#getMixinNodeTypes
   */
  public NodeType[] getMixinNodeTypes() throws RepositoryException {

    // should not be null
    if (mixinTypes == null)
      throw new RepositoryException("Data Container implementation error getMixinTypeNames == null");
    NodeType[] mixinNodeTypes = new NodeType[mixinTypes.length];
    for (int i = 0; i < mixinNodeTypes.length; i++) {
      mixinNodeTypes[i] = nodeType(mixinTypes[i].getInternalName());
    }

    return mixinNodeTypes;
  }

  /**
   * @see javax.jcr.Node#isNodeType
   */
  public boolean isNodeType(String nodeTypeName) throws RepositoryException {
    return isNodeType(locationFactory.parseJCRName(nodeTypeName).getInternalName());
  }
  
  public boolean isNodeType(InternalQName qName) throws RepositoryException {
    NodeType testNodeType;
    try {
      testNodeType = nodeType(qName);
    } catch (NoSuchNodeTypeException e) {
      //e.printStackTrace();
      log.warn("Node.isNodeType() No such nodetype: " + qName.getAsString());
      return false;
    }
    
    //System.out.println("IS TYPE >>> "+testNodeType+" "+qName.getAsString());
    
    NodeType[] nodetypes = getAllNodeTypes();
    for (int i = 0; i < nodetypes.length; i++) {
      //System.out.println("IS TYPE >>> "+testNodeType.getName()+" "+nodetypes[i]);
      if (NodeTypeImpl.isSameOrSubType(testNodeType, nodetypes[i]))
        return true;
    }
    return false;
  }


  /**
   * @see javax.jcr.Node#getDefinition
   */
  public NodeDefinition getDefinition() throws RepositoryException {

    if (definition == null)
      initDefinition();

    return definition;

  }

  /**
   * @see javax.jcr.Node#getCorrespondingNodePath
   */
  public String getCorrespondingNodePath(String workspaceName) throws ItemNotFoundException,
      NoSuchWorkspaceException, AccessDeniedException, RepositoryException {
    SessionImpl correspSession = (SessionImpl) ((RepositoryImpl) session.getRepository()).login(
        session.getCredentials(), workspaceName);

    return getCorrespondingNode(correspSession).getPath();
  }

  public NodeImpl getCorrespondingNode(SessionImpl correspSession) throws ItemNotFoundException,
      NoSuchWorkspaceException, AccessDeniedException, RepositoryException {
    if (this.isNodeType("mix:referenceable")) {
      try {
        return (NodeImpl) correspSession.getNodeByUUID(getUUID());
      } catch (ItemNotFoundException e) {
      }
    } else {
      //Node refAncestor = null;
      //List qpathEntries = new ArrayList();
      //NodeImpl curNode = this;
      for (int i = getDepth(); i >= 0; i--) {
        NodeImpl ancestor = (NodeImpl) getAncestor(i);
        if (ancestor.isNodeType("mix:referenceable")) {
          NodeImpl correspAncestor = (NodeImpl) correspSession.getNodeByUUID(ancestor.getUUID());
          JCRPath.PathElement[] relJCRPath = getLocation().getRelPath(getDepth() - i);
          // System.out.println("LOC. getRelPath "+(getDepth() - i)+" =
          // "+getRelPath(relJCRPath)+" "+getPath());
          try {
            return (NodeImpl) correspAncestor.getNode(getRelPath(relJCRPath));
          } catch (ItemNotFoundException e) {
          }
        }
      }
    }
    try {
      return (NodeImpl) correspSession.getItem(getPath());
    } catch (PathNotFoundException e) {
      throw new ItemNotFoundException("No corresponding path for " + getPath() + " in "
          + correspSession.getWorkspace().getName());
    }
  }

  private String getRelPath(JCRPath.PathElement[] relPath) throws RepositoryException {
    String path = "";
    for (int i = 0; i < relPath.length; i++) {
      path += relPath[i].getAsString(false);
      if (i < relPath.length - 1)
        path += "/";
    }
    return path;
  }

  /**
   * @see javax.jcr.Node#update
   */
  public void update(String srcWorkspaceName) throws NoSuchWorkspaceException,
      AccessDeniedException, InvalidItemStateException, LockException, RepositoryException {

    // Check pending changes
    if (session.hasPendingChanges())
      throw new InvalidItemStateException("Session has pending changes ");

    // Check locking
    if (!checkLocking())
      throw new LockException("Node " + getPath() + " is locked ");

    String srcPath;
    try {
      srcPath = getCorrespondingNodePath(srcWorkspaceName);
      // [PN] 27.07.06
      //remove();
      //save();
      removeFromParent();
      session.getTransientNodesManager().getTransactManager().saveItem(this);
    } catch (ItemNotFoundException e) {
      if (log.isDebugEnabled()) log.debug("No corresponding node in workspace: " + srcWorkspaceName);
      return;
    }
    session.getWorkspace().clone(srcWorkspaceName, srcPath, this.getPath(), true);
    reloadNodeData();
  }

  protected void reloadNodeData() throws RepositoryException {
    ItemData itemData = dataManager.getTransactManager().getItemData(getLocation().getInternalPath());
    if (itemData != null) { // || getPersistedVersion() > 0
      loadData(itemData);
    } 
  }

  /**
   * @see javax.jcr.Node#addMixin
   */
  public void addMixin(String mixinName) throws NoSuchNodeTypeException,
      ConstraintViolationException, VersionException, LockException, RepositoryException {

    doAddMixin(mixinName, true, true, true);
  }

  public void doAddMixin(String mixinName, boolean registerInDataManager,
      boolean doExternalValidation, boolean addAutoCreated) throws NoSuchNodeTypeException,
      ConstraintViolationException, VersionException, LockException, RepositoryException {

    // [PN] 19.04.06 magic, will be fixed in session refactoring
    // has place after changes in reloadNodeData()
    loadChildProperties();
    setRefreshToPersistedVersion(false);
    
    // Add both to mixinNodeTypes and to jcr:mixinTypes property
    NodeType type = nodeType(locationFactory.parseJCRName(mixinName).getInternalName());
    if (log.isDebugEnabled()) log.debug("Node.addMixin " + mixinName + " " + getPath());
    
    // Mixin or not
    if (type == null || !type.isMixin())
      throw new NoSuchNodeTypeException("Node Type " + mixinName + " not found or not mixin type.");

    if (doExternalValidation) {

      // Validate
      // if (!canAddMixin(mixinName))
      // throw new ConstraintViolationException("Can not add mixin type " +
      // mixinName);
      if (hasSameOrSubtypeMixin((NodeTypeImpl) type))
        throw new ConstraintViolationException("Can not add mixin type " + mixinName);

      if (getDefinition().isProtected())
        throw new ConstraintViolationException("Can not add mixin type. Node is protected "
            + getPath());

      // Check if versionable ancestor is not checked-in
      if (!isCheckedOut())
        throw new VersionException("Node " + getPath() + " or its nearest ancestor is checked-in");

      // Check locking
      if (!checkLocking())
        throw new LockException("Node " + getPath() + " is locked ");

    }
    
    for (int i = 0; i < mixinTypes.length; i++) {
      if (type.equals(nodeType(mixinTypes[i].getInternalName()))) {
        // we already have this mixin type 
        log.warn("Node.addMixin node already has this mixin type " + mixinName + " " + getPath());
        return;
      }
    }

    // Prepare mixin values
    JCRName[] newMixin = new JCRName[mixinTypes.length + 1];
    Value[] values = new Value[mixinTypes.length + 1];

    for (int i = 0; i < mixinTypes.length; i++) {
      newMixin[i] = mixinTypes[i];
      values[i] = valueFactory.createValue(mixinTypes[i]);
    }
    newMixin[mixinTypes.length] = locationFactory.parseJCRName(mixinName);
    values[mixinTypes.length] = valueFactory.createValue(newMixin[mixinTypes.length]);

    // Set mixin property and nodeData
    boolean firstMixin = (mixinTypes.length == 0);
    mixinTypes = newMixin;
    // Should register jcr:mixinTypes and autocreated items if node is not added
    PropertyImpl prop = doSetProperty(Constants.JCR_MIXINTYPES, values, PropertyType.NAME, true,
        registerInDataManager, false);

    log.debug("Node.addMixin Property " + prop.getPath() + " values " + mixinTypes.length + " "
        + addAutoCreated);

    // Add autocreated items for this nodeType
    if (addAutoCreated)
      addAutoCreatedItems(nodeType(locationFactory.parseJCRName(mixinName).getInternalName()), registerInDataManager);

  }

  /**
   * @see javax.jcr.Node#removeMixin
   */
  public void removeMixin(String mixinName) throws NoSuchNodeTypeException,
      ConstraintViolationException, RepositoryException {

    // find mixin
    int ind = -1;
    for (int i = 0; i < mixinTypes.length; i++) {
      if (mixinTypes[i].getAsString().equals(mixinName))
        ind = i;
    }
    // no mixin found
    if (ind < 0)
      throw new NoSuchNodeTypeException("No mixin type found " + mixinName + " for node "
          + getPath());

    // A ConstraintViolationException will be thrown either
    // immediately or on save if the removal of a mixin is not
    // allowed. Implementations are free to enforce any policy
    // they like with regard to mixin removal and may differ on
    // when this validation is done.

    // Check if versionable ancestor is not checked-in
    if (!isCheckedOut())
      throw new VersionException("Node " + getPath() + " or its nearest ancestor is checked-in");

    // Check locking
    if (!checkLocking())
      throw new LockException("Node " + getPath() + " is locked ");

    // Prepare mixin values
    JCRName[] newMixin = new JCRName[mixinTypes.length - 1];
    Value[] values = new Value[mixinTypes.length - 1];
    for (int i = 0; i < newMixin.length; i++) {
      if (i != ind) {
        newMixin[i] = mixinTypes[i];
        // JCRName mixinJCRName = locationFactory.createJCRName(newMixin[i]);
        values[i] = valueFactory.createValue(mixinTypes[i]);
      }
    }

    // Set mixin property and locally
    mixinTypes = newMixin;
    boolean registerInDataManager = (this.state != ItemChangeState.ADDED);
    PropertyImpl prop = doSetProperty("jcr:mixinTypes", values, PropertyType.NAME, true,
        registerInDataManager, false);

  }

  /**
   * @see javax.jcr.Node#canAddMixin
   */
  public boolean canAddMixin(String mixinName) throws RepositoryException {

//    if (hasSameOrSubtypeMixin((NodeTypeImpl) nodeType(mixinName))) {
    if (hasSameOrSubtypeMixin(nodeType(locationFactory.parseJCRName(mixinName).getInternalName()))) {
      return false;
    }

    if (getDefinition().isProtected()) {
      return false;
    }

    if (!isCheckedOut())
      return false;

    if (!checkLocking())
      return false;

    return true;
  }

  private boolean hasSameOrSubtypeMixin(NodeTypeImpl type) throws RepositoryException,
      ValueFormatException {

    for (int i = 0; i < mixinTypes.length; i++) {
      
//      if (NodeTypeImpl.isSameOrSubType(type, nodeType(mixinTypes[i].getAsString()))) {
      if (NodeTypeImpl.isSameOrSubType(type, nodeType(mixinTypes[i].getInternalName()))) {
        return true;
      }
    }
    return false;

  }

  /**
   * @see javax.jcr.Node#orderBefore
   */
  public void orderBefore(String srcName, String destName)
      throws UnsupportedRepositoryOperationException, ConstraintViolationException,
      ItemNotFoundException, RepositoryException {

    if (!getPrimaryNodeType().hasOrderableChildNodes())
      throw new UnsupportedRepositoryOperationException("Node does not support child ordering "
          + getPrimaryNodeType().getName());

    if (srcName.equals(destName))
      return;

    EntityCollection nodes = (EntityCollection) getNodes();
    NodeImpl srcNode = null;
    NodeImpl destNode = null;
    int size = (int) nodes.getSize();
    if (size < 2)
      throw new UnsupportedRepositoryOperationException("Nothing to order Count of child nodes "
          + size);

    // index number of each element equlal of NodeImpl.orderNumber
    List<NodeImpl> list = nodes.getList();
    List<NodeImpl> prevOrderList = new ArrayList<NodeImpl>(list);

    if (destName == null)
      destNode = list.get(size - 1);

    // find dest and src nodes
    while (nodes.hasNext()) {
      NodeImpl node = (NodeImpl) nodes.nextNode();
      // [PN]
//      if (node.getName().equals(srcName))
//        srcNode = node;
//      if (destName != null && node.getName().equals(destName))
//        destNode = node;
      boolean withIndex = srcName.endsWith("]");
      if (node.getLocation().getAsString(withIndex).endsWith(srcName)) {
        srcNode = node;
      }      
      if (destName != null) {
        withIndex = destName.endsWith("]");
        if (node.getLocation().getAsString(withIndex).endsWith(destName)) {
          destNode = node;
        }
      }
    }

    if (srcNode == null)
      throw new ItemNotFoundException("Src Node not found " + srcName);
    if (destNode == null)
      throw new ItemNotFoundException("Dest Node not found " + destName);
    
    // if sameNameSablings and one name nodes ordered swaping lock state
    // it's a play for next reordering 
    if (srcNode.getName().equals(destNode.getName()) && srcNode.getIndex() != destNode.getIndex()
        && srcNode.isLocked() && !destNode.isLocked()) {
      // make locked current node
      Lock lock = srcNode.getLock();
      destNode.lockInSession(lock.isDeep(), lock.isSessionScoped());
      // unclock prevNode
      srcNode.unlockInSession();
    }    

    // System.out.println("ORDER "+srcNode.getName()+"
    // "+srcNode.getOrderNumber()+" "+destNode.getName()+"
    // "+destNode.getOrderNumber());

    // reorder nodes
    list.remove(srcNode.getOrderNumber());            
    list.add(destNode.getOrderNumber(), srcNode);
    
    int srcOrderNumber = -1;
    int destOrderNumber = -1;
    if (destName == null) {
      srcNode.setOrderDirection(Event.NODE_ADDED); // order source node up to head
      srcOrderNumber = destNode.getOrderNumber() + 1;
      destOrderNumber = srcNode.getOrderNumber() + 1;
    } else if (srcNode.getName().equals(destNode.getName())) {
      srcNode.setOrderDirection(Event.NODE_ADDED); // order source node up to head
      srcOrderNumber = srcNode.getOrderNumber();
      destOrderNumber = destNode.getOrderNumber();
    } else {
      srcNode.setOrderDirection(Event.NODE_ADDED | Event.NODE_REMOVED); // order source node up to head      
    }
    // log.info("==== reorder source: " + srcNode.getPath() + " order numb: " + srcNode.getOrderNumber());
    // log.info("==== reorder dest: " + destNode.getPath() + " order numb: " + destNode.getOrderNumber());
    for (int i = 0; i < list.size(); i++) {
      NodeImpl node = list.get(i);
      // log.info("==== reorder node: " + node.getLocation().getAsString(true) + " from " + node.getOrderNumber() + " to " + i);
      if (srcNode.getName().equals(node.getName())
          && node.getOrderNumber() >= destOrderNumber
          && node.getOrderNumber() < srcOrderNumber) {
        // Same name sibs for source node
        node.setOrderDirection(Event.NODE_REMOVED); // order destination node down to end
        // log.info("==== reorder same name sibs node: " + node.getPath() + " from " + node.getOrderNumber() + " to " + i);
      }
      node.setOrderNumber(i);
      // [PN] 13.04.06, 14.04.06 setState(ItemChangeState.UPDATED) inside  
      // node.setOrderNumber(i) works wrong for events cycle in SDM.commit()
      node.setState(ItemChangeState.UPDATED);
      dataManager.update(node);      
    }
  }
  
  boolean isOrdered() {
    return getOrderDirection() != 0;
  }

  /**
   * @return Returns the orderDirection.
   */
  int getOrderDirection() {
    return orderDirection;
  }

  /**
   * @param orderDirection The orderDirection to set.
   */
  void setOrderDirection(int orderDirection) {
    //  [PN] 08.02.06 Set order status. This is runtime flag.
    this.orderDirection = orderDirection;
  }
  
  int getOrderNumber() {
    return orderNumber;
  }

  void setOrderNumber(int orderNumber) {
    this.orderNumber = orderNumber;
    // [PN] 13.04.06 WRONG
    //setState(ItemChangeState.UPDATED);
  }

  // //////////////////////// OPTIONAL

  // VERSIONING

  public Version checkin() throws VersionException, UnsupportedRepositoryOperationException,
      InvalidItemStateException, RepositoryException {
    
    // validation
    if (!this.isNodeType("mix:versionable"))
      throw new UnsupportedRepositoryOperationException(
          "Node.checkin() is not supported for not mix:versionable node ");
    if (!this.isCheckedOut())
      return this.getBaseVersion();
    if (session.getTransientNodesManager().hasPendingChanges(this.getLocation(), false))
      throw new InvalidItemStateException("Node has pending changes " + getPath());
    if (hasProperty("jcr:mergeFailed"))
      throw new VersionException("Node has jcr:mergeFailed " + getPath());

    // checking process
    //MutableItemDataChangesLog changes = new MutableItemDataChangesLog(session);
    //VersionImpl ver = ((VersionHistoryImpl) getVersionHistory()).addVersion(this.getActualItemData(), changes);
    
    VersionImpl ver = ((VersionHistoryImpl) getVersionHistory()).addVersion(this);
    PropertyImpl isCheckedOut = createChildProperty("jcr:isCheckedOut", new Value[] { valueFactory
        .createValue(false) }, PropertyType.BOOLEAN);
    PropertyImpl baseVersion = createChildProperty("jcr:baseVersion", new Value[] { valueFactory
            .createValue(ver) }, PropertyType.REFERENCE);
    PropertyImpl predecessors = createChildProperty("jcr:predecessors", new Value[0], PropertyType.REFERENCE);

    // [PN] 09.04.06, 12.04.06
    //----------- BEGIN SESSION BATCH ----------
    dataManager.getWorkspaceDataManager().beginSessionBatch(session);
    
    dataManager.getTransactManager().saveItem(ver); // add version
    dataManager.getTransactManager().saveItem(isCheckedOut);
    dataManager.getTransactManager().saveItem(baseVersion);
    dataManager.getTransactManager().saveItem(predecessors);

    dataManager.getWorkspaceDataManager().endSessionBatch(session);
    //----------- END SESSION BATCH ----------
    dataManager.resetItemState(ver);
    dataManager.resetItemState(isCheckedOut);
    dataManager.resetItemState(baseVersion);
    dataManager.resetItemState(predecessors);
    
    dataManager.resetItemState(this); // TODO [PN] ??? does we need it

    //VersionImpl v = new VersionImpl(); 
    //VersionImpl(NodeData data, NodeImpl parent, int state, SessionImpl session)
    return ver;
  }

  public void checkout() throws RepositoryException, UnsupportedRepositoryOperationException {
    if (!this.isNodeType("mix:versionable"))
      throw new UnsupportedRepositoryOperationException(
          "Node.checkout() is not supported for not mix:versionable node ");

    PropertyImpl isCheckedOut = createChildProperty("jcr:isCheckedOut", new Value[] { valueFactory
        .createValue(true) }, PropertyType.BOOLEAN);
    // The current value of N's jcr:baseVersion is copied to N's
    // jcr:predecessors property.    
    PropertyImpl predecessors = createChildProperty("jcr:predecessors", new Value[] { getProperty(
      "jcr:baseVersion").getValue() }, PropertyType.REFERENCE);
    
    // [PN] 12.04.06
    //----------- BEGIN SESSION BATCH ----------
    dataManager.getWorkspaceDataManager().beginSessionBatch(session);
    dataManager.getTransactManager().saveItem(isCheckedOut);
    dataManager.getTransactManager().saveItem(predecessors);
    dataManager.getWorkspaceDataManager().endSessionBatch(session);
    //----------- END SESSION BATCH ----------
    dataManager.resetItemState(isCheckedOut);
    dataManager.resetItemState(predecessors);
    
    //----------- SESSION CACHING ------------
    //dataManager.cacheItem(isCheckedOut, true);
    //dataManager.cacheItem(predecessors, true);
  }

  public boolean isCheckedOut() throws UnsupportedRepositoryOperationException, RepositoryException {

    if (isRoot())
      return true;

    if (this.isNodeType("mix:versionable")) {
      return getProperty("jcr:isCheckedOut").getBoolean();
    }

    NodeImpl ancestor = (NodeImpl) getParent();
    while (!ancestor.isRoot()) {
      if (ancestor.isNodeType("mix:versionable"))
        return ancestor.isCheckedOut();
      else
        ancestor = (NodeImpl) ancestor.getParent();
    }
    if (ancestor.isNodeType("mix:versionable"))
      return ancestor.isCheckedOut();
    else
      return true;
  }

  public VersionHistory getVersionHistory() throws UnsupportedRepositoryOperationException,
      RepositoryException {

    if (!this.isNodeType("mix:versionable"))
      throw new UnsupportedRepositoryOperationException("Node is not mix:versionable " + getPath());

    String versionHistory = getProperty("jcr:versionHistory").getString();

    return (VersionHistory) session.getNodeByUUID(versionHistory);
  }

  public Version getBaseVersion() throws UnsupportedRepositoryOperationException,
      RepositoryException {

    if (!this.isNodeType("mix:versionable"))
      throw new UnsupportedRepositoryOperationException("Node is not versionable " + getPath());

    Version v = (Version) session.getNodeByUUID(getProperty("jcr:baseVersion").getString());

    return v;
  }

  public void restore(Version version, boolean removeExisting) throws VersionException,
      ItemExistsException, UnsupportedRepositoryOperationException, LockException,
      RepositoryException, InvalidItemStateException {

    try {
    
    if (!this.isNodeType("mix:versionable"))
      throw new UnsupportedRepositoryOperationException("Node is not versionable " + getPath());

    if (session.hasPendingChanges())
      throw new InvalidItemStateException("Session has pending changes ");

    if (version.getName().equals("jcr:rootVersion"))
      throw new VersionException("It is illegal to call restore() on jcr:rootVersion");

    if (!((VersionHistoryImpl) getVersionHistory()).isVersionBelongToThis(version))
      throw new VersionException("Bad version " + version.getPath());

    // Check locking
    if (!checkLocking())
      throw new LockException("Node " + getPath() + " is locked ");

    NodeImpl restored = ((VersionImpl) version).restore(this, removeExisting);

    // System.out.println("RESTORED >>>> "+restored+"
    // "+restored.getProperty("jcr:isCheckedOut").getBoolean());

    // System.out.println(">>>>>>>>>>>>>>> " + dataManager.dump());

    // reload data
    loadData(restored.getActualItemData());
    
    } catch(UnsupportedRepositoryOperationException e) {
      //log.error("RESTORE: " + e.getMessage(), e);
      throw e;
    } catch(VersionException e) {
      //log.error("RESTORE: " + e.getMessage(), e);
      throw e;  
    } catch(InvalidItemStateException e) {
      //log.error("RESTORE: " + e.getMessage(), e);
      throw e;  
    } catch(RepositoryException e) {
      //log.error("RESTORE: " + e.getMessage(), e);
      throw e;
    }
  }

  public void restore(String versionName, boolean removeExisting) throws VersionException,
      ItemExistsException, UnsupportedRepositoryOperationException, LockException,
      RepositoryException, InvalidItemStateException {

    // throw VersionException if not found
    VersionImpl version = (VersionImpl) getVersionHistory().getVersion(versionName);
    restore(version, removeExisting);

  }

  public void restore(Version version, String relPath, boolean removeExisting)
      throws VersionException, ItemExistsException, UnsupportedRepositoryOperationException,
      LockException, RepositoryException, InvalidItemStateException {

    NodeImpl node;
    try {
      node = (NodeImpl) getNode(relPath);
    } catch (PathNotFoundException e) {
      JCRPath newPath = locationFactory.createJCRPath(getLocation(), relPath);
      // throw if not found
      NodeImpl pNode = (NodeImpl) session.getItem(newPath.makeParentPath().getAsString(false));
      // make dummy node in memory (nodeType will be replaced)
      node = this.createChildNode(pNode.getName(), "nt:base", true);
    }

    node.restore(version, removeExisting);
  }

  public void restoreByLabel(String versionLabel, boolean removeExisting) throws VersionException,
      ItemExistsException, UnsupportedRepositoryOperationException, LockException,
      RepositoryException, InvalidItemStateException {

    VersionImpl version = (VersionImpl) getVersionHistory().getVersionByLabel(versionLabel);
    restore(version, removeExisting);

  }

  public NodeIterator merge(String srcWorkspace, boolean bestEffort)
      throws UnsupportedRepositoryOperationException, NoSuchWorkspaceException,
      AccessDeniedException, MergeException, RepositoryException, InvalidItemStateException {

    if (session.hasPendingChanges())
      throw new InvalidItemStateException("Session has pending changes ");

    Map<String, String> failed = new HashMap<String, String>();

    // get corresponding node
    SessionImpl srcSession = (SessionImpl) ((RepositoryImpl) session.getRepository()).login(session
        .getCredentials(), srcWorkspace);
    NodeImpl correspNode;
    try {
      correspNode = getCorrespondingNode(srcSession);
    } catch (ItemNotFoundException e) {
      if (log.isDebugEnabled()) log.debug("No corresponding node in workspace: " + srcWorkspace);
      return new EntityCollection();
    }

    // TODO May be call: reloadNodeData()
    //loadData(this.getActualItemData());
    
    // Manually reloading data to actual state, 
    // force lazy loading of childs - they will be used in node removing/adding (inside WDM)  
    setRefreshToPersistedVersion(true);
    reloadNodeData();
    loadChildNodes();
    loadChildProperties();
    setRefreshToPersistedVersion(false);

    MergeVisitor visitor = new MergeVisitor(this, failed, bestEffort);
    correspNode.accept(visitor);
    
    // [PN] 08.04.06
    //----------- END SESSION BATCH ----------    
    dataManager.getWorkspaceDataManager().beginSessionBatch(session);
    state = ItemChangeState.DELETED;
    dataManager.getTransactManager().saveItem(this);
    
    state = ItemChangeState.ADDED;
    dataManager.getTransactManager().saveItem(this);
    dataManager.getWorkspaceDataManager().endSessionBatch(session);
    //----------- END SESSION BATCH ----------
    
    dataManager.resetItemState(this);

    NodeIterator failedIter = createMergeFailed(failed);
    
    // [PN] 01.08.06
//    for (ItemImpl cItem: visitor.getCleanupList()) {
//      // Session cache cleaning
//      // [PN] 01.08.06
//      cItem.setRefreshToPersistedVersion(true);
//      //dataManager.removeFromCache(cItem, true);
//      dataManager.removeFromCache(cItem);
//    }
    
    // Mark this node and her childs for reload with persistent state
    dataManager.reloadCache(getLocation(), false);
    
    return failedIter;
  }

  private NodeIterator createMergeFailed(Map<String, String> failed) throws RepositoryException {

    EntityCollection res = new EntityCollection();

    //for (Iterator it = failed.keySet().iterator(); it.hasNext();) {
    for (String uuid: failed.keySet()) {  

      NodeImpl versionable = (NodeImpl) session.getNodeByUUID(uuid);
      res.add(versionable);
      String offendingUuid = failed.get(uuid);
      Value[] mergeFailed;
      try {
        mergeFailed = versionable.getProperty("jcr:mergeFailed").getValues();
      } catch (PathNotFoundException e) {
        mergeFailed = new Value[0];
      }

      HashSet<String> mf = new HashSet<String>();
      // add "old" values
      for (int i = 0; i < mergeFailed.length; i++) {
        mf.add(mergeFailed[i].getString());
      }
      // if not exists add "new" value and refresh jcr:mergeFailed property
      if (!mf.contains(offendingUuid)) {
        mf.add(offendingUuid);
        Value[] vals = new Value[mf.size()];
        int i = 0;
        for (String ofs: mf) {
          vals[i] = valueFactory.createValue(ofs, PropertyType.REFERENCE);
        }

        PropertyImpl prop = doSetProperty("jcr:mergeFailed", vals, PropertyType.REFERENCE, true,
            true, false);
        getParent().save();
      }
    }

    return res;
  }

  public void doneMerge(Version version) throws VersionException, InvalidItemStateException,
      UnsupportedRepositoryOperationException, RepositoryException {
    VersionImpl base = (VersionImpl) getBaseVersion();
    base.addPredecessor(version);
    removeMergeFailed(version);
  }

  public void cancelMerge(Version version) throws VersionException, InvalidItemStateException,
      UnsupportedRepositoryOperationException, RepositoryException {
    removeMergeFailed(version);
  }

  private void removeMergeFailed(Version version) throws RepositoryException {
    
    boolean exists = false;
    if (!hasProperty("jcr:mergeFailed"))
      return;
    PropertyImpl mergeFailed = (PropertyImpl) getProperty("jcr:mergeFailed");
    Value[] mergeFailedRefs = mergeFailed.getValues();
    List<String> mf = new ArrayList<String>();
    for (Value mfv: mergeFailedRefs) {
      // System.out.println("REMOVE FAILED "+mergeFailed[i].getString()+" "+version.getUUID());
      if (!mfv.getString().equals(version.getUUID())) {
        mf.add(mfv.getString());
      }
    }
   
    if (mf.size() > 0) {
      mergeFailedRefs = new Value[mf.size()];
      for (int i = 0; i < mergeFailedRefs.length; i++) {
        mergeFailedRefs[i] = valueFactory.createValue(mf.get(i), PropertyType.REFERENCE);
      }
      // remove prev object stores prop jcr:mergeFailed
      mergeFailed.removeFromParent(ItemChangeState.UNCHANGED);
      // one new for prop jcr:mergeFailed
      mergeFailed = createChildProperty("jcr:mergeFailed", mergeFailedRefs, PropertyType.REFERENCE);
      //mergeFailed.markUpdated(); // as updated
      //saveItem(mergeFailed);
    } else {
      // [PN] 12.04.06 Remove whole prop
      mergeFailed.removeFromParent();
      //saveItem(mergeFailed);
    }
    //----------- BEGIN SESSION BATCH ----------      
    dataManager.getWorkspaceDataManager().beginSessionBatch(session);
    dataManager.getTransactManager().saveItem(mergeFailed);
    dataManager.getWorkspaceDataManager().endSessionBatch(session);
    //----------- END SESSION BATCH ----------
    dataManager.resetItemState(mergeFailed);
    dataManager.removeFromCache(mergeFailed);
  }

  // Locks
  public Lock lock(boolean isDeep, boolean isSessionScoped)
      throws UnsupportedRepositoryOperationException, LockException, AccessDeniedException,
      RepositoryException {

    if (!isNodeType("mix:lockable"))
      throw new LockException("Node is not lockable " + getPath());

    Lock lock = session.getLockManager().getLock(this, false);
    if (lock != null)
      throw new LockException("Node is already locked " + getPath() + " on "
          + lock.getNode().getPath());

    if (isDeep && session.getLockManager().getChildLock(this) != null)
      throw new LockException("Lock is deep and node already has locked child " + getPath());

    if (session.getTransientNodesManager().hasPendingChanges(getLocation(), false))
      throw new InvalidItemStateException("Node has pending unsaved changes " + getPath());

    PropertyImpl lockOwner = createChildProperty("jcr:lockOwner", new Value[] { valueFactory
        .createValue(session.getUserID()) }, PropertyType.STRING);
    PropertyImpl lockIsDeep = createChildProperty("jcr:lockIsDeep", new Value[] { valueFactory
        .createValue(isDeep) }, PropertyType.BOOLEAN);
    dataManager.getWorkspaceDataManager().beginSessionBatch(session);
    dataManager.getTransactManager().saveItem(lockOwner);
    dataManager.getTransactManager().saveItem(lockIsDeep);
    //dataManager.getWorkspaceDataManager().add(lockOwner);
    //dataManager.getWorkspaceDataManager().add(lockIsDeep);
    dataManager.getWorkspaceDataManager().endSessionBatch(session);

    Lock newLock = session.getLockManager().addLock(this, isDeep, isSessionScoped);
    return newLock;

  }
  
  /**
   * Lock node in session using SessionDataManager instead WorkspaceManager. 
   * Used in orderBefore, for relock nodes with sameNameSibs and index change.
   * We can't use Node.lock() because it works direct with workspace and needs save after that.
   * But we need one session scope in orderBefore for observation works true.
   * */
  private Lock lockInSession(boolean isDeep, boolean isSessionScoped)
      throws UnsupportedRepositoryOperationException, LockException, AccessDeniedException,
      RepositoryException {

    // [PN] 23.02.06 
    if (!isNodeType("mix:lockable"))
      throw new LockException("Node is not lockable " + getPath());

    Lock lock = session.getLockManager().getLock(this, false);
    if (lock != null)
      throw new LockException("Node is already locked " + getPath() + " on "
          + lock.getNode().getPath());

    if (isDeep && session.getLockManager().getChildLock(this) != null)
      throw new LockException("Lock is deep and node already has locked child " + getPath());

    PropertyImpl lockOwner = createChildProperty("jcr:lockOwner", new Value[] { valueFactory
        .createValue(session.getUserID()) }, PropertyType.STRING);
    PropertyImpl lockIsDeep = createChildProperty("jcr:lockIsDeep", new Value[] { valueFactory
        .createValue(isDeep) }, PropertyType.BOOLEAN);
    
    dataManager.update(lockOwner);
    dataManager.update(lockIsDeep);

    Lock newLock = session.getLockManager().addLock(this, isDeep, isSessionScoped);
    return newLock;

  }

  public Lock getLock() throws UnsupportedRepositoryOperationException, LockException,
      AccessDeniedException, RepositoryException {
    LockImpl lock = session.getLockManager().getLock(this, false);
    if (lock == null)
      throw new LockException("Lock not found " + getPath());
    return lock;

  }

  public void unlock() throws UnsupportedRepositoryOperationException, LockException,
      AccessDeniedException, RepositoryException {

    if (!session.hasLockToken(session.getLockManager().getLockToken(this)))
      throw new LockException("There are no permission to unlock the node or node is not locked "
          + getPath());

    if (session.getTransientNodesManager().hasPendingChanges(getLocation(), false))
      throw new InvalidItemStateException("Node has pending unsaved changes " + getPath());

    PropertyImpl lockOwner = (PropertyImpl) getProperty("jcr:lockOwner");
    PropertyImpl lockIsDeep = (PropertyImpl) getProperty("jcr:lockIsDeep");

    // [PN] 09.04.06
    //removeChildProperty(lockOwner);
    //removeChildProperty(lockIsDeep);

    //----------- BEGIN SESSION BATCH ----------    
    dataManager.getWorkspaceDataManager().beginSessionBatch(session);

    lockOwner.removeFromParent();
    dataManager.getTransactManager().saveItem(lockOwner);
    //dataManager.getWorkspaceDataManager().delete(lockOwner);
    
    lockIsDeep.removeFromParent();
    dataManager.getTransactManager().saveItem(lockIsDeep);
    //dataManager.getWorkspaceDataManager().delete(lockIsDeep);
    
    dataManager.getWorkspaceDataManager().endSessionBatch(session);
    //----------- END SESSION BATCH ----------    

    dataManager.resetItemState(lockOwner);
    dataManager.resetItemState(lockIsDeep);
    
    dataManager.removeFromCache(this);
    
    session.getLockManager().removeLock(this);
  }

  /**
   * Unlock node in session using SessionDataManager instead WorkspaceManager. 
   * Used in orderBefore, for relock nodes with sameNameSibs and index change.
   * We can't use Node.unlock() because it works direct with workspace and needs save after that.
   * But we need one session scope in orderBefore for observation works true.
   * */
  private void unlockInSession() throws UnsupportedRepositoryOperationException, LockException,
      AccessDeniedException, RepositoryException {

    if (!session.hasLockToken(session.getLockManager().getLockToken(this)))
      throw new LockException("There are no permission to unlock the node or node is not locked "
          + getPath());

    PropertyImpl lockOwner = (PropertyImpl) getProperty("jcr:lockOwner");
    PropertyImpl lockIsDeep = (PropertyImpl) getProperty("jcr:lockIsDeep");
    removeChildProperty(lockOwner);
    removeChildProperty(lockIsDeep);
    
    dataManager.delete(lockOwner);
    dataManager.delete(lockIsDeep);

    session.getLockManager().removeLock(this);
  }

  public boolean holdsLock() throws RepositoryException {
    return (session.getLockManager().getLock(this, true) != null);
  }

  public boolean isLocked() throws RepositoryException {
    return (session.getLockManager().getLock(this, false) != null);
  }

  boolean checkLocking() throws RepositoryException {
    Lock lock = session.getLockManager().getLock(this, false);
    return lock == null || // not locked 
      session.hasLockToken(lock.getLockToken()) || // has lock token 
      session.getUserID().equals(SystemIdentity.SYSTEM); // has system rights 
    
    // [PN] 21.09.06
//    return (!isLocked() || // not locked
//        session.hasLockToken(session.getLockManager().getLockToken(this)) || // has lock toc 
//        session.getUserID().equals(SystemIdentity.SYSTEM));
  }

  // ////////////////// Item implementation ////////////////////

  /**
   * @see javax.jcr.Item#accept
   */
  public void accept(ItemVisitor visitor) throws RepositoryException {
    visitor.visit(this);
  }

  /**
   * @see javax.jcr.Item#isNode
   */
  public boolean isNode() {
    return true;
  }

  // ////////////////////////////////////////////////////

  /**
   * [PN] 23.09.06 Issue JCR-139
   */
  protected Value[] initAutoCreatedValues(int requiredType, Value[] defValues) {
    if (defValues == null) {
      defValues = new Value[1];
      if (requiredType == PropertyType.DATE)
        defValues[0] = valueFactory.createValue(dataManager.getWorkspaceDataManager().getCurrentTime());
      else
        defValues[0] = null;
    }
    
    return defValues;
  }
  
  protected void addAutoCreatedItems(NodeType type, boolean registerInDataManager)
      throws RepositoryException, ConstraintViolationException {

    NodeDefinition[] nodeDefs = type.getChildNodeDefinitions();
    PropertyDefinition[] propDefs = type.getPropertyDefinitions();

    // Add autocreated child properties
    for (int i = 0; i < propDefs.length; i++) {

      if (propDefs[i] == null) // it is possible for not mandatory propDef
        continue;

      if (propDefs[i].isAutoCreated()) {
        // [PN] 23.09.06 Issue JCR-139
        
        //Value[] propVal = propDefs[i].getDefaultValues();
        // there can be null in definition but should not be null value
//        if (propVal == null) {
//          propVal = new Value[1];
//          propVal[0] = null;
//        }
        
        Value[] propVal = initAutoCreatedValues(propDefs[i].getRequiredType(), propDefs[i].getDefaultValues());
        
        boolean multiple = propDefs[i].isMultiple();
        String name = propDefs[i].getName();
        JCRPath loc = session.getLocationFactory().createJCRPath(getLocation(), name);
//        PropertyImpl prop = doSetProperty(InternalQName.parse(name), propVal, propDefs[i].getRequiredType(), 
//            multiple, registerInDataManager, false);
        PropertyImpl prop = doSetProperty(name, propVal, propDefs[i].getRequiredType(), multiple,
            registerInDataManager, false);
        //log.info("   Add autocteated property: " + prop.getPath());
      }
    }

    // Add autocreated child nodes
    for (int i = 0; i < nodeDefs.length; i++) {
      if (nodeDefs[i].isAutoCreated()) {
        //JCRName nodeName = locationFactory.parseJCRName(nodeDefs[i].getName());
        //JCRName primaryTypeName = locationFactory.parseJCRName(nodeDefs[i].getDefaultPrimaryType().getName());
        
        // [PN] 20.07.06
        NodeImpl node = doAddNode(this, InternalQName.parse(nodeDefs[i].getName()),
            InternalQName.parse(nodeDefs[i].getDefaultPrimaryType().getName()), 
            null, null, -1, // mixin, uuid, index
            registerInDataManager, 
            false, // doExternalValidation=false 
            true);
        //log.info("   Add autocteated node: " + node.getPath());
        //NodeImpl child = doAddNode(this, nodeName, primaryTypeName, null, true, true);
        // put to dataManager also if parent not new
        // [PN] 11.04.06 already placed in DM before
        //if (registerInDataManager) {
        //  dataManager.update(child);
        //}
      }
    }
    initAutoCreatedItems(type, registerInDataManager);
  }

  private NodeTypeImpl nodeType(InternalQName qName) throws NoSuchNodeTypeException, RepositoryException {
    NodeTypeImpl nt = ((NodeTypeManagerImpl)session.getWorkspace().getNodeTypeManager()).findNodeType(qName);
    if(nt == null)
      throw new NoSuchNodeTypeException("NodeTypeManager.getNodeType(): NodeType '" + qName.getAsString()
          + "' not found.");
    return nt;
  }

  public NodeType[] getAllNodeTypes() throws RepositoryException {
    NodeType primaryType = getPrimaryNodeType();
    NodeType[] mixinTypes = getMixinNodeTypes();
    NodeType[] types = new NodeType[mixinTypes.length + 1];
    types[0] = primaryType;
    for (int j = 0; j < mixinTypes.length; j++) {
      types[j + 1] = mixinTypes[j];
    }
    return types;
  }
  
  public PropertyDefinition findPropertyDefinition(String propertyName) throws RepositoryException {
    // [PN] 31.03.06 Fixed a property definition search.
    NodeType[] allTypes = getAllNodeTypes();
    PropertyDefinition pdResidual = null; // [PN] 31.03.06
    for (NodeType nt: allTypes) {
//      if (propertyName.equals("jcr:created")) {
//        PropertyDefinition[] pds1 = ((NodeTypeImpl) nt).getPropertyDefinitions();
//        for (PropertyDefinition pd: pds1) {
//          if (pd.getName().equals("jcr:created")) {
//            log.info("jcr:created >>> " + OnParentVersionAction.nameFromValue(pd.getOnParentVersion()));
//          }
//        }
//      }
      PropertyDefinitions pds = ((NodeTypeImpl) nt).getPropertyDefinitions(propertyName);
      PropertyDefinition pd = pds.getAnyDefinition();
      if ( pd != null) {
        if (((PropertyDefinitionImpl)pd).isResidualSet()) {
          pdResidual = pd;
        } else {
          return pd;
        }
      }
    }
    return pdResidual;
  }

  public String[] getMixinTypeNames() throws RepositoryException {
    NodeType[] mixinTypes = getMixinNodeTypes();
    String[] mtNames = new String[mixinTypes.length];
    for (int i = 0; i < mtNames.length; i++)
      mtNames[i] = mixinTypes[i].getName();
    return mtNames;
  }

  public void validateMandatoryChildren() throws ConstraintViolationException, RepositoryException {
    ArrayList mandatoryItemDefs = ((NodeTypeImpl) getPrimaryNodeType()).getManadatoryItemDefs();
    NodeType[] mixinTypes = getMixinNodeTypes();
    for (int i = 0; i < mixinTypes.length; i++)
      mandatoryItemDefs.addAll(((NodeTypeImpl) mixinTypes[i]).getManadatoryItemDefs());
    Iterator defs = mandatoryItemDefs.iterator();
    while (defs.hasNext()) {
      ItemDefinition def = (ItemDefinition) defs.next();
      if ((def instanceof NodeDefinition && !hasNode(def.getName()))
          || (def instanceof PropertyDefinition && !hasProperty(def.getName())))

        throw new ConstraintViolationException("Mandatory item " + def.getName()
            + " not found. Node [" + getPath() + " primary type: "
            + this.getPrimaryNodeType().getName() + "]");
    }
  }

  public void initAutoCreatedItems(NodeType type, boolean registerInDataManager)
      throws RepositoryException {

    if (type.isNodeType("nt:base")) {
      PropertyImpl prop = doSetProperty("jcr:primaryType", new Value[] { valueFactory
          .createValue(primaryType) }, PropertyType.NAME, false, false, false);
    }
    if (type.isNodeType("mix:referenceable")) {
      doSetProperty("jcr:uuid", new Value[] { valueFactory.createValue(getInternalUUID()) },
          PropertyType.STRING, false, false, false);
    }
    if (type.isNodeType("nt:hierarchyNode")) {
      Value[] values = new Value[] { valueFactory.createValue(dataManager.getWorkspaceDataManager()
          .getCurrentTime()) };
      doSetProperty("jcr:created", values, PropertyType.DATE, false, false, false);
    }
    if (!session.getAccessManager().getAccessControlPolicy().equals(AccessControlPolicy.DISABLE)
        && type.isNodeType("exo:accessControllable")) {
      Map perms = new HashMap();
      perms.put(SystemIdentity.ANY, PermissionType.DEFAULT_AC);
      setPermissions(perms);
    }
    if (type.isNodeType("mix:versionable")) {

      // Generate uuids
      Uuid versionHistoryUUID = uuidGenerator.generateUUID(getPath() + "/jcr:versionHistory");
      Uuid baseVersionUUID = uuidGenerator.generateUUID(getPath() + "/jcr:baseVersion");

      doSetProperty("jcr:versionHistory", new Value[] { valueFactory
          .createValue(versionHistoryUUID) }, PropertyType.REFERENCE, false, registerInDataManager,
          false);
      doSetProperty("jcr:baseVersion", new Value[] { valueFactory.createValue(baseVersionUUID) },
          PropertyType.REFERENCE, false, registerInDataManager, false);
      doSetProperty("jcr:predecessors", new Value[] { valueFactory.createValue(baseVersionUUID) },
          PropertyType.REFERENCE, true, registerInDataManager, false);
      doSetProperty("jcr:isCheckedOut", new Value[] { valueFactory.createValue(true) },
          PropertyType.BOOLEAN, false, registerInDataManager, false);
    }
  }

  // ExtendedNode -----------------------------

  /* (non-Javadoc)
   * @see org.exoplatform.services.jcr.core.ExtendedNode#setPermissions(java.util.Map)
   */
  public void setPermissions(Map permissions) throws RepositoryException,
      AccessDeniedException, AccessControlException {

    checkPermission(PermissionType.CHANGE_PERMISSION);

    String ownerName;
    // First who make it accessControllable is owner
    if (!isNodeType("exo:accessControllable")) {
      ownerName = session.getUserID();
      addMixin("exo:accessControllable");
    } else {
      if (!getACL().getOwner().equals(SystemIdentity.SYSTEM))
        ownerName = getACL().getOwner();
      else
        ownerName = session.getUserID();

    }
    setACL(new AccessControlList(ownerName, permissions));
    
    // populate exo:accessControllable
    Value[] owner = new Value[] { valueFactory.createValue(getACL().getOwner()) };

    boolean registerInDataManager = (this.state != ItemChangeState.ADDED);
    doSetProperty("exo:owner", owner, PropertyType.STRING, false, registerInDataManager, false);

    saveChildPermissionsProperty();
  }

  private void saveChildPermissionsProperty() throws RepositoryException {

    boolean registerInDataManager = (this.state != ItemChangeState.ADDED);
    //Map permissions = childrenACL.getPermissions();
    Map permissions = getACL().getPermissions();
    List permValues = new ArrayList();
    for (Iterator i = permissions.keySet().iterator(); i.hasNext();) {
      String identity = (String) i.next();
      String[] perm = (String[]) permissions.get(identity);
      try {
        for (int j = 0; j < perm.length; j++)
          permValues.add(new PermissionValue(identity, perm[j]));
      } catch (IOException e) {
        throw new RepositoryException(e);
      }
    }

    Value[] permArray = new Value[permValues.size()];
    for (int i = 0; i < permArray.length; i++)
      permArray[i] = (Value) permValues.get(i);

    doSetProperty("exo:permissions", permArray, ExtendedPropertyType.PERMISSION, true,
        registerInDataManager, false);

  }

  /* (non-Javadoc)
   * @see org.exoplatform.services.jcr.core.ExtendedNode#getACL()
   */
  public AccessControlList getACL() {
    return ((NodeData)data).getACL();
  }
  
  private void setACL(AccessControlList acl) {
    ((NodeData)data).setACL(acl);
  }

  /* (non-Javadoc)
   * @see org.exoplatform.services.jcr.core.ExtendedNode#clearACL()
   */
  public void clearACL() throws RepositoryException, AccessControlException {
    checkPermission(PermissionType.CHANGE_PERMISSION);
    if (isNodeType("exo:accessControllable")) {
      removeMixin("exo:accessControllable");
    }
    setACL(new AccessControlList(true));
  }

  /* (non-Javadoc)
   * @see org.exoplatform.services.jcr.core.ExtendedNode#removePermission(java.lang.String)
   */
  public void removePermission(String identity) throws RepositoryException,
      AccessControlException {
    checkPermission(PermissionType.CHANGE_PERMISSION);
    // remove default permission
    if (!isNodeType("exo:accessControllable")) {
      throw new AccessControlException(
          "It is impossible to remove default permissions. Use setPermission(s) instead");
    }
    getACL().getPermissions().remove(identity);
    saveChildPermissionsProperty();
  }

  /* (non-Javadoc)
   * @see org.exoplatform.services.jcr.core.ExtendedNode#setPermission(java.lang.String, java.lang.String[])
   */
  public void setPermission(String identity, String[] permission)
      throws RepositoryException, AccessControlException {
    // check if changing permission allowed
    checkPermission(PermissionType.CHANGE_PERMISSION);
    // remove default permission
    if (!isNodeType("exo:accessControllable")) {
      Map perms = new HashMap();
      perms.put(identity, permission);
      setPermissions(perms);
    } else {
      getACL().getPermissions().put(identity, permission);
      saveChildPermissionsProperty();
    }
  }

  /* (non-Javadoc)
   * @see org.exoplatform.services.jcr.core.ExtendedNode#checkPermission(java.lang.String)
   */
  public void checkPermission(String actions) throws AccessControlException,
      RepositoryException {
    if (!session.getAccessManager().hasPermission(getACL(), actions, session.getUserID()))
      throw new AccessControlException("Permission denied " + getPath() + " : " + actions);

  }


  public void checkChildrenPermission(String actions) throws AccessControlException, RepositoryException {
    checkPermission(actions);
  }

  public void clearChildrenACL() throws RepositoryException, AccessControlException {
    clearACL();  
  }

  public AccessControlList getChildrenACL() {
    return getACL();
  }

  public void removeChildrenPermission(String identity) throws RepositoryException, AccessControlException {
    removePermission(identity);
  }

  public void setChildrenPermission(String identity, String[] permissions) throws RepositoryException, AccessControlException {
    setPermission(identity, permissions);
  }

  public void setChildrenPermissions(Map permissions) throws RepositoryException, AccessControlException {
    setPermissions(permissions);    
  }

  // ---------------------------------------------------------


  /**
   * @return Returns the mixinTypes.
   */
  JCRName[] getMixinTypes() {
    return mixinTypes;
  }

  /**
   * @return Returns the primaryType.
   */
  JCRName getPrimaryType() {
    return primaryType;
  }

  boolean isChildNodesInitialized() {
    return childNodes != null;
  }

  boolean isChildPropertiesInitialized() {
    return childProperties != null;
  }

  //----------------------- Object.equals() ----------------------

  /**
   * @author [PN] 18.04.06
   * @see java.lang.Object#equals(java.lang.Object)
   */
  @Override
  public boolean equals(Object obj) {
    if (obj instanceof NodeImpl) {
      NodeImpl otherNode = (NodeImpl) obj;
      try {
        if (otherNode.isNodeType("mix:referenceable") && this.isNodeType("mix:referenceable")) {
          // by UUID
          //return isSame((Node) obj); // may be changed
          // getProperty("jcr:uuid") is more correct, but may decrease performance 
          return getInternalUUID().equals(otherNode.getInternalUUID());
        }
        // by path
        return getLocation().equals(otherNode.getLocation());
      } catch(RepositoryException e) {
        return false;
      }
    }
    return false;
  }
  
  //////////////////// NEW METHODS (since 1.2) ////////////////
  
  public PropertyDefinition findChildPropertyDefinition(InternalQName propertyName) 
  throws RepositoryException, ConstraintViolationException {
    NodeType[] allTypes = getAllNodeTypes();
    PropertyDefinition pdResidual = null; 
    for (NodeType nt: allTypes) {
      PropertyDefinitions pds = ((NodeTypeImpl) nt).getPropertyDefinitions(propertyName);
      PropertyDefinition pd = pds.getAnyDefinition();
      if ( pd != null) {
        if (((PropertyDefinitionImpl)pd).isResidualSet()) {
          pdResidual = pd;
        } else {
          return pd;
        }
      }
    }
    if(pdResidual == null)
      throw new ConstraintViolationException("Child property definition not found for "+propertyName+" "+getPath());
    return pdResidual;
  }

  public NodeDefinition findChildNodeDefinition(InternalQName nodeName) 
  throws RepositoryException, ConstraintViolationException {
    NodeDefinition ndResidual = null; 
    for (NodeType nt: getAllNodeTypes()) {
      NodeDefinitionImpl nd = (NodeDefinitionImpl)((NodeTypeImpl) nt).getChildNodeDefinition(nodeName); 
      if ( nd != null) {
        if (nd.isResidualSet()) {
          ndResidual = nd;
        } else {
          return nd;
        }
      }
    }
    if(ndResidual == null)
      throw new ConstraintViolationException("Child node definition not found for "+nodeName+" "+getPath());

    return ndResidual;
  }

  
//  /**
//   * [PN] 19.04.06 Don't used. For session refactoring.
//   *  
//   * Note (from J2SE API):
//   * It is not required that if two objects are unequal according to the equals(java.lang.Object) method, 
//   * then calling the hashCode method on each of the two objects must produce distinct integer results. 
//   * However, the programmer should be aware that producing distinct integer results for unequal objects 
//   * may improve the performance of hashtables.
//   * 
//   * TODO If a node (or property), as java object, located in repository, as JCR item, by path 
//   * will be relocated (e.g. move, copy in other workspace, etc.) it's will be true to change
//   * hash code on the item.
//   */
//  protected int calcHashCode() {
//    try {
//      if (this.isNodeType("mix:referenceable")) {
//        // by UUID
//        itemHashCode = (session.getWorkspace().getName() + getInternalUUID()).hashCode();
//      }
//      // by path
//      itemHashCode = (session.getWorkspace().getName() + getLocation().getAsString(false)).hashCode();
//    } catch(Exception e) {
//    }
//    return itemHashCode;
//  }  
  
}
