/***************************************************************************
 * Copyright 2001-2006 The eXo Platform SARL         All rights reserved.  *
 * Please look at license.txt in info directory for more license detail.   *
 **************************************************************************/

package org.exoplatform.services.jcr.impl.dataflow;

import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.UnsupportedEncodingException;
import java.util.Arrays;
import java.util.Calendar;

import org.exoplatform.services.jcr.access.AccessControlEntry;
import org.exoplatform.services.jcr.datamodel.AbstractValueData;
import org.exoplatform.services.jcr.datamodel.InternalQName;
import org.exoplatform.services.jcr.datamodel.InternalQPath;
import org.exoplatform.services.jcr.datamodel.Uuid;
import org.exoplatform.services.jcr.impl.Constants;
import org.exoplatform.services.jcr.impl.util.DateFormatHelper;
import org.exoplatform.services.jcr.impl.util.io.FileCleaner;

/**
 * Created by The eXo Platform SARL        .<br/>
 * 
 * session scoped shared value data
 *
 * @author Gennady Azarenkov
 * @version $Id: TransientValueData.java 9852 2006-10-28 11:34:29Z peterit $
 */

public class TransientValueData extends AbstractValueData {

  protected byte[] data;
  
  protected InputStream tmpStream;
  
  protected File spoolFile;
  
  protected FileCleaner fileCleaner;
  
  protected int maxBufferSize;
  
  protected File tempDirectory;
  

  static protected byte[] stringToBytes(final String value) {
    try {
      return value.getBytes(Constants.DEFAULT_ENCODING);
    } catch (UnsupportedEncodingException e) {
      throw new RuntimeException("FATAL ERROR Charset "
          + Constants.DEFAULT_ENCODING + " is not supported!");
    }
  }
  
  /**
   * creates TransientValueData with incoming byte array
   * @param value
   * @param orderNumber
   */
  public TransientValueData(byte[] value, int orderNumber) {
    super(orderNumber);
    this.data = value;
  }

  /**
   * creates TransientValueData with incoming input stream.
   * the stream will be lazily spooled to file or byte array depending on maxBufferSize 
   * @param value
   * @param orderNumber
   */
  public TransientValueData(InputStream stream, int orderNumber) {
    super(orderNumber);
    this.tmpStream = stream;
  }


  public TransientValueData(InputStream stream) {
    this(stream, 0);
  }  

  /**
   * Constructor for String value data
   * @param value
   */
  public TransientValueData(String value) {
    this(stringToBytes(value), 0);
  }

  /**
   * Constructor for boolean value data
   * @param value
   */
  public TransientValueData(boolean value) {
    this(Boolean.valueOf(value).toString().getBytes(), 0);
  }

  /**
   * Constructor for Calendar value data
   * @param value
   */
  public TransientValueData(Calendar value) {
    this(new DateFormatHelper().serialize(value), 0);
  }

  /**
   * Constructor for double value data
   * @param value
   */
  public TransientValueData(double value) {
    this(Double.valueOf(value).toString().getBytes(), 0);
  }
  
  /**
   * Constructor for long value data
   * @param value
   */
  public TransientValueData(long value) {
    this(Long.valueOf(value).toString().getBytes(), 0);
  }

  /**
   * Constructor for Name value data
   * @param value
   */
  public TransientValueData(InternalQName value) {
    this(value.getAsString().getBytes(), 0);
  }

  /**
   * Constructor for Path value data
   * @param value
   */
  public TransientValueData(InternalQPath value) {
    this(value.getAsString().getBytes(), 0);
  }

  /**
   * Constructor for Reference value data
   * @param value
   */
  public TransientValueData(Uuid value) {
    this(value.getString().getBytes(), 0);
  }
  
  /**
   * Constructor for Permission value data
   * @param value
   */
  public TransientValueData(AccessControlEntry value) {
    this(value.getAsString().getBytes(), 0);
  }


  /* (non-Javadoc)
   * @see org.exoplatform.services.jcr.datamodel.ValueData#getAsByteArray()
   */
  public byte[] getAsByteArray() throws IOException {
    spoolInputStream();
    if(data != null) {
      byte[] bytes = new byte[data.length];
      System.arraycopy(data, 0, bytes, 0, data.length);
      return bytes;
    } else {
      return fileToByteArray();
    }
  }

  /* (non-Javadoc)
   * @see org.exoplatform.services.jcr.datamodel.ValueData#getAsStream()
   */
  public InputStream getAsStream() throws IOException {
    spoolInputStream();
    if(data != null) {
      return new ByteArrayInputStream(data); // from bytes
    } else if (spoolFile != null){
      return new FileInputStream(spoolFile); // from spool file if initialized
    } else
      throw new NullPointerException("Null Stream data ");
    
  }

  /* (non-Javadoc)
   * @see org.exoplatform.services.jcr.datamodel.ValueData#getLength()
   */
  public long getLength() {
    try {
      spoolInputStream();
    } catch (IOException e) {
      e.printStackTrace();
      return -1;
    }
    if(data == null) {
      return spoolFile.length();
    } else {
      return data.length;
    }
  }

  /* 
   * returns true if this data is spooled to byte array, false otherwise (to file) 
   * @see org.exoplatform.services.jcr.datamodel.ValueData#isByteArray()
   */
  public boolean isByteArray() {
    try {
      spoolInputStream();
    } catch (IOException e) {
      e.printStackTrace();
      return false;
    }
    return data != null;
  }
  
  /**
   * @return spool file if any
   */
  public File getSpoolFile() {
    try {
      spoolInputStream();
    } catch (IOException e) {
      e.printStackTrace();
      return null;
    }
    return spoolFile;
  }
  
  /**
   * helper method to simplify operations that requires stringified data
   * @return
   * @throws IOException
   */
  public String getString() throws IOException {
    return new String(getAsByteArray(), Constants.DEFAULT_ENCODING);
  }
  
  /////////////////////////////////////
  /**
   * make sense for stream storage only
   * @param cleaner
   */
  public void setFileCleaner(FileCleaner cleaner) {
    this.fileCleaner = cleaner;
  }
  
  /**
   * @param tempDirectory
   */
  public void setTempDirectory(File tempDirectory) {
    this.tempDirectory = tempDirectory;
  }

  /**
   * @param maxBufferSize
   */
  public void setMaxBufferSize(int maxBufferSize) {
    this.maxBufferSize = maxBufferSize;
  }

  protected void finalize() throws Throwable {
    if(spoolFile != null)
      fileCleaner.addFile(spoolFile);
  }
  
  
  /* (non-Javadoc)
   * @see java.lang.Object#equals(java.lang.Object)
   */
  public boolean equals(Object obj) {
    
    if (this == obj) {
      return true;
    }
    
    if (obj instanceof TransientValueData) {

      TransientValueData other = (TransientValueData) obj;
      if(isByteArray() != other.isByteArray())
        return false;
      try {
        if(isByteArray()) {
          return Arrays.equals(getAsByteArray(), other.getAsByteArray());
        } else
          return getSpoolFile().equals(other.getSpoolFile());
      } catch (IOException e) {
        e.printStackTrace();
        return false;
      }
    }
    return false;
  }

  
  private boolean toBufferOnly() {
    return fileCleaner == null;
  }
  

  /////////////////////////////////////

  private void spoolInputStream() throws IOException {

    if(tmpStream == null) // already spooled 
      return;

    this.spoolFile = File.createTempFile("jcrvd", null, tempDirectory);
    ByteArrayOutputStream baos = new ByteArrayOutputStream();
    FileOutputStream fos = new FileOutputStream(spoolFile);

    byte[] buffer = new byte[0x2000];
    int len;
    int total = 0;
    try {
      while ((len = tmpStream.read(buffer)) > 0) {
        if(!toBufferOnly())
          fos.write(buffer, 0, len);
        total += len;
        if(total < maxBufferSize || toBufferOnly()) {
          baos.write(buffer, 0, len);
        } else {
          baos = null;
        }
      }
      fos.close();
      if(baos != null) { // spool to byte array
        this.spoolFile.delete();
        this.spoolFile = null;
        this.data = baos.toByteArray();
        baos.close();
      } else { // spool to file
        this.fileCleaner.addFile(spoolFile);
        this.data = null;
      }
      this.tmpStream = null;
    } catch (IOException e) {
      throw new IllegalStateException(e);
    }
  }

  /**
   * try to convert stream to byte array
   * WARNING: Potential lack of memory due to call getAsByteArray() on stream data
   * @return byte array
   */
  private byte[] fileToByteArray() throws IOException {
    ByteArrayOutputStream out = new ByteArrayOutputStream();
    
    byte[] buffer = new byte[0x2000];
    int len;
    int total = 0;
    FileInputStream stream = new FileInputStream(spoolFile); 
    while ((len = stream.read(buffer)) > 0) {
      out.write(buffer, 0, len);
      total += len;
      if (total > maxBufferSize)
        log.warn("Potential lack of memory due to call getAsByteArray() on stream data exceeded "
                + total + " bytes");
    }
    out.close();
    return out.toByteArray();
  }



}