/***************************************************************************
 * Copyright 2001-2006 The eXo Platform SARL         All rights reserved.  *
 * Please look at license.txt in info directory for more license detail.   *
 **************************************************************************/
package org.exoplatform.services.jcr.util.io;

import java.io.File;
import java.io.IOException;
import java.util.Map;
import java.util.WeakHashMap;

import org.apache.commons.logging.Log;
import org.exoplatform.services.log.ExoLogger;

/**
 * Created by The eXo Platform SARL
 * Author : Peter Nedonosko
 *          peter.nedonosko@exoplatform.com.ua
 * 24.10.2006
 * 
 * @author <a href="mailto:peter.nedonosko@exoplatform.com.ua">Peter Nedonosko</a>
 * @version $Id: SharedFile.java 10024 2006-11-02 17:52:33Z peterit $
 */
public class SharedFile extends File {
  
  static final public String DELETED_EXTENSION = "del";
  
  protected String FILE_SUBJECT_NAME = "shared";
  
  private static Log log = ExoLogger.getLogger("jcr.SharedFile");
  
  private Map<Object, Long> lockOwners = new WeakHashMap<Object, Long>();
  
  private volatile boolean deleted = false;
  
  protected boolean onFinalyzeDelete = false;
  private boolean finalyzed = false;
  
  protected SharedFile(String fileName) {
    super(fileName);
  }
  
  public synchronized boolean doDelete() {
    
    if (exists()) {
    
      if (isLockUsed()) {
        if (onFinalyzeDelete) {
          if (log.isDebugEnabled())
            log.debug("In use. Delete on finalyze " + getAbsolutePath());
          return false;
        }
        
        // Mark for delete on finalyze process.
        onFinalyzeDelete = true;
      } else {
        // Try delete now
        if (super.delete()) {
          deleteMarker();
          onFinalyzeDelete = false;
          if (log.isDebugEnabled())
            log.debug("Deleted " + getAbsolutePath());
        } else if (onFinalyzeDelete) {
          return false;
        } else {  
          // Mark for delete on finalyze process.
          onFinalyzeDelete = true;
        }
      }

      onDelete();
      
      if (onFinalyzeDelete) {
        // Perform additional logic giving us the file deletion guaranty in system crash situations etc.
        // Create delete marker with same name and .del extension.
        // SimpleFileValueStorage will check these files at start time and do cleanup.
        File deleteMarker = new File(getAbsolutePath() + "." + DELETED_EXTENSION);
        
        if (log.isDebugEnabled())
          log.debug("Mark for delete " + deleteMarker.getAbsolutePath());
        
        try {
          if (!deleteMarker.createNewFile()) {
            if (log.isDebugEnabled())
              log.debug("Delete marker exists or can't be created " + deleteMarker.getAbsolutePath());
          }
        } catch(IOException e) {
          // So, will goes without marker...
          log.warn("WARNING! Can't create delete marker for file " + getAbsolutePath());
        }
      }
    
      return deleted = !onFinalyzeDelete;
    }
    
    if (log.isDebugEnabled())
      log.debug("Not exists " + getAbsolutePath());
    
    return onFinalyzeDelete = false;
  }

  protected boolean deleteMarker() {
    File deleteMarker = new File(getAbsolutePath() + "." + DELETED_EXTENSION);
    if (deleteMarker.exists()) {
      if (!deleteMarker.delete())
        log.warn("Can't deleted marker file " + deleteMarker.getAbsolutePath());
      
      return true;
    }
    
    return false;
  }
  
  public boolean delete() {
    
    if (!deleted) {
      return doDelete();
    }
    
    if (log.isDebugEnabled())
      log.debug("Already deleted " + getAbsolutePath());
    return false;    
  }

  /**
   * On delete action.
   * This implementation provide no action.
   * 
   * Can be override for customizing java.io.File.delete() procedure after java.io.File.delete() call
   * but before the deletion marker creation (if needed, but before this occurs). 
   */
  protected void onDelete() {
    
  }
  
  // ------------ File users lock management -------------
  
  protected synchronized boolean isLockUsed() {
    return !isFinalyzed() && lockOwners.size()>0;
  }
  
  /**
   * Lock shared file as used in any consumer, e.g. LazyXXXFileInputStream
   * 
   * @return - true if opertion is successful
   */
  protected synchronized boolean lockUsed(Object lockOwner) {
    if (!isFinalyzed()) {
      lockOwners.put(lockOwner, System.currentTimeMillis());
      return true;
    }
    return false;
  }
  
  /**
   * Unlock shared file.
   * 
   * The method must be called from lock owner, 
   * i.e. consumer (LazyXXXFileInputStream) on stream close/finalization etc.
   * 
   * @return - true if opertion is successful
   */
  protected synchronized boolean unlockUsed(Object lockOwner) {
    if (!isFinalyzed()) {
      return lockOwners.remove(lockOwner) != null;
    }
    return false;
  }

  protected boolean isFinalyzed() {
    return finalyzed;
  }
  
  @Override
  protected void finalize() throws Throwable {
    
    finalyzed = true;
    try {
      if (onFinalyzeDelete) {
        if (log.isDebugEnabled())
          log.debug("Delete on finalyze " + getAbsolutePath());
        if(!super.delete())
          throw new IllegalStateException("Can not delete " + FILE_SUBJECT_NAME + " file " + getAbsolutePath() 
              + ". File is marked as deleted by java.io.File.delete() method call.");
        
        deleteMarker();
        
        if (log.isDebugEnabled())
          log.debug(FILE_SUBJECT_NAME + " file deleted " + getAbsolutePath());
      }
    } finally {
      super.finalize();
    }
  }    
}
