/***************************************************************************
 * Copyright 2001-2006 The eXo Platform SARL         All rights reserved.  *
 * Please look at license.txt in info directory for more license detail.   *
 **************************************************************************/
package org.exoplatform.services.jcr.impl.storage.jdbc;

import java.io.ByteArrayInputStream;
import java.io.File;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.sql.Connection;
import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.util.ArrayList;
import java.util.List;

import javax.jcr.InvalidItemStateException;
import javax.jcr.PropertyType;
import javax.jcr.RepositoryException;
import javax.sql.DataSource;

import org.apache.commons.logging.Log;
import org.exoplatform.services.jcr.access.AccessControlList;
import org.exoplatform.services.jcr.dataflow.persistent.PersistedNodeData;
import org.exoplatform.services.jcr.dataflow.persistent.PersistedPropertyData;
import org.exoplatform.services.jcr.datamodel.InternalQName;
import org.exoplatform.services.jcr.datamodel.InternalQPath;
import org.exoplatform.services.jcr.datamodel.ItemData;
import org.exoplatform.services.jcr.datamodel.NodeData;
import org.exoplatform.services.jcr.datamodel.PropertyData;
import org.exoplatform.services.jcr.datamodel.ValueData;
import org.exoplatform.services.jcr.impl.Constants;
import org.exoplatform.services.jcr.impl.dataflow.persistent.ByteArrayPersistedValueData;
import org.exoplatform.services.jcr.impl.dataflow.persistent.CleanableFileStreamValueData;
import org.exoplatform.services.jcr.impl.util.io.FileCleaner;
import org.exoplatform.services.jcr.storage.WorkspaceStorageConnection;
import org.exoplatform.services.jcr.storage.value.ValueIOChannel;
import org.exoplatform.services.jcr.storage.value.ValueStoragePluginProvider;
import org.exoplatform.services.log.ExoLogger;

/**
 * Created by The eXo Platform SARL        .
 * @author <a href="mailto:gennady.azarenkov@exoplatform.com">Gennady Azarenkov</a>
 * @version $Id: JDBCStorageConnection.java 9984 2006-11-01 15:05:38Z peterit $
 */

abstract public class JDBCStorageConnection extends DBConstants implements WorkspaceStorageConnection {

  protected static Log log = ExoLogger.getLogger("jcr.JDBCStorageConnection");

  protected PreparedStatement findItemById;
  protected PreparedStatement findItemByPath;

  @Deprecated
  protected PreparedStatement findChildItemByPath;
  protected PreparedStatement findChildPropertyByPath;

  protected PreparedStatement findDescendantNodes;
  protected PreparedStatement findDescendantProperties;

  //@Deprecated // [PN] TODO Refactor according JCR-56
  //protected PreparedStatement findPropertiesByValue;

  protected PreparedStatement findReferenceable;
  protected PreparedStatement findReferences;

  protected PreparedStatement findValuesByPropertyId;
  protected PreparedStatement findValueByPropertyIdOrderNumber;

  protected PreparedStatement findNodesByParentId;
  protected PreparedStatement findNodesIdsByParentId;
  protected PreparedStatement findPropertiesByParentId;

  protected PreparedStatement findNodesCountByParentId;
  protected PreparedStatement findPropertiesCountByParentId;

  protected PreparedStatement insertItem;
  protected PreparedStatement insertNode;
  protected PreparedStatement insertProperty;
  protected PreparedStatement insertReference;
  protected PreparedStatement insertValue;

  protected PreparedStatement updateItem;
  protected PreparedStatement updateItemPath;
  protected PreparedStatement updateNode;
  protected PreparedStatement updateProperty;
  //protected PreparedStatement updateValue;

  protected PreparedStatement deleteItem;
  protected PreparedStatement deleteNode;
  protected PreparedStatement deleteProperty;
  protected PreparedStatement deleteReference;
  protected PreparedStatement deleteValue;

  protected final ValueStoragePluginProvider valueStorageProvider;
  protected final int maxBufferSize;
  protected final File swapDirectory;
  protected final FileCleaner swapCleaner;

  protected final DataSource dataSource;
  protected final Connection dbConnection;

  protected final String containerName;

  protected final SQLExceptionHandler exceptionHandler;


  /**
   * @param dbConnection
   * @throws SQLException
   */
  protected JDBCStorageConnection(DataSource dataSource,
      String containerName, boolean multiDb,
      ValueStoragePluginProvider valueStorageProvider,
      int maxBufferSize, File swapDirectory, FileCleaner swapCleaner) throws SQLException {


    //super(dataSource.getConnection(), containerName, multiDb);
    this.dataSource = dataSource;
    this.valueStorageProvider = valueStorageProvider;

    this.maxBufferSize = maxBufferSize;
    this.swapDirectory = swapDirectory;
    this.swapCleaner = swapCleaner;
    this.containerName = containerName;


    this.dbConnection = dataSource.getConnection();

    // Fix for Sybase jConnect JDBC driver bug.
    // Which throws SQLException(JZ016: The AutoCommit option is already set to false)
    // if conn.setAutoCommit(false) called twise or more times with value 'false'.
    if (dbConnection.getAutoCommit()) {
      dbConnection.setAutoCommit(false);
    }

    prepareQueries();
    prepareStatements();

    this.exceptionHandler = new SQLExceptionHandler(containerName, this);

    //if (log.isDebugEnabled())
    //  log.debug("JDBCStorageConnection is opened using "+dbConnection.getClass().getName());
  }

  /**
   * For test only
   * Return JDBC connection obtained from initialized data source.
   * NOTE: Helper can obtain one new connection per each call of the method  or return one obtained once.
   *
   * @throws SQLException
   */
  public Connection getJdbcConnection() throws SQLException {
    return dbConnection;
  }

  /**
   * Prepared queries at start time
   * @throws SQLException
   */
  abstract protected void prepareQueries() throws SQLException;


  /**
   * Used in Single Db Connection classes for UUID related queries
   * @param uuid
   * @return
   */
  protected abstract String getInternalId(String uuid);

  /**
   * Used in loadXYZRecord methods for extract real UUID from container value.
   * @param internalId
   * @return
   */
  protected abstract String getUuid(String internalId);

  // ---------------- WorkspaceStorageConnection -------------

  /**
   * @throws IllegalStateException if connection is closed
   */
  protected void checkIfOpened() throws IllegalStateException {
    if(!isOpened())
      throw new IllegalStateException("Connection is closed");
  }

  /** Will be used as implementation of
   * @see org.exoplatform.services.jcr.storage.WorkspaceStorageConnection#isOpened()
   */
  public boolean isOpened() {
    try {
      return !dbConnection.isClosed();
    } catch (SQLException e) {
      log.error(e);
      return false;
    }
  }

  /** Will be used as implementation of
   * @see org.exoplatform.services.jcr.storage.WorkspaceStorageConnection#rollback()
   */
  public final void rollback() throws IllegalStateException, RepositoryException {
    checkIfOpened();
    try {
      //Connection con = getJdbcConnection();
      dbConnection.rollback();
      dbConnection.close();
      //dbConnection = null;
    } catch (SQLException e) {
      throw new RepositoryException(e);
    }
  }

  /** Will be used as implementation of
   * @see org.exoplatform.services.jcr.storage.WorkspaceStorageConnection#commit()
   */
  public final void commit() throws IllegalStateException, RepositoryException {
    checkIfOpened();
    try {
      //Connection con = getJdbcConnection();
      dbConnection.commit();
      dbConnection.close();
      //con = null;
    } catch (SQLException e) {
      throw new RepositoryException(e);
    }
  }

  /* (non-Javadoc)
   * @see org.exoplatform.services.jcr.storage.WorkspaceStorageConnection#add(org.exoplatform.services.jcr.datamodel.NodeData)
   */
  public void add(NodeData data) throws RepositoryException, UnsupportedOperationException, InvalidItemStateException, IllegalStateException {
    checkIfOpened();
    try {
      addNodeRecord(data);
      if (log.isDebugEnabled())
        log.debug("Node added " + data.getQPath().getAsString() + ", " + data.getUUID() + ", " + data.getPrimaryTypeName().getAsString());

    } catch (SQLException e) {
      log.error("Node add. Database error: " + e, e);
      exceptionHandler.handleAddException(e, data);
    }
  }

  /* (non-Javadoc)
   * @see org.exoplatform.services.jcr.storage.WorkspaceStorageConnection#add(org.exoplatform.services.jcr.datamodel.PropertyData)
   */
  public void add(PropertyData data) throws RepositoryException, UnsupportedOperationException, InvalidItemStateException, IllegalStateException {
    checkIfOpened();

    try {
      addPropertyRecord(data);

      if (data.getType() == PropertyType.REFERENCE) {
        try {
          addReference(data);
        } catch(IOException e) {
          throw new RepositoryException("Can't read REFERENCE property ("+data.getQPath()+" "+data.getUUID()+") value: " + e.getMessage(), e);
        }
      }

      ValueIOChannel channel = valueStorageProvider.getApplicableChannel(data);
      if (channel != null) {
        channel.write(data.getUUID(), data.getValues());
        channel.close();
      } else {
        addValues(getInternalId(data.getUUID()), data.getValues());
      }

      if (log.isDebugEnabled())
        log.debug("Property added " + data.getQPath().getAsString() + ", " + data.getUUID()
            + (data.getValues() != null ? ", values count: " + data.getValues().size() : ", NULL data")
            + ", use channel "+channel);

    } catch (IOException e) {
      log.error("Property add. IO error: " + e, e);
      exceptionHandler.handleAddException(e, data);
    } catch (SQLException e) {
      log.error("Property add. Database error: " + e, e);
      exceptionHandler.handleAddException(e, data);
    }
  }


  /* (non-Javadoc)
   * @see org.exoplatform.services.jcr.storage.WorkspaceStorageConnection#delete(org.exoplatform.services.jcr.datamodel.ItemData)
   */
  public void delete(ItemData data) throws RepositoryException, UnsupportedOperationException, InvalidItemStateException, IllegalStateException {
    checkIfOpened();

    final String cid = getInternalId(data.getUUID());

    try {
      if (data.isNode()) {
        if (deleteNode(cid) <= 0) {
          // [PN] it's error state, as we actually didn't delete the node (with sub-nodes)
          log.warn("Error state, the node (with sub-nodes) is not deleted " + data.getQPath().getAsString());
        }
      } else {
        if (deleteProperty(cid) <= 0) {
          // [PN] it's error state, as we actually didn't delete the property
          log.warn("Error state, the property item is not deleted " + data.getQPath().getAsString());
        }
      }

      if (log.isDebugEnabled())
        if (data.isNode())
          log.debug("Node deleted " + data.getQPath().getAsString() + ", " + data.getUUID() + ", " + ((NodeData) data).getPrimaryTypeName().getAsString());
        else
          log.debug("Property deleted " + data.getQPath().getAsString() + ", " + data.getUUID()
              + (((PropertyData) data).getValues() != null ? ", values count: " + ((PropertyData) data).getValues().size() : ", NULL data"));

    } catch (SQLException e) {
      log.error("Item remove. Database error: " + e, e);
      exceptionHandler.handleDeleteException(e, data);
    }
  }

  /**
   * @param cid, must be ready for container (e.g. if single-db then with prefix)
   */
  protected int deleteNode(String cid) throws SQLException, RepositoryException {
    int deleted = 0;

    // deleting from NODE table
    if (deleteNodeByUUID(cid) <= 0) {
      // [PN] it's error state, as we actually didn't delete the node
      log.warn("Error state, the node actually not deleted " + cid);
    }

    // deleting from ITEM table
    int nc = deleteItemByUUID(cid);
    if (nc <= 0) {
      // [PN] it's error state, as we actually didn't delete a item corresponding the node
      log.warn("Error state, a item corresponding the node actually not deleted " + cid);
    }
    deleted += nc;
    return deleted;
  }

  protected int deleteProperty(String cid) throws SQLException, RepositoryException {

    PropertyData data = (PropertyData) getItemByUUID(cid); // by container id

    try {
      // delete value
      ValueIOChannel channel = valueStorageProvider.getApplicableChannel(data);
      if (channel != null) {
        channel.delete(data.getUUID()); // by API UUI, not by cid
        channel.close();
      } else {
        deleteValues(cid);
      }

      //if (log.isDebugEnabled())
      //  log.debug("values deleted " + data.getUUID() + " use channel " + channel);

      deleteReference(cid);

      // delete property
      if (deletePropertyByUUID(cid) <= 0) {
        log.warn("Error state, the property actually is not deleted " + cid);
      }

      // delete item
      return deleteItemByUUID(cid);
    } catch (IOException e) {
      exceptionHandler.handleDeleteException(e, data);
      return 0; // will newer returns
    }
  }

  public void doReindex(String itemUuid, String oldQPath, int indexDelimPos, String newIndexStr) throws RepositoryException, UnsupportedOperationException, InvalidItemStateException, IllegalStateException, SQLException {

    // child nodes
    ResultSet descendantNodes = findDescendantNodes(itemUuid, oldQPath);
    while (descendantNodes.next()) {
      String oldDescPath = descendantNodes.getString(COLUMN_PATH);
      String newDescPath = oldDescPath.substring(0, indexDelimPos + 1) + newIndexStr
        + oldDescPath.substring(indexDelimPos + newIndexStr.length() + 1);
      if (updateItemPathByUUID(newDescPath, descendantNodes.getInt(COLUMN_VERSION) + 1, descendantNodes.getString(COLUMN_ID)) <= 0)
        log.warn("No nodes was updated during reindex " + oldDescPath + " -> " + newDescPath + ", uuid:" + descendantNodes.getString(COLUMN_ID));
      else {
        // DO IT RECURSIVE
        doReindex(descendantNodes.getString(COLUMN_ID), oldQPath, indexDelimPos, newIndexStr);

        if (log.isDebugEnabled())
          log.debug("Reindex node " + oldDescPath + " -> " + newDescPath + ", " + descendantNodes.getString(COLUMN_ID));
      }
    }
    descendantNodes.close();

    // child properties
    ResultSet descendantProperties = findDescendantProperties(itemUuid, oldQPath);
    while (descendantProperties.next()) {
      String oldDescPath = descendantProperties.getString(COLUMN_PATH);
      String newDescPath = oldDescPath.substring(0, indexDelimPos + 1) + newIndexStr
        + oldDescPath.substring(indexDelimPos + newIndexStr.length() + 1);
      if (updateItemPathByUUID(newDescPath, descendantProperties.getInt(COLUMN_VERSION) + 1, descendantProperties.getString(COLUMN_ID)) <= 0)
        log.warn("No nodes was updated during reindex " + oldDescPath + " -> " + newDescPath + ", uuid:" + descendantProperties.getString(COLUMN_ID));
      else if (log.isDebugEnabled())
        log.debug("Reindex property " + oldDescPath + " -> " + newDescPath + ", " + descendantProperties.getString(COLUMN_ID));
    }
    descendantProperties.close();
  }

  public void reindex(NodeData oldData, NodeData data) throws RepositoryException, UnsupportedOperationException, InvalidItemStateException, IllegalStateException {
    checkIfOpened();
    try {
      String cid = getInternalId(data.getUUID());

      String oldQPath = oldData.getQPath().getAsString();
      String newQPath = data.getQPath().getAsString();

      int indexDelimPos = oldQPath.lastIndexOf(":");
      if (newQPath.charAt(indexDelimPos) != ':')
        throw new RepositoryException("Reindex. An old and a new node has a different locations in workspace. "
            + oldQPath + ", " + newQPath);

      String newIndexStr = newQPath.substring(indexDelimPos + 1);

      // reindex node
      if (updateItemPathByUUID(newQPath, data.getPersistedVersion() + 1, cid) <= 0)
        log.warn("No nodes was updated during reindex " + oldQPath + " -> " + newQPath + ", uuid:" + cid);
      else if (log.isDebugEnabled())
        log.debug("Reindex root node " + oldQPath + " -> " + newQPath + ", " + cid + ", " + data.getPrimaryTypeName().getAsString());

      // reindex childs
      doReindex(cid, oldQPath, indexDelimPos, newIndexStr);

    } catch (RepositoryException e) {
      throw new RepositoryException(e);
    } catch (Exception e) { // SQL
      log.error("Node reindex. Database error: " + e, e);
      exceptionHandler.handleUpdateException(e, data);
    }
  }

  /* (non-Javadoc)
   * @see org.exoplatform.services.jcr.storage.WorkspaceStorageConnection#update(org.exoplatform.services.jcr.datamodel.NodeData)
   */
  public void update(NodeData data) throws RepositoryException, UnsupportedOperationException, InvalidItemStateException, IllegalStateException {
    checkIfOpened();
    try {
      String cid = getInternalId(data.getUUID());
      // version update
      updateItemVersionByUUID(data.getPersistedVersion(), cid);
      // order numb update
      updateNodeOrderNumbByUUID(data.getOrderNumber(), cid);

      if (log.isDebugEnabled())
        log.debug("Node updated " + data.getQPath().getAsString() + ", " + data.getUUID() + ", " + data.getPrimaryTypeName().getAsString());

    } catch (SQLException e) {
      log.error("Node update. Database error: " + e, e);
      exceptionHandler.handleUpdateException(e, data);
    }
  }

  /* (non-Javadoc)
   * @see org.exoplatform.services.jcr.storage.WorkspaceStorageConnection#update(org.exoplatform.services.jcr.datamodel.PropertyData)
   */
  public void update(PropertyData data) throws RepositoryException, UnsupportedOperationException,
  InvalidItemStateException, IllegalStateException {
    checkIfOpened();

    try {
      String cid = getInternalId(data.getUUID());

      // version update
      updateItemVersionByUUID(data.getPersistedVersion(), cid);

      // update type
      updatePropertyTypeByUUID(data.getType(), cid);

      // update reference
      try {
        deleteReference(cid);

        if (data.getType() == PropertyType.REFERENCE) {
          addReference(data);
        }
      } catch(IOException e) {
        throw new RepositoryException("Can't update REFERENCE property ("+data.getQPath()+" "+data.getUUID()+") value: " + e.getMessage(), e);
      }

      ValueIOChannel channel = valueStorageProvider.getApplicableChannel(data);
      if (channel != null) {
        channel.delete(data.getUUID()); // by API UUID, not by cid
        channel.write(data.getUUID(), data.getValues());
        channel.close();
      } else {
        deleteValues(cid);
        addValues(cid, data.getValues());
      }

      if (log.isDebugEnabled())
        log.debug("Property updated " + data.getQPath().getAsString() + ", " + data.getUUID()
            + (data.getValues() != null ? ", values count: " + data.getValues().size() : ", NULL data")
            + ", use channel "+channel);

    } catch (IOException e) {
      log.error("Property update. IO error: " + e, e);
      exceptionHandler.handleUpdateException(e, data);
    } catch (SQLException e) {
      log.error("Property update. Database error: " + e, e);
      exceptionHandler.handleUpdateException(e, data);
    }
  }

  /* (non-Javadoc)
   * @see org.exoplatform.services.jcr.storage.WorkspaceStorageConnection#getChildNodesData(org.exoplatform.services.jcr.datamodel.NodeData)
   */
  public List<NodeData> getChildNodesData(NodeData parent) throws RepositoryException, IllegalStateException {
    checkIfOpened();
    try {
      ResultSet node = findChildNodesByParentUUID(getInternalId(parent.getUUID()));
      List<NodeData> childrens = new ArrayList<NodeData>();
      while(node.next()) {
        if (node.getString(COLUMN_NID) != null) {
          childrens.add(loadNodeRecord(node));
        } else {
          throw new RepositoryException("FATAL: Not found child node for parent "+parent.getQPath().getAsString()
              + ", but child item found " + node.getString(COLUMN_PATH) + " " + getUuid(node.getString(COLUMN_ID)));
        }
      }
      return childrens;
    } catch (SQLException e) {
      e.printStackTrace();
      throw new RepositoryException(e);
    }
  }

  /* (non-Javadoc)
   * @see org.exoplatform.services.jcr.storage.WorkspaceStorageConnection#getChildPropertiesData(org.exoplatform.services.jcr.datamodel.NodeData)
   */
  public List<PropertyData> getChildPropertiesData(NodeData parent) throws RepositoryException, IllegalStateException {
    checkIfOpened();
    try {
      ResultSet prop = findChildPropertiesByParentUUID(getInternalId(parent.getUUID()));
      List<PropertyData> children = new ArrayList<PropertyData>();
      while(prop.next()) {
        if (prop.getString(COLUMN_PID) != null) {
          children.add(loadPropertyRecord(prop));
        } else {
          throw new RepositoryException("FATAL: Not found child property for parent "+parent.getQPath().getAsString()
              + ", but child item found " + prop.getString(COLUMN_PATH) + " " + getUuid(prop.getString(COLUMN_ID)));
        }

      }
      return children;
    } catch (SQLException e) {
      e.printStackTrace();
      throw new RepositoryException(e);
    } catch (IOException e) {
      e.printStackTrace();
      throw new RepositoryException(e);
    }
  }

  /* (non-Javadoc)
   * @see org.exoplatform.services.jcr.storage.WorkspaceStorageConnection#getItemData(org.exoplatform.services.jcr.datamodel.InternalQPath)
   */
  public ItemData getItemData(InternalQPath qPath) throws RepositoryException, IllegalStateException {
    checkIfOpened();
    ResultSet item = null;
    try {
      item = findItemByPath(qPath.getAsString());
      if (item.next())
        return itemData(item);
      return null;
    } catch (SQLException e) {
      e.printStackTrace();
      throw new RepositoryException(e);
    } catch (IOException e) {
      e.printStackTrace();
      throw new RepositoryException(e);
    } finally {
      try {
        if (item != null)
          item.close();
      } catch(SQLException e) {
        log.error("getItemData() Error close resultset " + e.getMessage());
      }
    }
  }

  /* (non-Javadoc)
   * @see org.exoplatform.services.jcr.storage.WorkspaceStorageConnection#getItemData(java.lang.String)
   */
  public ItemData getItemData(String uuid) throws RepositoryException, IllegalStateException {
    return getItemByUUID(getInternalId(uuid));
  }

  /* (non-Javadoc)
   * @see org.exoplatform.services.jcr.storage.WorkspaceStorageConnection#getReferencesData(java.lang.String)
   */
  public List<PropertyData> getReferencesData(String nodeUUID) throws RepositoryException, IllegalStateException {
    checkIfOpened();
    try {
      ResultSet refProps = findReferences(getInternalId(nodeUUID));
      List<PropertyData> references = new ArrayList<PropertyData>();
      while(refProps.next()) {
        references.add(loadPropertyRecord(refProps));
      }
      return references;
    } catch (SQLException e) {
      e.printStackTrace();
      throw new RepositoryException(e);
    } catch (IOException e) {
      e.printStackTrace();
      throw new RepositoryException(e);
    }
  }

  // ------------------ Private methods ---------------


  protected ItemData getItemByUUID(String cid) throws RepositoryException, IllegalStateException {
    checkIfOpened();
    ResultSet item = null;
    try {
      item = findItemByUUID(cid);
      if(item.next()) {
        return itemData(item);
      }
      return null;
    } catch (SQLException e) {
      throw new RepositoryException("getItemData() error", e);
    } catch (IOException e) {
      throw new RepositoryException("getItemData() error", e);
    } finally {
        try {
          if (item != null)
            item.close();
        } catch(SQLException e) {
          log.error("getItemData() Error close resultset " + e.getMessage());
        }
    }
  }

  /**
   * @param itemRecord
   * @return
   * @throws RepositoryException
   * @throws SQLException
   */
  private ItemData itemData(ResultSet itemRecord) throws RepositoryException, SQLException, IOException {
    String id = itemRecord.getString(COLUMN_ID);

    boolean isNode = itemRecord.getString(COLUMN_NID) != null;
    boolean isProperty = itemRecord.getString(COLUMN_PID) != null;

    if (isNode && isProperty) {
      // inconsistent data stored in db
      throw new RepositoryException("FATAL: Item stored as node and as property "+itemRecord.getString(COLUMN_PATH) + " " + getUuid(id));
    }

    if (!isNode && !isProperty) {
      // no data stored for item - node or property
      throw new RepositoryException("FATAL: Not found corresponding node nor property for item "+itemRecord.getString(COLUMN_PATH) + " " + getUuid(id));
    }

    if (isNode)
      return loadNodeRecord(itemRecord);

    // property
    return loadPropertyRecord(itemRecord);
  }

  /**
   * Prepared statemens at statr time as it recommends by JDBC API doc
   * @throws SQLException
   */
  protected final void prepareStatements() throws SQLException {
    findItemById = dbConnection.prepareStatement(FIND_ITEM_BY_ID);
    findItemByPath = dbConnection.prepareStatement(FIND_ITEM_BY_PATH);
    findChildItemByPath = dbConnection.prepareStatement(FIND_CHILDITEM_BY_PATH);
    findChildPropertyByPath = dbConnection.prepareStatement(FIND_CHILD_PROPERTY_BY_PATH);

    findDescendantNodes = dbConnection.prepareStatement(FIND_DESCENDANT_NODES_LIKE_PATH);
    findDescendantProperties = dbConnection.prepareStatement(FIND_DESCENDANT_PROPERTIES_LIKE_PATH);

    //findPropertiesByValue = dbConnection.prepareStatement(FIND_PROPERTIES_BY_VALUE);

    findReferences = dbConnection.prepareStatement(FIND_REFERENCES);
    findReferenceable = dbConnection.prepareStatement(FIND_REFERENCEABLE);

    findValuesByPropertyId = dbConnection.prepareStatement(FIND_VALUES_BY_PROPERTYID);
    findValueByPropertyIdOrderNumber = dbConnection.prepareStatement(FIND_VALUE_BY_PROPERTYID_OREDERNUMB);

    findNodesByParentId = dbConnection.prepareStatement(FIND_NODES_BY_PARENTID);
    findNodesIdsByParentId = dbConnection.prepareStatement(FIND_NODES_IDS_BY_PARENTID);
    findPropertiesByParentId = dbConnection.prepareStatement(FIND_PROPERTIES_BY_PARENTID);

    findNodesCountByParentId = dbConnection.prepareStatement(FIND_NODESCOUNT_BY_PARENTID);
    findPropertiesCountByParentId = dbConnection.prepareStatement(FIND_PROPERTIESCOUNT_BY_PARENTID);

    insertItem = dbConnection.prepareStatement(INSERT_ITEM);
    insertNode = dbConnection.prepareStatement(INSERT_NODE);
    insertProperty = dbConnection.prepareStatement(INSERT_PROPERTY);
    insertValue = dbConnection.prepareStatement(INSERT_VALUE);
    insertReference = dbConnection.prepareStatement(INSERT_REF);

    updateItem = dbConnection.prepareStatement(UPDATE_ITEM);
    updateItemPath = dbConnection.prepareStatement(UPDATE_ITEM_PATH);
    updateNode = dbConnection.prepareStatement(UPDATE_NODE);
    updateProperty = dbConnection.prepareStatement(UPDATE_PROPERTY);
    //updateValue = dbConnection.prepareStatement(UPDATE_VALUE);

    deleteItem = dbConnection.prepareStatement(DELETE_ITEM);
    deleteNode = dbConnection.prepareStatement(DELETE_NODE);
    deleteProperty = dbConnection.prepareStatement(DELETE_PROPERTY);
    deleteValue = dbConnection.prepareStatement(DELETE_VALUE);
    deleteReference = dbConnection.prepareStatement(DELETE_REF);
  }

  protected PersistedNodeData loadNodeRecord(ResultSet item) throws RepositoryException, SQLException {

    String cNID = item.getString(COLUMN_NID);
    String cNPARENTID = item.getString(COLUMN_NPARENTID);
    int cVERSION = item.getInt(COLUMN_VERSION);
    int cNORDERNUM = item.getInt(COLUMN_NORDERNUM);
    InternalQPath qpath = InternalQPath.parse(item.getString(COLUMN_PATH));

    // PRIMARY
    InternalQPath ptPath = InternalQPath.makeChildPath(qpath, Constants.JCR_PRIMARYTYPE);

    ResultSet ptProp = findPropertyByPath(cNID, ptPath.getAsString());

    if (!ptProp.next())
    //if (idPrimaryType == null)
      throw new PrimaryTypeNotFoundException("FATAL ERROR primary type record not found "
          + ptPath.getAsString() + ", parent id " + cNID +  ", container "+this.containerName, ptPath);

    ResultSet ptValue = findValuesByPropertyId(ptProp.getString(COLUMN_PID));

    if (!ptValue.next())
      throw new RepositoryException("FATAL ERROR primary type value not found "
          + ptPath.getAsString() + ", id " + ptProp.getString(COLUMN_ID) + ", container " + this.containerName);

    byte[] data = ptValue.getBytes(COLUMN_DATA);
    InternalQName ptName = InternalQName.parse(new String((data != null ? data : new byte[] {})));

    // MIXIN
    InternalQPath mtPath = InternalQPath.makeChildPath(qpath, Constants.JCR_MIXINTYPES);
    ResultSet mtProp = findPropertyByPath(cNID, mtPath.getAsString());

    InternalQName[] mixinNames = new InternalQName[0];
    if(mtProp.next()) {
      List<byte[]> mts = new ArrayList<byte[]>();
      ResultSet mtValues = findValuesByPropertyId(mtProp.getString(COLUMN_PID));
      while (mtValues.next()) {
        mts.add(mtValues.getBytes(COLUMN_DATA));
      }
      mixinNames = new InternalQName[mts.size()];
      for (int i = 0; i < mts.size(); i++) {
        mixinNames[i] = InternalQName.parse(new String(mts.get(i)));
      }
    }

    // ACL
    AccessControlList acl;
    if (isAccessControllable(mixinNames)) {

      InternalQPath ownerPath = InternalQPath.makeChildPath(qpath, Constants.EXO_OWNER);

      PropertyData ownerData = (PropertyData) getItemData(ownerPath);

      InternalQPath permPath = InternalQPath.makeChildPath(qpath, Constants.EXO_PERMISSIONS);

      PropertyData permData = (PropertyData) getItemData(permPath);

      acl = new AccessControlList(ownerData, permData);
    } else {
      if(qpath.equals(Constants.ROOT_PATH)) {
        // make default ACL for root
        acl = new AccessControlList(true);
      } else {
        acl = null;
      }
    }
//    System.out.println("--- load Node data------------>>>>> "+cNID+" "+containerName+" "+cNPARENTID);

    return new PersistedNodeData(getUuid(cNID),
        qpath,
        getUuid(cNPARENTID),
        cVERSION,
        cNORDERNUM,
        ptName, mixinNames, acl);
  }


  private boolean isAccessControllable(InternalQName[] mixinNames) {
    for (int i = 0; i < mixinNames.length; i++) {
      if (mixinNames[i].getAsString().equals(Constants.ACCESS_TYPE_URI)) {
        return true;
      }
    }
    return false;
  }


  protected PersistedPropertyData loadPropertyRecord(ResultSet item) throws RepositoryException, SQLException, IOException {

    List<ValueData> values = new ArrayList<ValueData>();
    InternalQPath path = InternalQPath.parse(item.getString(COLUMN_PATH));
    String cid = item.getString(COLUMN_PID);
    String uuid = getUuid(cid);
    PersistedPropertyData pdata = new PersistedPropertyData(uuid,
        path,
        getUuid(item.getString(COLUMN_PPARENTID)),
        item.getInt(COLUMN_VERSION),
        item.getInt(COLUMN_PTYPE),
        item.getBoolean(COLUMN_PMULTIVALUED)
    );

    ValueIOChannel channel = valueStorageProvider.getApplicableChannel(pdata);
    if (channel != null) {
      values = channel.read(uuid, this.maxBufferSize);
      channel.close();
    } else {
      values = readValues(cid);
    }

    pdata.setValues(values);
    return pdata;
  }

  private List<ValueData> readValues(String cid) throws IOException {

    List<ValueData> data = new ArrayList<ValueData>();

    try {

      //final PreparedStatement findValue = dbConnection.prepareStatement(FIND_VALUES_BY_PROPERTYID);

      final ResultSet valueRecords = findValuesByPropertyId(cid);
      try {
        // [PN] 12.12.06
        //findValue.setString(1, cid);
        //final ResultSet valueRecords = findValue.executeQuery();
        while (valueRecords.next()) {
          final int orderNum = valueRecords.getInt(COLUMN_ORDERNUM);

          ValueData vdata = readValueData(cid, orderNum);
          data.add(vdata);
        }
      } finally {
        valueRecords.close();
      }

    } catch (SQLException e) {
      String msg = "Can't read value data of property with id " + cid + ", error:" + e;
      log.error(msg, e);
      throw new IOException(msg);
    }

    return data;
  }

  protected ValueData readValueData(String cid, int orderNumber) throws SQLException, IOException {

    InputStream in = readDataStream(cid, orderNumber);

    byte[] buffer = new byte[0];
    byte[] spoolBuffer = new byte[0x2000];
    int read;
    int len = 0;
    OutputStream out = null;
    File spoolFile = null;
    try {
      if (in != null) 
        while ((read = in.read(spoolBuffer)) > 0) {
          if (out != null) {
            // spool to temp file
            out.write(spoolBuffer, 0, read);
            len += read;
          } else if (len + read > maxBufferSize) {
            // threshold for keeping data in memory exceeded;
            // create temp file and spool buffer contents
            spoolFile = new File(swapDirectory, cid+orderNumber);
            out = new FileOutputStream(spoolFile);
            out.write(buffer, 0, len);
            out.write(spoolBuffer, 0, read);
            buffer = null;
            len += read;
          } else {
            // reallocate new buffer and spool old buffer contents
            byte[] newBuffer = new byte[len + read];
            System.arraycopy(buffer, 0, newBuffer, 0, len);
            System.arraycopy(spoolBuffer, 0, newBuffer, len, read);
            buffer = newBuffer;
            len += read;
          }
        }
    } finally {
      if (out != null) {
        out.close();
      }
    }

    if(buffer == null)
      return new CleanableFileStreamValueData(spoolFile, orderNumber, false, swapCleaner);
    else
      return new ByteArrayPersistedValueData(buffer, orderNumber);
  }

  private InputStream readDataStream(String cid, int orderNumber) throws SQLException {

    // [PN] 12.12.06 - leak of database mem, a result set is not closed, stmt creating one new each time

//    final PreparedStatement findValueData = dbConnection.prepareStatement(FIND_VALUE_BY_PROPERTYID_OREDERNUMB);
//
//    findValueData.setString(1, cid);
//    findValueData.setInt(2, orderNumber);
//    final ResultSet vrs = findValueData.executeQuery();
//
//    if (vrs.next()) {
//      return vrs.getBinaryStream(COLUMN_DATA);
//      //return new JDBCIOHelperInputStream(vrs.getBinaryStream(COLUMN_DATA), findValueData);
//    }
//
//    return new ByteArrayInputStream(new byte[] {});

      final ResultSet vrs = findValueByPropertyIdOrderNumber(cid, orderNumber);
      try {
        if (vrs.next()) {
          return vrs.getBinaryStream(COLUMN_DATA);
        }
      } finally {
        vrs.close();
      }

      return new ByteArrayInputStream(new byte[] {});
  }

  protected void addValues(String cid, List<ValueData> data) throws IOException, SQLException {
    if(data == null) {
      log.warn("List of values data is NULL. Check JCR logic. PropertyId: " + getUuid(cid));
      return;
    }

    for (int i = 0; i < data.size(); i++) {
      addValueRecord(cid, data.get(i), i); // TODO data.get(i).getOrderNumber()
    }
  }

  protected void addValueRecord(String cid, ValueData data, int orderNumber) throws SQLException, IOException {

    InputStream stream = null;
    int streamLength = 0;
    if (data.isByteArray()) {
      byte[] dataBytes = data.getAsByteArray();
      stream = new ByteArrayInputStream(dataBytes);
      streamLength = dataBytes.length;
    } else {
      stream = data.getAsStream();
      streamLength = stream.available(); // for FileInputStream can be used channel.size() result
    }

    insertValue.setBinaryStream(1, stream, streamLength);

    data.setOrderNumber(orderNumber);
    insertValue.setInt(2, data.getOrderNumber());

    insertValue.setString(3, cid);
    insertValue.executeUpdate();
  }

  protected void deleteValues(String cid) throws SQLException {
    deleteValue.setString(1, cid);
    deleteValue.executeUpdate();
  }

  protected ResultSet findValuesByPropertyId(String cid) throws SQLException {
    findValuesByPropertyId.setString(1, cid);
    return findValuesByPropertyId.executeQuery();
  }

  protected ResultSet findValueByPropertyIdOrderNumber(String cid, int orderNumb) throws SQLException {
    findValueByPropertyIdOrderNumber.setString(1, cid);
    findValueByPropertyIdOrderNumber.setInt(2, orderNumb);
    return findValueByPropertyIdOrderNumber.executeQuery();
  }

  // ---- Data access methods (query wrappers) to override in concrete connection ------

  protected abstract void addNodeRecord(NodeData data) throws SQLException;

  protected abstract void addPropertyRecord(PropertyData prop) throws SQLException;

  protected abstract ResultSet findItemByPath(String path) throws SQLException;

  @Deprecated
  protected abstract ResultSet findItemByPath(String parentId, String path) throws SQLException;

  protected abstract ResultSet findPropertyByPath(String parentId, String path) throws SQLException;

  protected abstract ResultSet findDescendantNodes(String parentId, String parentPath) throws SQLException;

  protected abstract ResultSet findDescendantProperties(String parentId, String parentPath) throws SQLException;

  protected abstract ResultSet findItemByUUID(String uuid) throws SQLException;

  protected abstract void addReference(PropertyData data) throws SQLException, IOException;

  protected abstract void deleteReference(String propertyUuid) throws SQLException;

  protected abstract ResultSet findReferences(String nodeUuid) throws SQLException;

  protected abstract ResultSet findReferenceable(String propertyUuid) throws SQLException;

  protected abstract ResultSet getChildNodesCountByParentUUID(String parentUUID) throws SQLException;

  protected abstract ResultSet getChildPropertiesCountByParentUUID(String parentUUID) throws SQLException;

  protected abstract ResultSet findChildNodesByParentUUID(String parentUUID) throws SQLException;

  // [PN] 22.05.06 For deleteNode() method
  protected abstract ResultSet findChildNodesIdsByParentUUID(String parentUUID) throws SQLException;

  protected abstract ResultSet findChildPropertiesByParentUUID(String parentUUID) throws SQLException;

  protected abstract int deleteItemByUUID(String uuid) throws SQLException;
  protected abstract int deleteNodeByUUID(String uuid) throws SQLException;
  protected abstract int deletePropertyByUUID(String uuid) throws SQLException;

  protected abstract int updateItemVersionByUUID(int versionValue, String uuid) throws SQLException;
  protected abstract int updateItemPathByUUID(String qpath, int version, String uuid) throws SQLException;
  protected abstract int updateNodeOrderNumbByUUID(int orderNumb, String uuid) throws SQLException;
  protected abstract int updatePropertyTypeByUUID(int type, String uuid) throws SQLException;


}
