/**
 * Copyright 2001-2006 The eXo Platform SARL         All rights reserved.  *
 * Please look at license.txt in info directory for more license detail.   *
 */

package org.exoplatform.services.jcr.impl.storage.jdbc;

import java.io.File;
import java.io.IOException;
import java.sql.SQLException;

import javax.jcr.RepositoryException;
import javax.naming.Context;
import javax.naming.InitialContext;
import javax.naming.NamingException;
import javax.sql.DataSource;

import org.apache.commons.logging.Log;
import org.exoplatform.services.database.jdbc.DBSchemaCreator;
import org.exoplatform.services.jcr.config.RepositoryConfigurationException;
import org.exoplatform.services.jcr.config.WorkspaceEntry;
import org.exoplatform.services.jcr.impl.storage.WorkspaceDataContainerBase;
import org.exoplatform.services.jcr.impl.storage.jdbc.update.StorageUpdateManager;
import org.exoplatform.services.jcr.impl.storage.value.StandaloneStoragePluginProvider;
import org.exoplatform.services.jcr.impl.util.io.FileCleaner;
import org.exoplatform.services.jcr.storage.WorkspaceStorageConnection;
import org.exoplatform.services.jcr.storage.value.ValueStoragePluginProvider;
import org.exoplatform.services.log.ExoLogger;
import org.picocontainer.Startable;

/**
 * Created by The eXo Platform SARL .
 * 
 * @author Gennady Azarenkov
 * @version $Id: JDBCWorkspaceDataContainer.java 12924 2007-02-21 22:32:12Z peterit $
 */
public class JDBCWorkspaceDataContainer extends WorkspaceDataContainerBase 
 implements Startable{

  public final static String CONTAINER_NAME = "containerName";
  public final static String SOURCE_NAME = "sourceName";
  public final static String MULTIDB = "multi-db";
  public final static String MAXBUFFERSIZE = "max-buffer-size";
  public final static String SWAPDIR = "swap-directory";
  
  public final static int DEF_MAXBUFFERSIZE = 1024 * 200; // 200k
  public final static String DEF_SWAPDIR = System.getProperty("java.io.tmpdir");
  
  protected static Log log = ExoLogger.getLogger("jcr.JDBCWorkspaceDataContainer");
  
  protected String containerName;
  protected String sourceName;
  protected boolean multiDb;
  private DataSource dataSource;
  
  private String storageVersion;
  
  private final ValueStoragePluginProvider valueStorageProvider;
  
  protected int maxBufferSize;
  protected File swapDirectory;
  
  private FileCleaner swapCleaner;
  

  /**
   * Constructor with value storage plugins
   * @param wsConfig
   * @param dbCreator
   * @param valueStrorageProvider
   * @throws RepositoryConfigurationException
   * @throws NamingException
   */
  public JDBCWorkspaceDataContainer(WorkspaceEntry wsConfig, 
        DBSchemaCreator dbCreator,
        ValueStoragePluginProvider valueStorageProvider) 
    throws RepositoryConfigurationException, NamingException, RepositoryException, IOException {
    
    this.containerName = wsConfig.getName();
    this.sourceName = wsConfig.getContainer().getParameterValue(SOURCE_NAME);    
    this.multiDb = Boolean.parseBoolean(wsConfig.getContainer().getParameterValue(MULTIDB));
    this.valueStorageProvider = valueStorageProvider;
    
    try {
      String bsParam = wsConfig.getContainer().getParameterValue(MAXBUFFERSIZE);
      this.maxBufferSize = Integer.parseInt(bsParam);
    } catch (RepositoryConfigurationException e) {
      this.maxBufferSize = DEF_MAXBUFFERSIZE;
    }

    try {
      String sdParam = wsConfig.getContainer().getParameterValue(SWAPDIR);
      this.swapDirectory = new File(sdParam);
    } catch (RepositoryConfigurationException e1) {
      this.swapDirectory = new File(DEF_SWAPDIR);
    }
    if(!swapDirectory.exists())
      swapDirectory.mkdirs();

    Context context = new InitialContext();
    dataSource = (DataSource) context.lookup(sourceName);
    
    String suParam = null;
    boolean enableStorageUpdate = false;
    try {
      suParam = wsConfig.getContainer().getParameterValue("update-storage");
      enableStorageUpdate = Boolean.parseBoolean(suParam);
    } catch (RepositoryConfigurationException e) {
      log.debug("update-storage parameter is not set "+sourceName);
    }
    
    checkVersion(enableStorageUpdate);
    
    // check for FileValueStorage
    if (valueStorageProvider instanceof StandaloneStoragePluginProvider) {
      WorkspaceStorageConnection conn = null;
      try {
        conn = openConnection();
        ((StandaloneStoragePluginProvider) valueStorageProvider).checkConsistency(conn);
      } finally {
        if (conn != null)
          conn.rollback();
      }
    }
    
    log.info(getInfo());  
//    log.info("JDBCWorkspaceDataContainer (cEntry): " + containerName + ", dataSource: " + sourceName);
  }
  
  public String checkVersion(boolean enableStorageUpdate) throws RepositoryException {
    this.storageVersion = StorageUpdateManager.checkVersion(sourceName, dataSource, multiDb, enableStorageUpdate);
    
    return this.storageVersion;
  }
    
  /* (non-Javadoc)
   * @see org.exoplatform.services.jcr.storage.WorkspaceDataContainer#openConnection()
   */
  public WorkspaceStorageConnection openConnection() throws RepositoryException {
    
    try {

      if (multiDb) 
        return new MultiDbJDBCConnection(dataSource, containerName, multiDb, valueStorageProvider, maxBufferSize, swapDirectory, swapCleaner);
      else
        return new SingleDbJDBCConnection(dataSource, containerName, multiDb, valueStorageProvider, maxBufferSize, swapDirectory, swapCleaner);

    } catch (SQLException e) {
      throw new RepositoryException(e);
    }
  }

  /* (non-Javadoc)
   * @see org.exoplatform.services.jcr.storage.WorkspaceDataContainer#getName()
   */
  public String getName() {
    return containerName;
  }

  /* (non-Javadoc)
   * @see org.exoplatform.services.jcr.storage.WorkspaceDataContainer#getInfo()
   */
  public String getInfo() {
    String str = "JDBC based JCR Workspace Data container \n"+
      "container name: "+containerName+" \n"+
      "data source JNDI name: "+sourceName+"\n"+
      "is multi database: "+multiDb+"\n"+
      "storage version: "+storageVersion+"\n"+
      "value storage provider: "+valueStorageProvider+"\n"+
      "max buffer size (bytes): "+maxBufferSize+"\n"+
      "swap directory path: "+swapDirectory.getAbsolutePath()+"\n"+
      "instance: " + this;
    return str;
  }

  
  /* (non-Javadoc)
   * @see org.exoplatform.services.jcr.storage.DataContainer#getStorageVersion()
   */
  public String getStorageVersion() {
    return storageVersion;
  }

  /* (non-Javadoc)
   * @see org.picocontainer.Startable#start()
   */
  public void start() {
    this.swapCleaner = new FileCleaner();
  }

  /* (non-Javadoc)
   * @see org.picocontainer.Startable#stop()
   */
  public void stop() {
    this.swapCleaner.halt();
    this.swapCleaner.interrupt();
  }

}