/**
 * Copyright (c) 2001-2006 eXo Platform SARL, http://www.exoplatform.com/
 *
 * All rights reserved. Please look at license.txt in info directory for
 * more license detail.
 *
 * -------------------------------------------------------------------------
 * $Id: Loader.java 10466 2006-11-21 19:48:45Z rmarins $
 * -------------------------------------------------------------------------
 */
package org.exoplatform.applet.install;

import java.applet.Applet;
import java.io.File;
import java.net.MalformedURLException;
import java.net.URL;
import java.util.Collection;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Vector;

import org.exoplatform.applet.deploy.Component;
import org.exoplatform.applet.deploy.LauncherConfiguration;
import org.exoplatform.applet.install.event.InstallDispatcher;
import org.exoplatform.applet.install.event.InstallEvent;
import org.exoplatform.applet.logging.Logger;
import org.exoplatform.applet.logging.file.FileLoggerFactory;
import org.exoplatform.applet.plugin.AppletPlugin;
import org.exoplatform.applet.plugin.Pluggable;
import org.exoplatform.applet.security.RestrictedActionFactory;
import org.exoplatform.applet.security.RestrictedActor;

/**
 * @version $Rev: 10466 $ $Date: 2006-11-22 02:48:45 +0700 (Wed, 22 Nov 2006) $
 * @author <a href="mailto:rafael.marins@exoplatform.com">Rafael Marins</a>
 */
public class Loader {

	private static final String LOG_SOURCE_CLASS = "Loader";

	private static final Logger logger = FileLoggerFactory.getInstance().getLogger(Loader.class.getName());

	private LauncherConfiguration config;

	private ClassLoader sharedLoader;
	private Applet realApplet;
	private AppletPlugin[] plugins;
	
    protected Loader(LauncherConfiguration configuration) {
	    super();
	    this.config = configuration;
    }

    public void load(InstallDispatcher dispatcher) {

    	final String log_method = "load".intern();

    	logger.entering(LOG_SOURCE_CLASS, log_method);

    	noteStartLoad(dispatcher);

    	try {

        	ClassLoader shared = prepareSharedLoader();
        	noteLibrariesLoad(dispatcher, shared);
        	logger.finest("Shared Classloader bootstrapped");

        	Component[] comps = getDependenciesByScope(Component.APPLET_SCOPE);

        	Map compLoaders = prepareComponentsLoaders(comps, shared);
        	logger.finest("Components classloaders bootstrapped");

    		this.realApplet = loadApplet(compLoaders);
	    	noteAppletLoad(dispatcher, realApplet);
	    	logger.finest("Real Applet bootstrapped");

	    	this.plugins = loadPlugins(dispatcher, comps, compLoaders);
	    	addAccessoriesToApplet(realApplet);
	    	logger.finest("Accessories added to Real Applet");

        } catch (Exception e) {
        	noteLoadFailure(dispatcher);
        	logger.throwing(LOG_SOURCE_CLASS, log_method, e);
        }

        noteCompleteLoad(dispatcher);
        logger.exiting(LOG_SOURCE_CLASS, log_method);
    }

    /**
     * @param real 
     * 
     */
    private void addAccessoriesToApplet(Applet real) {

    	if (real == null)
    		return;

    	// add loaded plugins to the pluggable interface
    	AppletPlugin[] plugins = getPlugins();
    	
    	if ((real instanceof Pluggable) && (plugins != null) && (plugins.length > 0)) {
	    	for (int i=0; i<plugins.length; i++) {
	    		AppletPlugin p = plugins[i];
	    		((Pluggable) real).addPlugin(p);
	    	}
    	}

    	// set a trusted access instance to an Actor
    	if (real instanceof RestrictedActor) {
    		((RestrictedActor) real).setTrustAccess(
    				RestrictedActionFactory.getInstance());
    	}
    }

    /**
     * @return the distro
     */
    public LauncherConfiguration getLauncherConfiguration() {
	    return config;
    }

    /**
     * @return the plugins
     */
    public AppletPlugin[] getPlugins() {
	    return plugins;
    }

    /**
     * @return the sharedLoader
     */
    public ClassLoader getSharedLoader() {
	    return sharedLoader;
    }

    /**
     * @return the targetApplet
     */
    public Applet getRealApplet() {
	    return realApplet;
    }

	/**
     * @param dispatcher
     * @param targetApplet
     */
    private void noteAppletLoad(InstallDispatcher d, Applet targetApplet) {
    	d.fire(new InstallEvent(targetApplet, InstallEvent.LOAD_APPLET));
    }

	/**
	 * Create a private classloader for each applet component using the shared
	 * classloader as parent.
	 *
     * @param shared
     * @return
     * @throws MalformedURLException 
     */
    private Map prepareComponentsLoaders(Component[] components, ClassLoader shared) throws MalformedURLException {

    	final String log_method = "prepareComponentsLoaders".intern();

    	logger.entering(LOG_SOURCE_CLASS, log_method, "Count: " + components.length);

    	Map loadersTable = new HashMap();

    	for (int i=0; i<components.length; i++) {
    		Component c = components[i];
    		String path = getDependencyInstallPath(
    				config.getInstallBaseDirName(), c);
    		URL[] urls = pathToUrls(path);

    		logger.finest("Bootstrapping classloader for lib: " + path);

    		ClassLoader cl = createClassLoader(urls, shared);

    		loadersTable.put(c.getId(), cl);
    	}

    	logger.exiting(LOG_SOURCE_CLASS, log_method, "Map items: " + loadersTable.size());

	    return loadersTable;
    }

	/**
     * @param dispatcher
     */
    private void noteLoadFailure(InstallDispatcher d) {
    	d.fire(new InstallEvent(this, InstallEvent.LOAD_FAILURE));
    }

	/**
     * @param dispatcher 
	 * @param comps 
	 * @param loaders
     * @return
	 * @throws ClassNotFoundException 
	 * @throws IllegalAccessException 
	 * @throws InstantiationException 
     */
    private AppletPlugin[] loadPlugins(InstallDispatcher dispatcher, Component[] components, Map loaders) throws ClassNotFoundException, InstantiationException, IllegalAccessException {

    	Collection plugins = new Vector();

    	for (int i=0; i<components.length; i++) {

    		Component c = components[i];

    		ClassLoader pluginLoader = (ClassLoader) loaders.get(c.getId());

    		List pluginClassNames = c.getPluginClassNames();

            Iterator it = pluginClassNames.iterator();
            while (it.hasNext()) {
                String className = (String) it.next();

                Class clazz = pluginLoader.loadClass(className);
                AppletPlugin p = (AppletPlugin) clazz.newInstance();

                notePluginLoad(dispatcher, p);

                plugins.add(p);
            }
    	}
    	
	    return (AppletPlugin[]) plugins.toArray(new AppletPlugin[plugins.size()]);
    }

	/**
     * @param componentsLoaders
     * @return
	 * @throws ClassNotFoundException 
	 * @throws InstantiationException 
	 * @throws IllegalAccessException 
     */
    private Applet loadApplet(Map componentsLoaders) throws ClassNotFoundException, InstantiationException, IllegalAccessException {

    	final String log_method = "loadApplet".intern();

    	logger.entering(LOG_SOURCE_CLASS, log_method);

    	ClassLoader appletLoader = (ClassLoader) componentsLoaders.get(
    			config.getAppletDependencyId());

    	logger.finest("Loading applet: " + config.getAppletClassName() +
    			", with classloader : " + appletLoader);

    	Class clazz = null;
        try {
	        clazz = appletLoader.loadClass(config.getAppletClassName());
        } catch (Throwable e) {
        	logger.finest("ERROR: " + e.getMessage());
        	logger.throwing(LOG_SOURCE_CLASS, log_method, e);
//        	throw e;
        	return null;
        }

    	Applet real;
        try {
	        real = (Applet) clazz.newInstance();
        } catch (InstantiationException e) {
        	logger.finest("ERROR: " + e.getMessage());
        	throw e;
        } catch (IllegalAccessException e) {
        	logger.finest("ERROR: " + e.getMessage());
        	throw e;
        }

    	logger.exiting(LOG_SOURCE_CLASS, log_method, real);
    	
	    return real;
    }

	private ClassLoader prepareSharedLoader() throws MalformedURLException {

		final String log_method = "prepareSharedLoader".intern();

		logger.entering(LOG_SOURCE_CLASS, log_method);

    	Component[] libs = getDependenciesByScope(Component.SHARED_SCOPE); 
    	String[] paths = getDependenciesPath(libs);
    	URL[] locations = pathsToUrls(paths);

    	for (int i=0; i<paths.length; i++)
    		logger.finest("Shared Classloader lib: " + paths[i]);

    	ClassLoader cl = createClassLoader(locations, getParentClassLoader());

    	logger.exiting(LOG_SOURCE_CLASS, log_method, cl);

    	return cl;
    }

    private URL[] pathToUrls(String path) throws MalformedURLException {
    	return pathsToUrls(new String[] {path});
    }

    private URL[] pathsToUrls(String[] paths) throws MalformedURLException {

		Collection urls = new Vector();

    	for (int i=0; i<paths.length; i++) {

    		String p = paths[i];
    		File f = new File(p);
    		URL u = f.toURL();

    		urls.add(u);
    	}

    	URL[] result = (URL[]) urls.toArray(new URL[paths.length]);

	    return result;
    }

    /**
     * @param libs
     * @return
     */
    private String[] getDependenciesPath(Component[] libs) {

    	Collection paths = new Vector();

    	for (int i=0; i<libs.length; i++) {

    		Component d = libs[i];

    		String fileInstallPath = getDependencyInstallPath(
    				config.getInstallBaseDirName(), d);

    		paths.add(fileInstallPath);
    	}

	    return (String[]) paths.toArray(new String[paths.size()]);
    }

	/**
     * @param installBaseDirName
     * @param d
     */
    private static String getDependencyInstallPath(String installBaseDirName, Component d) {

		String path = Installer.getInstallPath(
				installBaseDirName,
				d.getScope(),
				d.getGroupId(),
				d.getArtifactId(),
				d.getVersion(),
				d.getClassifier(),
				d.getType());

		return path;
    }

	/**
     * @param scope
     * @return
     */
    private Component[] getDependenciesByScope(String scope) {

    	Collection filtered = new Vector();
    	List deps = config.getComponents();
        Iterator it = deps.iterator();

        while (it.hasNext()) {
            Component d = (Component) it.next();
            if (scope.equals(d.getScope()))
                filtered.add(d);
        }

	    return (Component[]) filtered.toArray(new Component[filtered.size()]);
    }

	private ClassLoader getParentClassLoader() {
    	ClassLoader cl = Thread.currentThread().getContextClassLoader();
    	if (cl == null)
    		cl = this.getClass().getClassLoader();
    	return cl;
    }

    private static ClassLoader createClassLoader(URL[] urls, ClassLoader parent) {
    	ExoClassLoader ecl = new ExoClassLoader(urls, parent);
    	ecl.setParentFirst(false);
    	return ecl;
    }

    private void notePluginLoad(InstallDispatcher d, AppletPlugin plugin) {
    	d.fire(new InstallEvent(plugin, InstallEvent.LOAD_PLUGIN));
    }

    private void noteLibrariesLoad(InstallDispatcher d, ClassLoader shared) {
		d.fire(new InstallEvent(shared, InstallEvent.LOAD_LIBRARIES));
    }

    private void noteStartLoad(InstallDispatcher d) {
    	d.fire(new InstallEvent(this, InstallEvent.LOAD_START));
    }

    private void noteCompleteLoad(InstallDispatcher d) {
    	d.fire(new InstallEvent(this, InstallEvent.LOAD_COMPLETE));
    }
}
