package org.esupportail.cas.server.handlers.nis;

import org.dom4j.Element;
import org.esupportail.cas.server.util.MisconfiguredHandlerException;
import org.esupportail.cas.server.util.RedundantHandler;
import org.esupportail.cas.server.util.crypt.Crypt;

// TODO Nis broadcast
// The NIS domain can not be bound by broadcast. The 'nis:///domain' should work according to 
// http://rzm-hamy-wsx.rz.uni-karlsruhe.de/Training/JNDI/instances/common/nis-provider/doc/providers/jndi-nis.html, 
// but it does not :-(

/**
 * This class implements a NIS (Network Information Service) handler 
 * class. It is used by GenericHandler.
 *
 * @author Pascal Aubry <pascal.aubry at univ-rennes1.fr>
 * @author Jean-Baptiste Daniel <danielj at sourceforge.net>
 */
public final class NisHandler extends RedundantHandler {

	/**
	 * The NIS domain to bind to.
	 */
	private String domain;
	/**
	 * The map to search into.
	 */
	private String map;
	/**
	 * The encryption used to store the passwords.
	 */
	private String encryption;
	
	/**
	 * Constructor.
	 *
	 * @param handlerElement the XML element that declares the handler 
	 * in the configuration file
	 * @param configDebug debugging mode of the global configuration
	 * @throws Exception Exception
	 */
	public NisHandler(
			final Element handlerElement, 
			final Boolean configDebug) throws Exception {
		super(handlerElement, configDebug);
		traceBegin();

		checkConfigElement(true);
		
		domain = getConfigSubElementContent("domain", true/*needed*/);
		trace("domain = " + domain);

		map = getConfigSubElementContent("map", false/*not needed*/);
		if (map.equals("")) {
			map = "passwd.byname";
		}
		trace("map = " + map);

		encryption = getConfigSubElementContent("encryption", false/*not needed*/);
		if (encryption.equals("")) {
			encryption = "pammd5";
		}
		if (!Crypt.isEncryptionSupported(encryption)) {
			traceThrow(new MisconfiguredHandlerException(
					"Encryption \"" 
					+ encryption 
					+ "\" is not supported."));
		}
		trace("encryption = " + encryption);
		if (encryption.equals("plain")) {
			traceThrow(new MisconfiguredHandlerException("Can not use plain passwords with NIS."));
		}

		// add the NisServer instances
		addServers(true/*serverElementNeeded*/, getClass().getPackage().getName() + ".NisServer");

		// check that the JNDI class exists
		checkClass("com.sun.jndi.nis.NISCtxFactory");

		traceEnd();
	}

	/**
	 * Retrieve the map name.
	 *
	 * @return a String Object.
	 */
	String getMap() {
		return map;
	}
	
	/**
	 * Retrieve the NIS domain name.
	 *
	 * @return a String Object.
	 */
	 String getDomain() {
		return domain;
	}
	
	/**
	 * Retrieve the encryption used to store passwords.
	 *
	 * @return a String Object.
	 */
	String getEncryption() {
		return encryption;
	}
	
}

